/* mobility.h
 */

#ifndef EVAL_EXPERIMENTAL_MOBILITY_H
#define EVAL_EXPERIMENTAL_MOBILITY_H

#include "osl/eval/experimental/weights.h"
#include "osl/eval/experimental/midgame.h"
#include "osl/ptype.h"
#include "osl/misc/carray.h"
#include "osl/state/numEffectState.h"
#include "osl/container/tripleInt.h"

namespace osl
{
  namespace eval
  {
    namespace experimental
    {
      class RookMobilityAll
      {
	friend class RookMobility;
	friend class RookMobilityX;
	friend class RookMobilityY;
	friend class RookMobilitySum;
      public:
	template<int Sign>
	static void adjust(bool promoted,
			   int vertical, int horizontal,
			   Position pos, 
			   MultiInt& value);
	static void eval(const NumEffectState&, MultiInt& out);
      private:
	static int indexX(Position rook, bool promoted,
			  int count, bool vertical)
	{
	  const int x = (rook.x() > 5 ?
			 10 - rook.x() : rook.x());
	  return x - 1 + 5 * ((promoted ? 1 : 0) +
			      2 * (vertical ? 1 : 0));
	}
	template <int Sign>
	static int indexY(Position rook, bool promoted,
			  int count, bool vertical)
	{
	  const int y = (Sign > 0 ? rook.y() : 10 - rook.y());
	  return y - 1 + 9 * ((promoted ? 1 : 0) +
			      2 * (vertical ? 1 : 0));
	}
	static CArray<MultiInt, 18> rook_vertical_table;
	static CArray<MultiInt, 18> rook_horizontal_table;
	static CArray<MultiInt, 34> sum_table;
	static CArray<MultiInt, 324> x_table;
	static CArray<MultiInt, 324> y_table;
      };

      class RookMobility
      {
      public:
	enum { DIM = 36 };
	static void setUp(const Weights &weights,int stage);
      };

      class RookMobilitySum
      {
      public:
	enum { ONE_DIM = 34, DIM = ONE_DIM * EvalStages };
	static void setUp(const Weights &weights);
      };
      class RookMobilityX
      {
      public:
	enum { ONE_DIM = 180, DIM = ONE_DIM * EvalStages };
	static void setUp(const Weights &weights);
      };
      class RookMobilityY
      {
      public:
	enum { ONE_DIM = 324, DIM = ONE_DIM * EvalStages };
	static void setUp(const Weights &weights);
      };

      struct BishopMobilityAll
      {
	template<int Sign>
	static void adjust(bool promoted, int index, MultiInt& value);
	static void eval(const NumEffectState&, MultiInt& out);
      };
      class BishopMobility
      {
	friend class BishopMobilityAll;
	static CArray<MultiInt, 36> bishop_table;
      public:
	enum { DIM = 36 };
	BishopMobility() { };
	static void setUp(const Weights &weights,int stage);
      };

      struct LanceMobilityAll
      {
	template<int Sign>
	static void adjust(int index, MultiInt& value);
	static void eval(const NumEffectState&, MultiInt& out);
      };
      class LanceMobility
      {
	static CArray<MultiInt, 9> lance_table;
	friend class LanceMobilityAll;
      public:
	enum { DIM = 9 };
	LanceMobility() { };
	static void setUp(const Weights &weights,int stage);
      };
    }
  }
}
#endif // EVAL_EXPERIMENTAL_MOBILITY_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
