#ifndef _HASH_STATE_H
#define _HASH_STATE_H

#include "osl/hash/hashKey.h"
#include "osl/apply_move/applyDoUndoXMove.h"
#include "osl/ptype.h"
#include "osl/piece.h"
#include "osl/move.h"
#include "osl/misc/carray2d.h"

namespace osl
{
namespace hash
{
  template<class State>
  class HashState : public State
  {
  public:
    /** Ѿ饹 hash ĤФ˻Ѥ */
    typedef HashState hash_state_t;
  private:
    HashKey hash;
    friend class apply_move::ApplyDoUndoSimpleMove<BLACK,HashState>;
    friend class apply_move::ApplyDoUndoCaptureMove<BLACK,HashState>;
    friend class apply_move::ApplyDoUndoDropMove<BLACK,HashState>;
    friend class apply_move::ApplyDoUndoSimpleMove<WHITE,HashState>;
    friend class apply_move::ApplyDoUndoCaptureMove<WHITE,HashState>;
    friend class apply_move::ApplyDoUndoDropMove<WHITE,HashState>;
  public:
    typedef HashKey HashType;
    typedef State base_t;
    
    explicit HashState(const SimpleState& state) 
      : State(state), hash(state)
    {
    }
    virtual ~HashState();
  private:
  public:
    void changeTurn()
    {
      State::changeTurn();
      hash.changeTurn();
    }
    // move methods
    void doSimpleMove(Position from, Position to, int promoteMask);
    void doDropMove(Position to,Ptype ptype);
    void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
    HashKey const& getHash() const{ return hash;}
#ifndef MINIMAL
    bool isConsistent(bool showError=true) const;
    void print(std::ostream&) const;
#endif
  };

  template<class State> inline
  bool operator==(const HashState<State>& s1, const HashState<State>& s2)
  {
    if (s1.getHash() != s2.getHash())
      return false;
    // assert줿
    return true;
  }

  template<class State> inline
  bool operator!=(const HashState<State>& s1, const HashState<State>& s2)
  {
    return !(s1==s2);
  }
} // namespace hash

  namespace apply_move
  {
  /**
   * default Ǥ٤ hash κƷ׻򥵥ܤäƤ 
   */
  template <Player P,typename Base>
  struct ApplyDoUndoSimpleMove<P,hash::HashState<Base> >
  {
    typedef hash::HashState<Base> state_t;
    template <typename F>
    static void doUndoSimpleMove(state_t& s,
				 Position from, Position to, int promoteMask,F& func)
    {
      HashKey oldHash = s.hash;
      s.hash.doSimpleMove(s, from, to, promoteMask);
      s.hash.changeTurn();
      ApplyDoUndoSimpleMove<P,Base>::template doUndoSimpleMove<F>
	(s,from,to,promoteMask,func);
      s.hash = oldHash;
    }
  };
  
  template <Player P,class Base>
  struct ApplyDoUndoDropMove<P,hash::HashState<Base> >
  {
    typedef hash::HashState<Base> state_t;
    template <typename F>
    static void doUndoDropMove(state_t& s,
			       Position to, Ptype ptype,F& func)
    {
      HashKey oldHash = s.hash;
      s.hash.doDropMove(s, to,ptype);
      s.hash.changeTurn();
      ApplyDoUndoDropMove<P,Base>::template doUndoDropMove<F>
	(s,to,ptype,func);
      s.hash = oldHash;
    }
  };

  template <Player P,class Base>
  struct ApplyDoUndoCaptureMove<P,hash::HashState<Base> >
  {
    typedef hash::HashState<Base> state_t;
    template <typename F>
    static void doUndoCaptureMove(state_t& s, Position from, Position to, 
				  Piece target, int promoteMask, F& func)
    {
      HashKey oldHash = s.hash;
      s.hash.doCaptureMove(s, from, to, target, promoteMask);
      s.hash.changeTurn();
      ApplyDoUndoCaptureMove<P,Base>::template doUndoCaptureMove<F>
	(s, from, to, target, promoteMask, func);
      s.hash = oldHash;
    }
  };

  } // namespace hash
} // namespace osl

template<class State>
osl::hash::HashState<State>::~HashState()
{
}

#endif /* _HASH_STATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
