#ifndef _HASH_TCC
#define _HASH_TCC

#include "osl/hash/hashState.h"
#include "osl/apply_move/applyMove.h"
#include "osl/pieceTable.h"
#include <iostream>

#ifndef MINIMAL
template<class State>
bool osl::hash::HashState<State>::isConsistent(bool showError) const
{
  if (! State::isConsistent(showError)) 
  {
    if (showError)
      std::cerr << "error before hash" << std::endl;
    return false;
  }
  if (hash != HashKey(*this))
  {
    if(showError) 
      std::cerr << "hash mismatch" << std::endl;
    return false;
  }
  PieceStand blackStand;
  for (int num=0;num<40;num++)
  {
    if (State::standMask(BLACK).test(num))
      blackStand.add(Piece_Table.getPtypeOf(num));
  }
  if (hash.blackStand() != blackStand)
  {
    if(showError) 
      std::cerr << "stand mismatch" << std::endl;
    return false;
  }
  return true;
}

template<class State>
void osl::hash::HashState<State>::print(std::ostream& os) const {
  os << hash << std::endl;
  os<<static_cast<State>(*this);
}
#endif

template<class State>
void osl::hash::HashState<State>::
doSimpleMove(Position from, Position to, int promoteMask)
{
  hash.doSimpleMove(*this, from,to,promoteMask);
  State::doSimpleMove(from,to,promoteMask);
}

template<class State>
void osl::hash::HashState<State>::
doDropMove(Position to,Ptype ptype)
{
  hash.doDropMove(*this, to,ptype);
  State::doDropMove(to,ptype);
}

template<class State>
void osl::hash::HashState<State>::
doCaptureMove(Position from, Position to, Piece p1,int promoteMask)
{
  hash.doCaptureMove(*this, from,to,p1,promoteMask);
  State::doCaptureMove(from,to,p1,promoteMask);
}

template<class State>
std::ostream& operator<<(std::ostream& os,const osl::hash::HashState<State>& state)
{
  state.print(os);
  return os;
}

#endif /* _HASH_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
