#ifndef _PIECE_H
#define _PIECE_H
#include "osl/misc/loki.h"
#include "osl/player.h"
#include "osl/position.h"
#include "osl/ptype.h"

#include <iosfwd>
namespace osl
{
  class Piece;
  inline bool operator==(Piece l, Piece r);
  const int EMPTY_NUM=0xff;
  const int EDGE_NUM=0x40;
  /**
   * .
   * ptypeo(-15 - 15), ֹ(0-39), ݥ(0-0xff)ʤ 
   * 16 bitptypeo, 8bitֹ, 8bitǥݥȤƤߤ?
   * ɤˤ 16 bitǤϼޤʤ
   */
  class Piece
  {
    int piece;
    Piece(int p) : piece(p)
    {
    }
  public:
    static const int SIZE=40;
    static const Piece makeDirect(int value) { return Piece(value); }
    int intValue() const { return piece; }
    static const Piece EMPTY()  { return Piece(BLACK,PTYPE_EMPTY,EMPTY_NUM,Position::STAND()); }
    static const Piece EDGE() { return Piece(WHITE,PTYPE_EDGE,EDGE_NUM,Position::STAND()); }
    
    Piece(Player owner, Ptype ptype, int num, Position position)
      : piece((static_cast<int>(owner)<<20)
	      |(static_cast<int>(ptype)<<16)
	      |((num)<<8)| position.uintValue())
    {
    }
    Piece() : piece(EMPTY().piece)
    {
    }
    /**
     * ̤
     */
    static const Piece makeKing(Player owner, Position position);

    Ptype ptype() const {
      return static_cast<Ptype>((piece>>16)&0xf);
    }
    PtypeO ptypeO() const {
      return static_cast<PtypeO>(piece>>16);
    }

    int number() const {
      return ((piece&0xff00)>>8);
    }

    const Position position() const {
      return Position::makeDirect(piece&0xff);
    }

    Piece& operator+=(Offset offset) {
      piece += offset.intValue();
      return *this;
    }

    void setPosition(Position position) {
      piece = (piece&0xffffff00)+position.uintValue();
    }
  private:
    bool isOnBoardByOwner(Int2Type<BLACK>) const {
      return static_cast<int>(static_cast<unsigned int>(piece)&0xfff000ff)>0;
    }
    bool isOnBoardByOwner(Int2Type<WHITE>) const {
      return (static_cast<unsigned int>(piece)&0xfff000ff)>0xfff00000;
    }
  public:
    /**
     * piece ץ쥤䡼PλʪǤĥܡɾˤξ true.
     * ŨζäꡤζäꡤPiece::EMPTY(), PIECE_EDGEξ false
     * @param P(template) - ץ쥤䡼
     * @param piece - 
     */
    template<Player P>
    bool isOnBoardByOwner() const { return isOnBoardByOwner(Int2Type<P>()); }
    /**
     * isOnBoardByOwner ̾ؿΥС.
     */
    bool isOnBoardByOwner(Player owner) const
    {
      if (owner == BLACK)
	return isOnBoardByOwner<BLACK>();
      else
	return isOnBoardByOwner<WHITE>();
    }

    /* .  PROMOTEԲĤpieceŬԲ */
    const Piece promote() const {
      assert(canPromote(ptype()));
      return Piece(piece-0x80000);
    }

    /* ᤹.  PROMOTEԲĤpieceŬѲ  */
    const Piece unpromote() const {
      return Piece((int)piece|0x80000);
    }

    /**
     * 줿piece. unpromoteơPosition0
     * λΤˤ
     */
    const Piece captured() const {
      // return (Piece)((((int)piece|0x80000)&0xffffff00)^0xfff00000);
      // optimize
      return Piece((piece&0xfff7ff00)^0xfff80000);
    }

    const Piece promoteWithMask(int promote_mask) const {
      assert(! (isPromoted() && promote_mask));
      return Piece(piece - promote_mask);
    }
    const Piece unpromoteWithMask(int promote_mask) const {
      assert(isPromoted() || (promote_mask == 0));
      return Piece(piece + promote_mask);
    }

    const Piece checkPromote(bool promotep) const {
      return promoteWithMask(promotep<<19);
    }
    const Piece checkUnpromote(bool promotep) const {
      return unpromoteWithMask(promotep<<19);
    }

    /**
     * promote𤫤ɤå
     */
    bool isPromoted() const { return (piece&(1<<19))==0; }

    /**
     * promoteǽʼζpromoteƤ뤫ɤå
     * @param piece - å(, ʳ)
     */
    /**
     * promoteƤʤOnBoardζǤ뤳ȤΥå
     * Lance̤Ȥƻʤ?
     */
    bool isOnBoardNotPromoted() const{
      int mask=piece&((1<<19)|0xff);
      return mask>(1<<19);
    }
    bool isPromotedNotKingGold() const {
      assert(ptype()!=KING && ptype()!=GOLD);
      return isPromoted();
    }

    bool isEmpty() const { return *this == EMPTY(); }
    static bool isEmptyNum(int num) {
      return (num&0x80)!=0;
    }
    bool isEmptyNum() const {
      return isEmptyNum(number());
    }
    bool isEdge() const { return *this == EDGE(); }
    static bool isEdgeNum(int num){
       return (num&0x40)!=0;
    }
    bool isEdgeNum() const { 
      return isEdgeNum(number());
    }
    static bool isPieceNum(int num){
      return (num&0xc0)==0;
    }
    bool isPiece() const {
      return isPieceNum(number());
    }
    /**
     * pieceǤ뤳ȤʬäƤˡBlackɤå롥
     */
    bool pieceIsBlack() const{
      assert(isPiece());
      return static_cast<int>(piece)>=0;
    }
    Player owner() const
    {
      assert(isPiece());
      return static_cast<Player>(piece>>20);
    }

  private:
    /**
     * PIECE_EMPTY 0x0000ff00
     * BLACK_PIECE 0x000Xxxxx X>=2
     * PIECE_EDGE  0xfff14000
     * WHITE_PIECE 0xfffXxxxx X>=2
     * PlayerBLACKλϡ
     * PIECE_EMPTY 0x0000ff00
     * WHITE_PIECE 0xfffXxxxx X>=2
     * TRUEˤʤ褦ˤ
     */
    bool canMoveOn(Int2Type<BLACK>) const {
      return ((static_cast<unsigned int>(EDGE().piece)-static_cast<unsigned int>(EMPTY().piece))
	      < (static_cast<unsigned int>(piece)-(1+static_cast<unsigned int>(EMPTY().piece))));
    }
    /**
     * PlayerWHITEλϡ
     * PIECE_EMPTY 0x0000ff00
     * BLACK_PIECE 0x000Xxxxx X>=2
     * TRUEˤʤ褦ˤ
     */
    bool canMoveOn(Int2Type<WHITE>) const {
      return EMPTY().piece <= piece;
    }
  public:
    /** Player Pζ𤬡thisξ˰ưǤ뤫?
     * @return thisζEMPTYʤtrue
     * @param P 
     */
    template<Player P>
    bool canMoveOn() const { return canMoveOn(Int2Type<P>()); }

    bool canMoveOn(Player pl) const{
      if(pl==BLACK) 
	return canMoveOn<BLACK>();
      else 
	return canMoveOn<WHITE>();
    }

    bool isOnBoard() const {
      assert(position().isValid());
      return ! position().isPieceStand();
    }
    /**
     * pgetPtypeOptypeOȰפƤ0,
     * פƤʤä֤ͤ
     */
    unsigned int ptypeOMask(PtypeO ptypeO) const {
      const int ret= -((piece&0x00ff0000)^
		       ((static_cast<int>(ptypeO)<<16)&0x00ff0000));
      assert( (this->ptypeO()==ptypeO && ret==0) ||
	      (this->ptypeO()!=ptypeO && ret<0));
      return static_cast<unsigned int>(ret);
    }
  };

  inline bool operator<(Piece l, Piece r)
  {
    return l.intValue() < r.intValue();
  }
  inline bool operator==(Piece l, Piece r)
  {
    return l.intValue() == r.intValue();
  }
  inline bool operator!=(Piece l, Piece r)
  {
    return ! (l == r);
  }

  std::ostream& operator<<(std::ostream& os,const Piece piece);
}

#endif /* _PIECE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
