#ifndef _COMPACT_BOARD_H
#define _COMPACT_BOARD_H
#include "osl/state/simpleState.h"
#include "osl/stl/vector.h"
#include <string>

namespace osl
{
  namespace record
  {
    /// Converts a position to an integer (bits)
    int position2Bits(const Position& pos);
    /// Converts an integer (bits) to Position
    Position bits2Position(const int bit_position);

    class OPiece
    {
    public:
      OPiece(Piece p)
      {
	const Position pos = p.position();
	const int bitPos = position2Bits(pos);
	value = (static_cast<int>(p.owner()) << 20 |
		 static_cast<int>(p.ptype()) << 16 | bitPos);
      }
      OPiece(int i)
      {
	value = i;
      }
      Position getPosition() const
      {
        return bits2Position(value);
      }
      Ptype getPtype() const
      {
	return static_cast<Ptype>((value >> 16) & 0xf);
      }
      Player getOwner() const
      {
	return static_cast<Player>(value >> 20);
      }
      operator int() const { return value; }
    private:
      int value;
    };

    class CompactBoard
    {
    public:
      CompactBoard() {}
      explicit CompactBoard(const SimpleState& state);
      SimpleState getState() const;
      const osl::vector<OPiece>& getPieces() const {return pieces;};
      Player getTurn() const {return turn;}
      /**
       * ̤Ӥ. 
       * ֤ˤ¸ǡȤƤζ̡ʼ֤ޤˤӤ. 
       * ⤽CompactBoard϶֤Ȥʤ.
       */
      bool isEqual(const CompactBoard& rhs) const;
      friend std::ostream& operator<<(std::ostream& os, const CompactBoard& c);
      friend std::istream& operator>>(std::istream& os, CompactBoard& c);
    private:
      osl::vector<OPiece> pieces;
      Player turn;
    };

    std::string toBase64(const CompactBoard& c);
    CompactBoard fromBase64(const std::string& str);
  }
}

#endif // _COMPACT_BOARD_H
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
