/* dualThreatmateState.h
 */
#ifndef OSL_SEARCH__DUALTHREATMATESTATE_H
#define OSL_SEARCH__DUALTHREATMATESTATE_H

#include "osl/search/threatmateState.h"
#include <iosfwd>

namespace osl
{
  namespace search
  {
    class DualThreatmateState
    {
      CArray<ThreatmateState,2> king_status;
      ThreatmateState& wstatus(Player king)
      {
	return king_status[king];
      }
    public:
      explicit DualThreatmateState(ThreatmateState::Status b=ThreatmateState::UNKNOWN,
				   ThreatmateState::Status w=ThreatmateState::UNKNOWN) 
      {
	wstatus(BLACK) = b;
	wstatus(WHITE) = w;
      }
      const ThreatmateState& status(Player king) const
      {
	return king_status[king];
      }
      void setThreatmate(Player king, Move m) { 
	wstatus(king).setThreatmate(m);
      }

      bool isThreatmate(Player king) const { 
	return status(king).isThreatmate();
      }
      bool maybeThreatmate(Player king) const { 
	return status(king).maybeThreatmate();
      }
      bool mayHaveCheckmate(Player king) const {
	return status(king).mayHaveCheckmate();
      }
      const Move threatmateMove(Player king) const { 
	return status(king).threatmateMove(); 
      }
      void update(Player turn, const DualThreatmateState *parent, bool in_check)
      {
	if (parent)
	{
	  if (! maybeThreatmate(turn))
	    wstatus(turn).update(&parent->status(turn), in_check);
	  if (! mayHaveCheckmate(alt(turn)))
	    wstatus(alt(turn)).update(&parent->status(alt(turn)), in_check);
	}
      }
    };
    std::ostream& operator<<(std::ostream&, DualThreatmateState);
  }
} // namespace osl

#endif /* OSL_SEARCH__DUALTHREATMATESTATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
