/* numEffectState.h
 */
#ifndef _NUM_EFFECT_STATE_H
#define _NUM_EFFECT_STATE_H

#include "osl/effect/numSimpleEffect.h"
#include "osl/mobility/kingMobility.h"
#include "osl/misc/fastCopier.h"
#include "osl/misc/align16New.h"

namespace osl
{
  namespace state
  {
    class NumEffectState;
    /**
     * ֤˰¸̡ʥ󥹥󥹡Ӥ򤹤.
     * ʤ֤¸ʶӤ򤷤ϡosl::record::CompactBoard
     * osl::hash::HashKeyѤ.
     */
    bool operator==(const NumEffectState& st1, const NumEffectState& st2);

    /**
     * Ķ
     * - effects (NumSimpleEffectTable) 
     * - onBoardMask (PieceMask) ׾ˤ
     */
    class NumEffectState : public SimpleState
#if OSL_WORDSIZE == 32
			 , public misc::Align16New
#endif
    {
      effect::NumSimpleEffectTable effects;
      CArray<PieceMask,2> onBoardMask;
      /**  */
      PieceMask promoted;
      CArray<PieceMask,2> pin_or_open;
      KingMobility king_mobility;
      mutable CArray<unsigned int,2> king8info_cache;
      mutable CArray<bool,2> valid_king8info;

      friend bool operator==(const NumEffectState& st1,const NumEffectState& st2);
      typedef NumEffectState state_t;
      friend class apply_move::ApplyDoUndoSimpleMove<BLACK,NumEffectState>;
      friend class apply_move::ApplyDoUndoCaptureMove<BLACK,NumEffectState>;
      friend class apply_move::ApplyDoUndoDropMove<BLACK,NumEffectState>;
      friend class apply_move::ApplyDoUndoSimpleMove<WHITE,NumEffectState>;
      friend class apply_move::ApplyDoUndoCaptureMove<WHITE,NumEffectState>;
      friend class apply_move::ApplyDoUndoDropMove<WHITE,NumEffectState>;
      friend class osl::misc::FastCopier;
      PieceMask& getOnBoardMask(Player p) {
	return onBoardMask[p];
      }
    public:
      const PieceMask& getOnBoardMask(Player p) const {
	return onBoardMask[p];
      }
      bool isOnBoardNum(int num) const
      {
	return getOnBoardMask(BLACK).test(num) ||
	  getOnBoardMask(WHITE).test(num);
      }

      /** T  isBasic ǤʤƤư unpromote(T) η̤Ʊ.
       * 64bitǤʤϤ
       */ 
      template<Player P,Ptype T,typename F>
      void forEachOnBoard(F& func) const {
	mask_t onMask=getOnBoardMask(P).selectBit<T>() ;    
	while (onMask.any())
	{
	  int num=onMask.takeOneBit()+((PtypeFuns<T>::indexNum)<<5);
	  Piece p = getPieceOf(num);
	  func(p);
	}
      }

      explicit NumEffectState(const state::SimpleState& st=SimpleState(HIRATE));
      ~NumEffectState();
      
      NumBitmapEffect getEffect(Position pos) const
      {
	return effects.getEffect(pos);
      }
      Position getMobility(Direction d,int num) const
      {
	return effects.mobilityTable.get(d,num);
      }
      void setMobility(Direction d,int num,Position pos)
      {
	effects.mobilityTable.set(d,num,pos);
      }
      Position kingMobilityAbs(Player p, Direction d) const
      {
	return Position::makeDirect(king_mobility[p][d]);
      }
      /** 
       * ̤dˤɤޤư뤫֤
       * @param p ܤ̤Υץ쥤
       * @param d piece ߤ
       */
      Position kingMobilityOfPlyaer(Player p, Direction d) const
      {
	if (p == BLACK)
	  d = inverse(d);
	return kingMobilityAbs(p, d);
      }
      const EffectedNumTable& longEffectNumTable() const
      {
	return effects.effectedNumTable;
      }
      /** 
       * piecedĹˤζ֤
       * @param d piece ߤ
       */
      const Piece longEffectOfDirection(Player owner, int piece, Direction d) const
      {
	assert(getPieceOf(piece).isOnBoardByOwner(owner));
	if (owner == BLACK)
	  d = inverse(d);
	const int num = effects.effectedNumTable[piece][d];
	if (num == EMPTY_NUM)
	  return Piece::EMPTY();
	return getPieceOf(num);
      }
      const Piece longEffectOfDirection(Player owner, Piece piece, Direction d) const
      {
	assert(piece.isPiece());
	assert(piece.owner() == owner);
	return longEffectOfDirection(owner, piece.number(), d);
      }
      const Piece longEffectOfDirection(Piece piece, Direction d) const
      {
	assert(piece.isPiece());
	return longEffectOfDirection(piece.owner(), piece, d);
      }
      const Piece longEffectOfDirection(Position square, Direction d) const
      {
	return longEffectOfDirection(getPieceOnBoard(square), d);
      }
      template <Ptype PTYPE>
      const mask_t effectBit(Player P, Position target) const 
      {
	return getEffect(target).selectBit<PTYPE>() & getOnBoardMask(P).getMask<PTYPE>();
      }
      template <Ptype PTYPE> const mask_t selectLong(Position target) const 
      {
	return getEffect(target).selectLong<PTYPE>() >> 8;
      }
      template <Ptype PTYPE> const mask_t selectLong(Position target, Player owner) const 
      {
	return selectLong<PTYPE>(target) & getOnBoardMask(owner).getMask(1);
      }
      const mask_t selectLong(Position target, Player owner) const 
      {
	return (getEffect(target).selectLong() >> 8)
	  & getOnBoardMask(owner).getMask(1);
      }

      const BoardMask changedEffects(Player pl) const{
	assert(! effects.changedEffects(pl).isInvalid());
	return effects.changedEffects(pl);
      }
      const BoardMask changedEffects() const{
	BoardMask ret = changedEffects(BLACK);
	return ret |= changedEffects(WHITE);
      }
      const NumBitmapEffect changedPieces() const{
	return effects.changedPieces();
      }
      template <Ptype PTYPE> bool longEffectChanged() const 
      {
	return changedPieces().template hasLong<PTYPE>();
      }
      template <Ptype PTYPE> bool anyEffectChanged() const 
      {
	return changedPieces().template hasAny<PTYPE>();
      }
      PieceMask pin(Player king) const
      {
	return pin_or_open[king]&getOnBoardMask(king);
      }
      PieceMask pinOrOpen(Player king) const
      {
	return pin_or_open[king];
      }
      const checkmate::King8Info king8Info(Player king) const;
      const PieceMask promotedPieces() const { return promoted; }
    private:
      template<Direction DIR>
      void makePinOpenDir(Position target,
			  PieceMask& pins, PieceMask const& onBoard,Player defense)
      {
	const Offset offset = DirectionTraits<DIR>::blackOffset();
	Position pos=target-offset;
	int num;
	while(Piece::isEmptyNum(num=getPieceAt(pos).number()))
	  pos-=offset;
	king_mobility[defense][DIR]=static_cast<unsigned char>(pos.uintValue());
	if(Piece::isEdgeNum(num)) return;
	int num1=longEffectNumTable()[num][DIR];
	if(Piece::isPieceNum(num1) && onBoard.test(num1)){
	  pins.set(num);
	}
      }
      void recalcPinOpen(Position changed, Direction &lastDir, Player defense)
      {
	Position target=getKingPosition(defense);
	const Direction longD=Board_Table.getLongDirection<BLACK>(changed,target);
	if(!isLong(longD) || (lastDir!=UL && longD==lastDir)) return;
	lastDir=longD;
	Direction shortD=longToShort(longD);
	{
	  // reset old pins
	  Position oldPos=Position::makeDirect(king_mobility[defense][shortD]);
	  int oldNum=getPieceAt(oldPos).number();
	  if(Piece::isPieceNum(oldNum))
	    pin_or_open[defense].reset(oldNum);
	}
	const Offset offset = Board_Table.getOffsetForBlack(longD);
	Position pos=target-offset;
	int num;
	while(Piece::isEmptyNum(num=getPieceAt(pos).number()))
	  pos-=offset;
	king_mobility[defense][shortD]=static_cast<unsigned char>(pos.uintValue());
	if(Piece::isEdgeNum(num)) return;
	int num1=longEffectNumTable()[num][shortD];
	if(Piece::isPieceNum(num1) && getOnBoardMask(alt(defense)).test(num1)){
	  pin_or_open[defense].set(num);
	}
      }
      PieceMask makePinOpen(Position target,Player defense);
    public:
      void makePinOpen(Player defense);
    public:
      bool isConsistent(bool showError=true) const;
      /** ̹˴ؤƥ */
      bool isConsistent(const NumEffectState& prev, Move moved, bool show_error=true) const;
      // move methods
      void doSimpleMove(Position from, Position to, int promoteMask);
      void doDropMove(Position to,Ptype ptype);
      void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
      /**
       * ޥˤDirectionǤĹ뤫ɤ.
       * 64bitбѤ
       */
      template<Direction Dir,Player P>
      bool hasEffectDir(Position to) const {
	BOOST_STATIC_ASSERT( (DirectionTraits<Dir>::isLong) );
	const PieceMask& onBoardMask=getOnBoardMask(P);
	mask_t mask1=onBoardMask.getMask(1);
	mask1 &= ((PtypeDirectionTraits<LANCE,Dir>::canMove 
		   ? mask_t::makeDirect(PtypeFuns<LANCE>::indexMask) 
		   : mask_t::makeDirect(0)) 
		  | (PtypeDirectionTraits<BISHOP,Dir>::canMove 
		     ?  mask_t::makeDirect(PtypeFuns<BISHOP>::indexMask) 
		     :  mask_t::makeDirect(0))
		  | (PtypeDirectionTraits<ROOK,Dir>::canMove 
		     ? mask_t::makeDirect(PtypeFuns<ROOK>::indexMask)
		     : mask_t::makeDirect(0)));
	mask1 <<= 8;
	// ûӽ
	mask1&=getEffect(to).getMask(1)& NumBitmapEffect::longEffectMask();
	while (mask1.any())
	{
	  int num=mask1.takeOneBit()+NumBitmapEffect::longToNumOffset;
	  Position from = getPieceOf(num).position();
	  Direction dir=Board_Table.getLongDirection<BLACK>(Offset32(to,from));
	  if (dir==DirectionPlayerTraits<Dir,P>::directionByBlack)
	    return true;
	}
	return false;
      }
      /**
       * ޥPTYPEĹ뤫ɤ.
       */
      template <Ptype PTYPE>
      bool hasEffectLong(Player P, Position to) const {
	BOOST_STATIC_ASSERT( (PTYPE == LANCE || PTYPE == BISHOP || PTYPE == ROOK) );
	const PieceMask& onBoardMask=getOnBoardMask(P);
	mask_t mask1=onBoardMask.getMask(1);
	mask1 &= mask_t::makeDirect(PtypeFuns<PTYPE>::indexMask);
	mask1 <<= 8;
	// ûӽ
	mask1&=getEffect(to).getMask(1)& NumBitmapEffect::longEffectMask();
	return mask1.any();
      }
      /** 
       * оݤȤޥˤץ쥤䡼뤫ɤ.
       * @param player ¦
       * @param target оݤΥޥ
       */
      bool hasEffectBy(Player player,Position target) const {
	assert(target.isOnBoard());
	mask_t mask=getEffect(target).getMask(1);
	mask&=NumBitmapEffect::playerEffectMask(player);
	return !mask.none();
      }


      /** 
       * оݤȤޥˤץ쥤䡼(ʳ)뤫ɤ.
       * ꤷ뤳Ȥ.
       * @param player ¦
       * @param piece ¦ζ
       * @param target оݤΥޥ
       */
      bool hasEffectNotBy(Player player,Piece piece,Position target) const {
	assert(piece.owner()==player);
	PieceMask onBoardMask=getOnBoardMask(player);
	int num=piece.number();
	onBoardMask.reset(num);
	return (onBoardMask&getEffect(target)).any();
      }
      /** 
       * İʾζ𤫤.
       * ⤽ʤƤФ
       * @param player ¦
       * @param target оݤΥޥ
       */
      bool hasMultipleEffectBy(Player player,Position target) const 
      {
	mask_t mask=getEffect(target).getMask(1);
	mask&=NumBitmapEffect::playerEffectMask(player);
	return NumBitmapEffect::playerEffect(player).getMask(1) < mask;
      }

      /** 
       * attack  target Ĥ
       * @param target оݤΥޥ
       */
      bool hasEffectByPiece(Piece attack, Position target) const 
      {
	assert(target.isOnBoard());
	return getEffect(target).test(attack.number());
      }

      /**
       * target  ptype 뤫? ̤ʤ
       */
      template <Ptype PTYPE>
      bool hasEffectByPtype(Player attack, Position target) const
      {
	return effectBit<PTYPE>(attack, target).any();
      }
      /**
       * target  ptype 뤫? 
       */
      template <Ptype PTYPE>
      bool hasEffectByPtypeStrict(Player attack, Position target) const
      {
	mask_t mask=effectBit<PTYPE>(attack, target);
	if (isPromoted(PTYPE)) 
	  mask &= promoted.getMask<PTYPE>();
	else
	  mask &= ~(promoted.getMask<PTYPE>());
	return mask.any();
      }
      /** return a piece s.t. owner == attack, ptype == PTYPE, has effect on target.  return Piece::EMPTY() otherwise */
      template <Ptype PTYPE>
      const Piece effectPtype(Player attack, Position target) const
      {
	mask_t mask=effectBit<PTYPE>(attack, target);
	if (mask.none())
	  return Piece::EMPTY();
	return getPieceOf(mask.takeOneBit()+PtypeFuns<PTYPE>::indexNum*32);
      }

      /**
       * ο2ޤǿ
       * @return 0,1,2 (2ξ2ʾ)
       */
      int countEffect2(Player player,Position target) const 
      {
	assert(target.isOnBoard());
	return std::min(2,countEffect(player,target));
      }
      /**
       * ο. 
       * targetפϤ߽ФƤϤʤ
       */
      int countEffect(Player player,Position target) const 
      {
	assert(target.isOnBoard());
	return getEffect(target).countEffect(player);
      }
      /**
       * ο. 
       * targetפϤ߽ФƤϤʤ
       * @param pins ζϿʤ
       */
      int countEffect(Player player,Position target, PieceMask pins) const 
      {
	assert(target.isOnBoard());
	const NumBitmapEffect effect = getEffect(target);
	const int all = effect.countEffect(player);
	pins &= effect;
	return all - pins.countBit();
      }
    private:
      int countKingEffect(Player P, PieceMask pieces) const 
      {
	if (pieces.test(PtypeTable::getKingIndex(P)))
	  return pieces.countBit2();
	return pieces.none() ? 0 : 2;
      }
    public:
      /**
       * ̵̤ͭȤʳζ2ޤǿ
       * @return 0 ʤˤʤ, 1 ̤Τ, 2 ʳ1İʾ夢
       */
      int countKingEffect(Player P, Position target) const 
      {
	assert(target.isOnBoard());
	const PieceMask pieces
	  = getOnBoardMask(P) & getEffect(target);
	return countKingEffect(P, pieces);
      }
      int countKingEffect(Player P, Position target, PieceMask dont_use) const 
      {
	assert(target.isOnBoard());
	PieceMask pieces = getOnBoardMask(P);
	pieces ^= dont_use;
	return countKingEffect(P, pieces & getEffect(target));
      }
      /**
       * 1Ĥpiece֤
       * ʤPiece::EMPTY()֤
       * @param P(template) ¦
       * @param p åȤΰ
       */
      template<Player P> 
      Piece hasEffectPiece(Position p) const
      {
	assert(p.isOnBoard());
#if OSL_WORDSIZE == 64
	const PieceMask pieceMask=getOnBoardMask(P)&getEffect(p);
	if (pieceMask.none()) return Piece::EMPTY();
	return getPieceOf(pieceMask.getMask(0).bsf());
#elif OSL_WORDSIZE == 32
	const PieceMask& onBoardMask=getOnBoardMask(P);
	mask_t mask0=onBoardMask.getMask(0);
	NumBitmapEffect effect=getEffect(p);
	mask0&=effect.getMask(0);
	if (mask0.any())
	{
	  return getPieceOf(mask0.bsf());
	}
	else{
	  mask_t mask1=effect.getMask(1);
	  mask1|=(mask1>>8);
	  mask1&=onBoardMask.getMask(1);
	  if (mask1.any())
	  {
	    int num=mask1.bsf()+32;
	    assert(32<=num && num<=39);
	    return getPieceOf(num);
	  }
	  else{
	    return Piece::EMPTY();
	  }
	}
#endif
      }
      Piece hasEffectPiece(Player pl,Position pos) const
      {
	if (pl==BLACK) return hasEffectPiece<BLACK>(pos);
	else return hasEffectPiece<WHITE>(pos);
      }
      /**
       * @param P ¦
       * @param Type 
       */
      template<Player P,Ptype Type,class Action>
      void forEachEffectPtype(Position pos,Action & action) const {
	mask_t mask=effectBit<Type>(P, pos);
	while (mask.any())
	{
	  int num=mask.takeOneBit()+(PtypeFuns<Type>::indexNum<<5);
	  action.template doActionPtype<P,Type>(getPieceOf(num),pos);
	}
      }

      /** 
       * pos ˤ move  action  member ƤӽФ.
       * @param action Ȥ AlwaysMoveAction
       */
      template<Player P,class Action>
      void forEachEffect(Position pos,Action & action) const
      {
	const PieceMask pieceMask=getOnBoardMask(P)&getEffect(pos);
#if OSL_WORDSIZE == 64
	mask_t mask=pieceMask.getMask(0);
	while (mask.any())
	{
	  const int num=mask.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#elif OSL_WORDSIZE == 32
	mask_t mask0=pieceMask.getMask(0);
	while (mask0.any())
	{
	  const int num=mask0.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
	mask_t mask1=pieceMask.getMask(1);
	while (mask1.any())
	{
	  const int num=mask1.takeOneBit()+32;
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#endif
      }
      /** 
       * pos ˤ move  action  member ƤӽФ.
       * @param pin ̵뤹
       * @param action Ȥ AlwaysMoveAction
       */
      template<Player P,class Action>
      void forEachEffect(Position pos,Action & action,const PieceMask& pin) const
      {
	PieceMask pieceMask=getOnBoardMask(P)&getEffect(pos);
	pieceMask &= ~pin;
#if OSL_WORDSIZE == 64
	mask_t mask=pieceMask.getMask(0);
	while (mask.any())
	{
	  const int num=mask.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#elif OSL_WORDSIZE == 32
	mask_t mask0=pieceMask.getMask(0);
	while (mask0.any())
	{
	  const int num=mask0.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
	mask_t mask1=pieceMask.getMask(1);
	while (mask1.any())
	{
	  const int num=mask1.takeOneBit()+32;
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#endif
      }

      /** 
       * pos ˰ư move  action  member ƤӽФ
       * @param action Ȥ AlwayMoveAction
       * @param piece  ʳζȤ
       */
      template<Player P,class Action>
      void forEachEffectNotBy(Position pos,Piece piece,Action & action) const {
	PieceMask pieceMask=getOnBoardMask(P);
	pieceMask.reset(piece.number());
	PieceMask effectMask=(PieceMask)getEffect(pos);
	pieceMask=(pieceMask&effectMask);
#if OSL_WORDSIZE == 64
	mask_t mask=pieceMask.getMask(0);
	while (mask.any())
	{
	  int num=mask.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#elif OSL_WORDSIZE == 32
	mask_t mask0=pieceMask.getMask(0);
	while (mask0.any())
	{
	  int num=mask0.takeOneBit();
	  action.template doAction<P>(getPieceOf(num),pos);
	}
	mask_t mask1=pieceMask.getMask(1);
	while (mask1.any())
	{
	  int num=mask1.takeOneBit()+32;
	  action.template doAction<P>(getPieceOf(num),pos);
	}
#endif
      }    

      /** ®ٽŻǤϤʤֽŻ 
       * pos ˤ move  action  member ƤӽФ
       * @param action Ȥ AlwayMoveAction
       */
      template<Player P,class Action>
      void forSortedEffect(Position pos,Action & action) const ;
      /**
       * attackerPiecetargetPositionĤɤ.
       * targetPosition϶Ǥɤ
       * @param attackerPiece 
       * @param targetPosition
       */
      bool hasEffectTo(Piece attackerPiece,Position targetPosition) const
      {
	assert(attackerPiece.isPiece() && targetPosition.isOnBoard());
	return getEffect(targetPosition).test(attackerPiece.number());
      }

      /** 
       * оݤȤޥˤץ쥤䡼뤫ɤ.
       * @param player ¦
       * @param target оݤΥޥ
       */
      template<Player P>
      bool hasEffectBy(Position target) const {
	assert(target.isOnBoard());
	mask_t mask=getEffect(target).getMask(1);
	mask&=NumBitmapEffect::playerEffectMask<P>();
	return !mask.none();
      }
    
      bool hasEffectBy(Player P, Position target,Piece& attackerPiece) const 
      {
	if (P == BLACK)
	  return hasEffectBy<BLACK>(target, attackerPiece);
	else
	  return hasEffectBy<WHITE>(target, attackerPiece);
      }
      /**
       * @param P(template) - ĤƤ¦Υץ쥤
       * @param target - Ĥ줿
       * @param attackerPiece - multiple attackξPiece::EMPTY()
       *        ǤʤʤĤƤ֤
       */
      template<Player P>
      bool hasEffectBy(Position target,Piece& attackerPiece) const {
	attackerPiece=Piece::EMPTY();
	const PieceMask& pieceMask=getOnBoardMask(P)&getEffect(target);
#if OSL_WORDSIZE == 64
	mask_t mask=pieceMask.getMask(0);
	if (mask.none()) return false;
	/**
	 * mask|=0x8000000000000000ll;
	 * if((mask&(mask-1))!=0x8000000000000000ll) ʤ1ĤifʸǺѤ
	 */
	if (mask.hasMultipleBit())
	  return true;
	int num=mask.bsf();
	attackerPiece=getPieceOf(num);
	return true;
#elif OSL_WORDSIZE == 32
	mask_t mask0=pieceMask.getMask(0);
	mask_t mask1=pieceMask.getMask(1);
	if (mask0.any())
	{
	  if (mask1.any())
	    return true;
	  int num=mask0.bsf();
	  if (mask0 == PieceMask::numToMask(num))
	    attackerPiece=getPieceOf(num);
	  return true;
	}
	else if (mask1.any())
	{
	  int num=mask1.bsf();
	  if (mask1==PieceMask::numToMask(num))
	    attackerPiece=getPieceOf(num+32);
	  return true;
	}
	else 
	  return false;
#endif
      }

      /** Pζ̤ */
      bool inCheck(Player P) const 
      {
	return hasEffectBy(alt(P), getKingPosition(P));
      }
      /** ֤ζ̤ */
      bool inCheck() const { return inCheck(getTurn()); }
      /**
       * õ
       * @return ꤫ɤ
       * @param attack_piece
       * Ĥζˤ벦ξattck_pieceˤζ
       * ʣζˤ벦ξPiece::EMPTY()
       * @param P(template) 
       */
      template<Player P>
      bool findCheckPiece(Piece& attack_piece) const
      {
	return hasEffectBy<PlayerTraits<P>::opponent>(getKingPosition(P),attack_piece);
      }
      /**
       * ˡ꤫ɤñ˸롥̤˰¸åΤߡ
       * 롼ؤʤǤǽϡisValidMove Ѥ롥
       *
       * PtemplateˤǤȾ®ʤꤽ
       * ̤˰¸븡Ǥ⡤̤ȴ䲦ɤǤ뤫
       * ꡤͤɤϸʤ
       * @param showError(template) - false֤ˤϲɽ.
       * @param move - 
       */
      template <bool show_error>
      bool isAlmostValidMove(Move move) const;
      bool isAlmostValidMove(Move move,bool show_error=true) const;

    private:
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position, Action&, Int2Type<false>) const {}
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position piecePosition,Action & action,Int2Type<true>) const {
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset();
	action.template doAction<P>(this->getPieceAt(piecePosition),piecePosition+offset);
      }

      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position piecePosition,Action & action) const {
	forEachEffectOfPieceDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
      }
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position, Action&, Int2Type<false>) const {}
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position piecePosition,Action & action,Int2Type<true>) const {
	Piece piece=this->getPieceAt(piecePosition);
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset();
	assert(offset.intValue() != 35);
	Position pos=piecePosition+offset;
	for (;this->getPieceAt(pos).isEmpty();pos+=offset)
	  action.template doAction<P>(piece,pos);
	action.template doAction<P>(piece,pos);
      }

      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position piecePosition,Action & action) const {
	forEachEffectOfPieceLongDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
      }
    public:
      /**
       * piecePositionˤˤä뤹٤Ƥposition
       * (ޤ)ˤĤ
       * action¹Ԥ
       *
       * SimpleState ܿ: ImmediateAddSupportʤоä
       */
      template<Player P,Ptype Type,class Action>
      void forEachEffectOfPiece(Position piecePosition,Action & action) const;
      template<class Action>
      void forEachEffectOfPiece(Piece piece,Action & action) const;

      /**
       * attackerptypeoζ𤬤Ȳꤷtarget뤫ɤ
       *  stateupdateʤǳΤ.
       * targetPosition϶Ǥɤ
       * ׾ζưƤƤȤϤˡ
       * ʬȤαäʤȸʤ뤳Ȥ
       */
      bool hasEffectFromTo(PtypeO ptypeo,Position attacker,
			   Position target) const
      {
	Offset32 offset32=Offset32(target,attacker);
	EffectContent effect=Ptype_Table.getEffect(ptypeo,offset32);
	if (! effect.hasEffect()) 
	  return false;
	if (effect.hasUnblockableEffect())
	  return true;
	assert(Board_Table.getShortOffset(offset32) == effect.offset());
	return this->isEmptyBetween(attacker,target,effect.offset());
      }
      /**
       * @param P(template) - ¦
       * @param pos - оݤΥޥ
       * @param action - ¹Ԥ
       */
      template<Player P,class Action,bool checkDone,class PositionCheck>
      void forEachEffectWithCheck(Position pos,Action & action,PositionCheck const& positionCheck) const;
      /**
       * 
       */
      template<Player P>
      bool hasEffectByWithRemove(Position target,Position removed) const;
    
      bool hasEffectByWithRemove(Player player, Position target,Position removed) const{
	if (player==BLACK)
	  return hasEffectByWithRemove<BLACK>(target,removed);
	else
	  return hasEffectByWithRemove<WHITE>(target,removed);
      }

      void showEffect(std::ostream& os) const;

      /** pl (1İʾ) */
      const PieceMask effectedMask(Player pl) const
      {
	return effects.effected_mask[pl];
      }
      /** λؼeffectedMask(pl)Ѳ.
       * 줿ϸߤμǤϥꥹȤʤ褦.
       */
      const PieceMask effectedChanged(Player pl) const
      {
	return effects.effected_changed_mask[pl];
      }
      void makeMovePass()
      {
	changeTurn();
	effects.clearChangedEffects();
	effects.clearEffectedChanged();
      }
    };

    template<Player P,class Action>
    void NumEffectState::forSortedEffect(Position pos,Action & action) const {
      forEachEffectPtype<P,PAWN,Action>(pos,action);
      forEachEffectPtype<P,LANCE,Action>(pos,action);
      forEachEffectPtype<P,KNIGHT,Action>(pos,action);
      forEachEffectPtype<P,SILVER,Action>(pos,action);
      forEachEffectPtype<P,GOLD,Action>(pos,action);
      forEachEffectPtype<P,BISHOP,Action>(pos,action);
      forEachEffectPtype<P,ROOK,Action>(pos,action);
      forEachEffectPtype<P,KING,Action>(pos,action);
    }

    inline bool operator!=(const NumEffectState& s1,
			   const NumEffectState& s2)
    {
      return !(s1==s2);
    }  
  } // namespace effect
  using state::NumEffectState;

  namespace apply_move
  {
    template<Player P>
    struct ApplyDoUndoSimpleMove<P,NumEffectState>
    {
      typedef NumEffectState state_t;
      template <typename F>
      static void doUndoSimpleMove(state_t& s, 
				   Position from, Position to, int promoteMask,F& func);

      static 
      void prologue(state_t& s, Position from, Position to, int promoteMask,
		    Piece& oldPiece, int& num, 
		    PtypeO& oldPtypeO, PtypeO& new_ptypeo,
		    CArray<PieceMask,2>& pin_or_open_backup,
		    KingMobility& king_mobility_backup,
		    PieceMask& promoted_backup,
		    CArray<PieceMask,2>& effected_mask_backup,
		    CArray<PieceMask,2>& effected_changed_mask_backup,
		    MobilityTable &mobility_backup
		    );
      static 
      void epilogue(state_t& s, Position from, Position to, Piece oldPiece, 
		    int num, PtypeO oldPtypeO, PtypeO newPtypeO,
		    const CArray<PieceMask,2>& pin_or_open_backup,
		    const KingMobility& king_mobility_backup,
		    const PieceMask& promoted_backup,
		    const CArray<PieceMask,2>& effected_mask_backup,
		    const CArray<PieceMask,2>& effected_changed_mask_backup,
		    const MobilityTable &mobility_backup
);
    };
  
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoSimpleMove<P,NumEffectState>::
    doUndoSimpleMove(state_t& s, 
		     Position from, Position to, int promoteMask,F& func)
    {
      Piece oldPiece;
      int num;
      PtypeO oldPtypeO, newPtypeO;
      CArray<PieceMask,2> pin_or_open_backup;
      KingMobility king_mobility_backup;
      PieceMask promoted_backup;
      CArray<PieceMask,2> effected_mask_backup;
      CArray<PieceMask,2> effected_changed_mask_backup;
      MobilityTable mobility_backup;
      prologue(s, from, to, promoteMask, oldPiece, num, oldPtypeO, newPtypeO, 
	       pin_or_open_backup, 
	       king_mobility_backup,
	       promoted_backup,
		     effected_mask_backup, effected_changed_mask_backup,
	       mobility_backup);

      if (promoteMask!=0 && num < PtypeTraits<PAWN>::indexLimit)
	{
	  s.clearPawn(P,from);
	  s.changeTurn();
	  func(to);
	  s.changeTurn();
	  s.setPawn(P,from);
	}
      else
	{
	  s.changeTurn();
	  func(to);
	  s.changeTurn();
	}

      epilogue(s, from, to, oldPiece, num, oldPtypeO, newPtypeO, 
	       pin_or_open_backup, 
	       king_mobility_backup,
promoted_backup, effected_mask_backup, effected_changed_mask_backup,mobility_backup);
    }

    template<Player P>
    struct ApplyDoUndoDropMove<P,NumEffectState>
    {
      typedef NumEffectState state_t;
      template <typename F>
      static void doUndoDropMove(state_t& s, 
				 Position to, Ptype ptype,F& func);
      static 
      void prologue(state_t& s, Position to, Ptype ptype,
		    Piece& oldPiece, int& num, PtypeO& ptypeO, 
		    int& numIndex, mask_t& numMask,
		    CArray<PieceMask,2>& pin_or_open_backup,
		    KingMobility& king_mobility_backup,
		    CArray<PieceMask,2>& effected_mask_backup,
		    CArray<PieceMask,2>& effected_changed_mask_backup,
 		    MobilityTable &mobility_backup
		    );
      static 
      void epilogue(state_t& s, Position to, Ptype ptype, Piece oldPiece, 
		    int num, PtypeO ptypeO, int numIndex, mask_t numMask,
		    const CArray<PieceMask,2>& pin_or_open_backup,
		    const KingMobility& king_mobility_backup,
		    const CArray<PieceMask,2>& effected_mask_backup,
		    const CArray<PieceMask,2>& effected_changed_mask_backup,
		    const MobilityTable &mobility_backup
		    );
    };
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoDropMove<P,NumEffectState>::
    doUndoDropMove(state_t& s, 
		   Position to, Ptype ptype,F& func)
    {
      Piece oldPiece;
      PtypeO ptypeO;
      int num, numIndex;
      mask_t numMask;
      CArray<PieceMask,2> pin_or_open_backup;
      KingMobility king_mobility_backup;
      CArray<PieceMask,2> effected_mask_backup;
      CArray<PieceMask,2> effected_changed_mask_backup;
      MobilityTable mobility_backup;
      prologue(s, to, ptype, oldPiece, num, ptypeO, numIndex, numMask, 
	       pin_or_open_backup, king_mobility_backup,

effected_mask_backup,effected_changed_mask_backup,mobility_backup);

      if (ptype==PAWN)
	{
	  s.setPawn(P,to);
	  s.changeTurn();
	  func(to);
	  s.changeTurn();
	  s.clearPawn(P,to);
	}
      else
	{
	  s.changeTurn();
	  func(to);
	  s.changeTurn();
	}
      epilogue(s, to, ptype, oldPiece, num, ptypeO, numIndex, numMask, 
	       pin_or_open_backup, king_mobility_backup,
	       effected_mask_backup,effected_changed_mask_backup,mobility_backup);
    }

    template<Player P>
    struct ApplyDoUndoCaptureMove<P,NumEffectState>
    {
      typedef NumEffectState state_t;
      template <typename F>
      static void doUndoCaptureMove(state_t& s, Position from,Position to, 
				    Piece target, int promoteMask,F& func);
      static
      void prologue(state_t& s, Position from, Position to, Piece target, 
		    int promoteMask,
		    Piece& oldPiece, PtypeO& oldPtypeO, PtypeO& capturePtypeO, 
		    PtypeO& new_ptypeo, int& num0, int& num1, 
		    int& num1Index, mask_t& num1Mask,
		    CArray<PieceMask,2>& pin_or_open_backup,
		    KingMobility& king_mobility_backup,
		    PieceMask& promoted_backup,
		    CArray<PieceMask,2>& effected_mask_backup,
		    CArray<PieceMask,2>& effected_changed_mask_backup,
		    MobilityTable &mobility_backup
);
      static
      void epilogue(state_t& s, Position from, Position to, Piece target, 
		    Piece oldPiece, PtypeO oldPtypeO, PtypeO capturePtypeO, 
		    PtypeO newPtypeO, int num0, int num1, 
		    int num1Index, mask_t num1Mask,
		    const CArray<PieceMask,2>& pin_or_open_backup,
		    const KingMobility& king_mobility_backup,
		    const PieceMask& promoted_backup,
		    const CArray<PieceMask,2>& effected_mask_backup,
		    const CArray<PieceMask,2>& effected_changed_mask_backup,
		    const MobilityTable &mobility_backup
);
    };
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoCaptureMove<P,NumEffectState>::
    doUndoCaptureMove(state_t& s, Position from,Position to, Piece target, 
		      int promoteMask,F& func)
    {
      Piece oldPiece;
      PtypeO oldPtypeO, capturePtypeO, newPtypeO;
      int num0, num1, num1Index;
      mask_t num1Mask;
      CArray<PieceMask,2> pin_or_open_backup;
      KingMobility king_mobility_backup;
      PieceMask promoted_backup;
      CArray<PieceMask,2> effected_mask_backup;
      CArray<PieceMask,2> effected_changed_mask_backup;
      MobilityTable mobility_backup;
      prologue(s, from, to, target, promoteMask, oldPiece, oldPtypeO, 
	       capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask, 
	       pin_or_open_backup, king_mobility_backup,
	       promoted_backup,
	       effected_mask_backup, effected_changed_mask_backup,mobility_backup);

      s.changeTurn();
      const Ptype capturePtype=target.ptype();
      if (capturePtype==PAWN)
	{
	  s.clearPawn(PlayerTraits<P>::opponent,to);
	  if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
	    {
	      s.clearPawn(P,from);
	      func(to);
	      s.setPawn(P,from);
	    }
	  else
	    {
	      func(to);
	    }
	  s.setPawn(PlayerTraits<P>::opponent,to);
	}
      else if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
	{
	  s.clearPawn(P,from);
	  func(to);
	  s.setPawn(P,from);
	}
      else
	{
	  func(to);
	}
      s.changeTurn();

      epilogue(s, from, to, target, oldPiece, oldPtypeO, capturePtypeO, newPtypeO, 
	       num0, num1, num1Index,num1Mask, 
	       pin_or_open_backup, king_mobility_backup,
	       promoted_backup,effected_mask_backup, effected_changed_mask_backup,mobility_backup);
    }
} // namespace state
  using state::NumEffectState;
} // namespace osl

template <class Action>
void osl::NumEffectState::
forEachEffectOfPiece(Piece piece,Action & action) const 
{
  Position piecePosition = piece.position();
  switch (piece.ptypeO()) {
  case NEW_PTYPEO(WHITE,PAWN): forEachEffectOfPiece<WHITE,PAWN,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,LANCE): forEachEffectOfPiece<WHITE,LANCE,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,KNIGHT): forEachEffectOfPiece<WHITE,KNIGHT,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,SILVER): forEachEffectOfPiece<WHITE,SILVER,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PPAWN): forEachEffectOfPiece<WHITE,PPAWN,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PLANCE): forEachEffectOfPiece<WHITE,PLANCE,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PKNIGHT): forEachEffectOfPiece<WHITE,PKNIGHT,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PSILVER): forEachEffectOfPiece<WHITE,PSILVER,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,GOLD): forEachEffectOfPiece<WHITE,GOLD,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,BISHOP): forEachEffectOfPiece<WHITE,BISHOP,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PBISHOP): forEachEffectOfPiece<WHITE,PBISHOP,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,ROOK): forEachEffectOfPiece<WHITE,ROOK,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,PROOK): forEachEffectOfPiece<WHITE,PROOK,Action>(piecePosition,action); break;
  case NEW_PTYPEO(WHITE,KING): forEachEffectOfPiece<WHITE,KING,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PAWN): forEachEffectOfPiece<BLACK,PAWN,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,LANCE): forEachEffectOfPiece<BLACK,LANCE,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,KNIGHT): forEachEffectOfPiece<BLACK,KNIGHT,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,SILVER): forEachEffectOfPiece<BLACK,SILVER,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PPAWN): forEachEffectOfPiece<BLACK,PPAWN,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PLANCE): forEachEffectOfPiece<BLACK,PLANCE,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PKNIGHT): forEachEffectOfPiece<BLACK,PKNIGHT,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PSILVER): forEachEffectOfPiece<BLACK,PSILVER,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,GOLD): forEachEffectOfPiece<BLACK,GOLD,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,BISHOP): forEachEffectOfPiece<BLACK,BISHOP,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PBISHOP): forEachEffectOfPiece<BLACK,PBISHOP,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,ROOK): forEachEffectOfPiece<BLACK,ROOK,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,PROOK): forEachEffectOfPiece<BLACK,PROOK,Action>(piecePosition,action); break;
  case NEW_PTYPEO(BLACK,KING): forEachEffectOfPiece<BLACK,KING,Action>(piecePosition,action); break;
  default: assert(0);
  }
}

template <osl::Player P, osl::Ptype Type, class Action>
void osl::NumEffectState::
forEachEffectOfPiece(Position piecePosition,Action & action) const 
{
  forEachEffectOfPieceDir<P,Type,Action,UL>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,U>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,UR>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,L>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,R>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,DL>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,D>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,DR>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,UUL>(piecePosition,action);
  forEachEffectOfPieceDir<P,Type,Action,UUR>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_UL>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_U>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_UR>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_L>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_R>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_DL>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_D>(piecePosition,action);
  forEachEffectOfPieceLongDir<P,Type,Action,LONG_DR>(piecePosition,action);
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoSimpleMove<P,osl::state::NumEffectState>::
prologue(state_t& s, Position from, Position to, int promoteMask,
	      Piece& oldPiece, int& num, 
	      PtypeO& oldPtypeO, PtypeO& new_ptypeo,
	 CArray<PieceMask,2>& pin_or_open_backup,
	 KingMobility& king_mobility_backup,
	      PieceMask& promoted_backup,
	 CArray<PieceMask,2>& effected_mask_backup,
	 CArray<PieceMask,2>& effected_changed_mask_backup,
	 MobilityTable &mobility_backup)
{
  mobility_backup = s.effects.mobilityTable;
  pin_or_open_backup = s.pin_or_open;
  king_mobility_backup = s.king_mobility;
  effected_mask_backup = s.effects.effected_mask;
  effected_changed_mask_backup = s.effects.effected_changed_mask;

  oldPiece=s.getPieceAt(from);
  Piece newPiece=oldPiece.promoteWithMask(promoteMask);
  newPiece+=(to-from);
  num=oldPiece.number();

  oldPtypeO=oldPiece.ptypeO();
  new_ptypeo=newPiece.ptypeO();
  // ʬȤθ򳰤
  s.setPieceOf(num,newPiece);
  s.effects.clearChangedEffects();
  s.effects.clearEffectedChanged();
  s.effects.template doEffect<NumBitmapEffect::Sub,true>(s,oldPtypeO,from,num);
  // ʬȤ֥åƤpromote?αĹ
  // 뤤ϼʬȤΥ֥å
  s.effects.effectedNumTable[num].clear();
  s.setBoard(to,newPiece);
  s.effects.template doBlockAt<NumBitmapEffect::Sub,true>(s,to,num);
  s.setBoard(from,Piece::EMPTY());
  s.effects.template doBlockAt<NumBitmapEffect::Add,true>(s,from,num);
  s.effects.template doEffect<NumBitmapEffect::Add,true>(s,new_ptypeo,to,num);

  if (oldPtypeO == newPtypeO(P,KING))
    s.makePinOpen(P);
  else {
    Direction lastD=UL;
    s.pin_or_open[P].reset(num);
    s.recalcPinOpen(from,lastD,P);
    s.recalcPinOpen(to,lastD,P);
  }
  {
    Direction lastD=UL;
    s.pin_or_open[alt(P)].reset(num);
    s.recalcPinOpen(from,lastD,alt(P));
    s.recalcPinOpen(to,lastD,alt(P));
  }
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  promoted_backup = s.promoted;
  if (promoteMask)
    s.promoted.set(num);
  if(s.hasEffectBy(BLACK,to))
    s.effects.effected_mask[BLACK].set(num);
  else
    s.effects.effected_mask[BLACK].reset(num);
  if(s.hasEffectBy(WHITE,to))
    s.effects.effected_mask[WHITE].set(num);
  else
    s.effects.effected_mask[WHITE].reset(num);
  s.effects.effected_changed_mask[BLACK].set(num);
  s.effects.effected_changed_mask[WHITE].set(num);
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoSimpleMove<P,osl::state::NumEffectState>::
epilogue(state_t& s, Position from, Position to, Piece oldPiece, 
	      int num, PtypeO oldPtypeO, PtypeO newPtypeO,
	      const CArray<PieceMask,2>& pin_or_open_backup,
	 const KingMobility& king_mobility_backup,
	      const PieceMask& promoted_backup,
	 const CArray<PieceMask,2>& effected_mask_backup,
	 const CArray<PieceMask,2>& effected_changed_mask_backup,
	 const MobilityTable & mobility_backup
	 )
{
  s.setPieceOf(num,oldPiece);
  s.effects.template doEffect<NumBitmapEffect::Sub,false>(s,newPtypeO,to,num);
  s.setBoard(from,oldPiece);
  s.effects.effectedNumTable[num].clear();
  s.effects.template doBlockAt<NumBitmapEffect::Sub,false>(s,from,num);
  s.setBoard(to,Piece::EMPTY());
  s.effects.template doBlockAt<NumBitmapEffect::Add,false>(s,to,num);
  s.effects.template doEffect<NumBitmapEffect::Add,false>(s,oldPtypeO,from,num);
  s.effects.invalidateChangedEffects();
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  s.pin_or_open = pin_or_open_backup;
  s.king_mobility = king_mobility_backup;
  s.promoted = promoted_backup;
  s.effects.effected_mask = effected_mask_backup;
  s.effects.effected_changed_mask = effected_changed_mask_backup;
  s.effects.mobilityTable = mobility_backup;
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoDropMove<P,osl::state::NumEffectState>::
prologue(state_t& s, Position to, Ptype ptype,
	      Piece& oldPiece, int& num, PtypeO& ptypeO, 
	      int& numIndex, mask_t& numMask,
	      CArray<PieceMask,2>& pin_or_open_backup,
	 KingMobility& king_mobility_backup,
	 CArray<PieceMask,2>& effected_mask_backup,
	 CArray<PieceMask,2>& effected_changed_mask_backup,
	 MobilityTable &mobility_backup)
{
  mobility_backup = s.effects.mobilityTable;
  pin_or_open_backup = s.pin_or_open;
  king_mobility_backup = s.king_mobility;
  effected_mask_backup = s.effects.effected_mask;
  effected_changed_mask_backup = s.effects.effected_changed_mask;

#if OSL_WORDSIZE == 64
  numIndex=0;
#elif OSL_WORDSIZE == 32
  numIndex=Ptype_Table.getIndex(ptype);
#endif
  const mask_t ownMochigoma=
    s.standMask(P).getMask(numIndex) & Ptype_Table.getMaskLow(ptype);
  assert(ownMochigoma.any());
  numMask=ownMochigoma.lowestBit();
  int numLow = ownMochigoma.bsf();
  num = numLow|(numIndex<<5);
  oldPiece=s.getPieceOf(num);
  Piece newPiece=oldPiece;
  newPiece+=to-Position::STAND();
  ptypeO=newPiece.ptypeO();
  s.setPieceOf(num,newPiece);
  s.effects.clearChangedEffects();
  s.effects.clearEffectedChanged();
  s.effects.template doBlockAt<NumBitmapEffect::Sub,true>(s,to,num);
  s.effects.template doEffect<NumBitmapEffect::Add,true>(s,ptypeO,to,num);
  s.setBoard(to,newPiece);
  s.standMask(P).xorMask(numIndex,numMask);
  s.stand_count[P][ptype-PTYPE_BASIC_MIN]--;
  s.getOnBoardMask(P).xorMask(numIndex,numMask);
  {
    Direction lastD=UL;
    s.recalcPinOpen(to,lastD,P);
  }
  {
    Direction lastD=UL;
    s.recalcPinOpen(to,lastD,alt(P));
  }
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  if(s.hasEffectBy(BLACK,to))
    s.effects.effected_mask[BLACK].set(num);
  else
    s.effects.effected_mask[BLACK].reset(num);
  if(s.hasEffectBy(WHITE,to))
    s.effects.effected_mask[WHITE].set(num);
  else
    s.effects.effected_mask[WHITE].reset(num);
  s.effects.effected_changed_mask[BLACK].set(num);
  s.effects.effected_changed_mask[WHITE].set(num);
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoDropMove<P,osl::state::NumEffectState>::
epilogue(state_t& s, Position to, Ptype ptype, Piece oldPiece, 
	      int num, PtypeO ptypeO, int numIndex, mask_t numMask,
	      const CArray<PieceMask,2>& pin_or_open_backup,
	 const KingMobility& king_mobility_backup,
	 const CArray<PieceMask,2>& effected_mask_backup,
	 const CArray<PieceMask,2>& effected_changed_mask_backup,
	 const MobilityTable& mobility_backup
	 )
{
  s.standMask(P).xorMask(numIndex,numMask);
  s.stand_count[P][ptype-PTYPE_BASIC_MIN]++;
  s.getOnBoardMask(P).xorMask(numIndex,numMask);
  s.setBoard(to,Piece::EMPTY());
  s.effects.template doEffect<NumBitmapEffect::Sub,false>(s,ptypeO,to,num);
  s.effects.template doBlockAt<NumBitmapEffect::Add,false>(s,to,num);
  s.setPieceOf(num,oldPiece);
  s.effects.effectedNumTable[num].clear();
  s.effects.invalidateChangedEffects();
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  s.pin_or_open = pin_or_open_backup;
  s.king_mobility = king_mobility_backup;
  s.effects.effected_mask = effected_mask_backup;
  s.effects.effected_changed_mask = effected_changed_mask_backup;
  s.effects.mobilityTable = mobility_backup;
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoCaptureMove<P,osl::state::NumEffectState>::
prologue(state_t& s, Position from, Position to, Piece target, 
	 int promoteMask,
	 Piece& oldPiece, PtypeO& oldPtypeO, PtypeO& capturePtypeO, 
	 PtypeO& new_ptypeo, int& num0, int& num1, 
	 int& num1Index, mask_t& num1Mask,
	 CArray<PieceMask,2>& pin_or_open_backup,
	 KingMobility& king_mobility_backup,
	 PieceMask& promoted_backup,
	 CArray<PieceMask,2>& effected_mask_backup,
	 CArray<PieceMask,2>& effected_changed_mask_backup,
	 MobilityTable &mobility_backup)
{
  mobility_backup = s.effects.mobilityTable;
  pin_or_open_backup = s.pin_or_open;
  king_mobility_backup = s.king_mobility;
  effected_mask_backup = s.effects.effected_mask;
  effected_changed_mask_backup = s.effects.effected_changed_mask;

  num1=target.number();
  num1Index=PieceMask::numToIndex(num1);
  num1Mask=PieceMask::numToMask(num1);
  s.getOnBoardMask(PlayerTraits<P>::opponent).xorMask(num1Index,num1Mask);
  s.standMask(P).xorMask(num1Index,num1Mask);
  oldPiece=s.getPieceAt(from);
  Piece newPiece=oldPiece.promoteWithMask(promoteMask);
  newPiece+=(to-from);
  num0=oldPiece.number();
  s.setPieceOf(num0,newPiece);
  s.setPieceOf(num1,target.captured());
      
  oldPtypeO=oldPiece.ptypeO();
  new_ptypeo=newPiece.ptypeO();
  capturePtypeO=target.ptypeO();
  s.stand_count[P][unpromote(getPtype(capturePtypeO))-PTYPE_BASIC_MIN]++;
  s.effects.clearChangedEffects();
  s.effects.clearEffectedChanged();
  s.effects.setChangedPieces(s.getEffect(to));
  s.effects.template doEffect<NumBitmapEffect::Sub,true>(s,capturePtypeO,to,num1);
  s.effects.template doEffect<NumBitmapEffect::Sub,true>(s,oldPtypeO,from,num0);
  s.setBoard(from,Piece::EMPTY());
  s.effects.template doBlockAt<NumBitmapEffect::Add,true>(s,from,num0);
  s.effects.effectedNumTable[num0]=s.effects.effectedNumTable[num1];
  s.effects.effectedNumTable[num1].clear();
  s.setBoard(to,newPiece);
  s.effects.template doEffect<NumBitmapEffect::Add,true>(s,new_ptypeo,to,num0);

  if (oldPtypeO == newPtypeO(P,KING))
    s.makePinOpen(P);
  else {
    Direction lastD=UL;
    s.pin_or_open[P].reset(num0);
    s.pin_or_open[P].reset(num1); // captured is not pin
    s.recalcPinOpen(from,lastD,P);
    s.recalcPinOpen(to,lastD,P);
  }
  {
    Direction lastD=UL;
    s.pin_or_open[alt(P)].reset(num0);
    s.pin_or_open[alt(P)].reset(num1); // captured is not pin
    s.recalcPinOpen(from,lastD,alt(P));
    s.recalcPinOpen(to,lastD,alt(P));
  }
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  promoted_backup = s.promoted;
  s.promoted.reset(num1);
  s.effects.effected_mask[BLACK].reset(num1);
  s.effects.effected_mask[WHITE].reset(num1);
  if (promoteMask)
    s.promoted.set(num0);
  if(s.hasEffectBy(BLACK,to))
    s.effects.effected_mask[BLACK].set(num0);
  else
    s.effects.effected_mask[BLACK].reset(num0);
  if(s.hasEffectBy(WHITE,to))
    s.effects.effected_mask[WHITE].set(num0);
  else
    s.effects.effected_mask[WHITE].reset(num0);
  s.effects.effected_changed_mask[BLACK].set(num0);
  s.effects.effected_changed_mask[WHITE].set(num0);
}

template<osl::Player P>
void osl::apply_move::ApplyDoUndoCaptureMove<P,osl::state::NumEffectState>::
epilogue(state_t& s, Position from, Position to, Piece target, 
	      Piece oldPiece, PtypeO oldPtypeO, PtypeO capturePtypeO, 
	      PtypeO newPtypeO, int num0, int num1, 
	      int num1Index, mask_t num1Mask,
	      const CArray<PieceMask,2>& pin_or_open_backup,
	 const KingMobility& king_mobility_backup,
	      const PieceMask& promoted_backup,
	 const CArray<PieceMask,2>& effected_mask_backup,
	 const CArray<PieceMask,2>& effected_changed_mask_backup,
	 const MobilityTable &mobility_backup
)
{
  s.standMask(P).xorMask(num1Index,num1Mask);
  s.stand_count[P][unpromote(getPtype(capturePtypeO))-PTYPE_BASIC_MIN]--;
  s.getOnBoardMask(PlayerTraits<P>::opponent).xorMask(num1Index,num1Mask);
  s.effects.effectedNumTable[num1]=s.effects.effectedNumTable[num0];
  s.effects.effectedNumTable[num0].clear();
  s.setPieceOf(num0,oldPiece);
  s.setPieceOf(num1,target);
  s.effects.template doEffect<NumBitmapEffect::Sub,false>(s,newPtypeO,to,num0);
  s.setBoard(from,oldPiece);
  s.setBoard(to,target);
  s.effects.template doBlockAt<NumBitmapEffect::Sub,false>(s,from,num0);
  s.effects.template doEffect<NumBitmapEffect::Add,false>(s,capturePtypeO,to,num1);
  s.effects.template doEffect<NumBitmapEffect::Add,false>(s,oldPtypeO,from,num0);
  s.effects.invalidateChangedEffects();
  s.valid_king8info[P] = s.valid_king8info[alt(P)] = false;
  s.pin_or_open = pin_or_open_backup;
  s.king_mobility = king_mobility_backup;
  s.promoted = promoted_backup;
  s.effects.effected_mask = effected_mask_backup;
  s.effects.effected_changed_mask = effected_changed_mask_backup;
  s.effects.mobilityTable = mobility_backup;
}

#endif /* _NUM_EFFECT_STATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
