#include "osl/eval/experimental/majorPiece.h"
using osl::MultiInt;

template <bool Opening, osl::Ptype MajorBasic>
osl::misc::CArray<int, 18>
osl::eval::experimental::MajorY<Opening, MajorBasic>::table;

template <bool Opening, osl::Ptype MajorBasic>
void osl::eval::experimental::
MajorY<Opening, MajorBasic>::setUp(const Weights &weights)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i] = weights.value(i);
  }
}


template <bool Opening>
int osl::eval::experimental::RookPawn<Opening>::weight;

template <bool Opening>
void osl::eval::experimental::
RookPawn<Opening>::setUp(const Weights &weights)
{
  weight = weights.value(0);
}

template <bool Opening>
int osl::eval::experimental::RookPawn<Opening>::eval(const NumEffectState &state)
{
  int result = 0;
  for (int i = PtypeTraits<ROOK>::indexMin;
       i < PtypeTraits<ROOK>::indexLimit;
       ++i)
  {
    const Piece piece = state.getPieceOf(i);
    if (piece.isOnBoard() && !piece.position().canPromote(piece.owner()) &&
	!state.isPawnMaskSet(piece.owner(), piece.position().x()))
    {
      if (piece.owner() == BLACK)
	result += weight;
      else
	result -= weight;
    }
  }
  return result;
}


osl::misc::CArray<MultiInt, 180> osl::eval::experimental::RookPawnY::table;
osl::misc::CArray<MultiInt, 1620> osl::eval::experimental::
RookPawnY::y_attack_table;
osl::misc::CArray<MultiInt, 1620> osl::eval::experimental::
RookPawnY::y_defense_table;

void osl::eval::experimental::
RookPawnYX::setUp(const Weights &weights)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s) 
    {  
      RookPawnY::y_attack_table[i][s] = weights.value(i + ONE_DIM * 2 * s);
      RookPawnY::y_defense_table[i][s] = weights.value(i + ONE_DIM * 2 * s + ONE_DIM);
    }
  }
}


void osl::eval::experimental::
RookPawnY::setUp(const Weights &weights)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

MultiInt osl::eval::experimental::
RookPawnY::eval(const NumEffectState &state,
		const CArray2d<int, 2, 9> &pawns)
{
  MultiInt result;
  const CArray<Position,2> kings = {{
    state.getKingPosition<BLACK>(),
    state.getKingPosition<WHITE>(),
  }};
  for (int i = PtypeTraits<ROOK>::indexMin;
       i < PtypeTraits<ROOK>::indexLimit;
       ++i)
  {
    const Piece piece = state.getPieceOf(i);
    if (piece.isOnBoard())
    {
      if (piece.owner() == BLACK)
      {
	const int pawn_y =
	  pawns[PlayerTraits<BLACK>::index][piece.position().x() - 1];
	result +=
	  table[index(piece, pawn_y)] +
	  y_attack_table[indexY(kings[WHITE], piece, pawn_y)] +
	  y_defense_table[indexY(kings[BLACK], piece, pawn_y)];
      }
      else
      {
	int y = pawns[PlayerTraits<WHITE>::index][piece.position().x() - 1];
	if (y != 0)
	  y = 10 - y;
	result -=
	  table[index(piece, y)] +
	  y_attack_table[indexY(kings[BLACK], piece, y)] +
	  y_defense_table[indexY(kings[WHITE], piece, y)];
      }
    }
  }
  return result;
}


MultiInt osl::eval::experimental::AllMajor::weight;

void osl::eval::experimental::
AllMajor::setUp(const Weights &weights,int stage)
{
  weight[stage] = weights.value(0);
}


template <bool Opening>
osl::misc::CArray<int, 32> osl::eval::experimental::MajorGoldSilverAttacked<Opening>::table;

template <bool Opening>
void osl::eval::experimental::
MajorGoldSilverAttacked<Opening>::setUp(const Weights &weights)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i] = weights.value(i);
  }
}

template <bool Opening>
int osl::eval::experimental::
MajorGoldSilverAttacked<Opening>::index(
  const NumEffectState &state, Piece piece)
{
  return piece.ptype() + (state.getTurn() == piece.owner() ? 0 : PTYPE_SIZE);
}

template <bool Opening>
template <osl::Ptype PTYPE>
int osl::eval::experimental::
MajorGoldSilverAttacked<Opening>::evalOne(const NumEffectState &state)
{
  int result = 0;
  for (int i = PtypeTraits<PTYPE>::indexMin;
       i < PtypeTraits<PTYPE>::indexLimit;
       ++i)
  {
    const Piece piece = state.getPieceOf(i);
    if (piece.isOnBoard() &&
	state.hasEffectBy(alt(piece.owner()), piece.position()))
    {
      const int weight = table[index(state, piece)];
      if (piece.owner() == BLACK)
	result += weight;
      else
	result -= weight;
    }
  }
  return result;
}

template <bool Opening>
int osl::eval::experimental::
MajorGoldSilverAttacked<Opening>::eval(const NumEffectState &state)
{
  int result = 0;
  result += evalOne<ROOK>(state);
  result += evalOne<BISHOP>(state);
  result += evalOne<GOLD>(state);
  result += evalOne<SILVER>(state);

  return result;
}

osl::misc::CArray<MultiInt, 612> osl::eval::experimental::RookEffectBase::attack_table;
osl::misc::CArray<MultiInt, 612> osl::eval::experimental::RookEffectBase::defense_table;
osl::misc::CArray<MultiInt, 32> osl::eval::experimental::RookEffectBase::piece_table;

osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::attack_u;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::attack_d;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::attack_r;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::attack_l;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::defense_u;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::defense_d;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::defense_r;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::RookEffectBase::defense_l;
osl::misc::CArray<MultiInt, 722> osl::eval::experimental::RookEffectBase::attack_nospace;
osl::misc::CArray<MultiInt, 722> osl::eval::experimental::RookEffectBase::defense_nospace;


template<osl::Player P>
inline
MultiInt osl::eval::experimental::RookEffectBase::evalOne(
 const NumEffectState& state,
 Position rook,
 Position myKing,
 Position opKing,
 Position up,
 Position dp,
 Position rp,
 Position lp,
 bool isP)
{
  MultiInt result;
  PtypeO uPtypeO=state.getPieceAt(up).ptypeO();
  PtypeO dPtypeO=state.getPieceAt(dp).ptypeO();
  PtypeO rPtypeO=state.getPieceAt(rp).ptypeO();
  PtypeO lPtypeO=state.getPieceAt(lp).ptypeO();
  if(P==WHITE){
    uPtypeO=(PtypeO)(static_cast<int>(uPtypeO)^(~15));
    dPtypeO=(PtypeO)(static_cast<int>(dPtypeO)^(~15));
    rPtypeO=(PtypeO)(static_cast<int>(rPtypeO)^(~15));
    lPtypeO=(PtypeO)(static_cast<int>(lPtypeO)^(~15));
    up=up.rotate180EdgeOK();
    dp=dp.rotate180EdgeOK();
    rp=rp.rotate180EdgeOK();
    lp=lp.rotate180EdgeOK();
    rook=rook.rotate180();
    myKing=myKing.rotate180();
    opKing=opKing.rotate180();
  }
  assert((myKing.y()-dp.y())<(myKing.y()-up.y()));
  assert((myKing.x()-lp.x())<(myKing.x()-rp.x()));
  result+=attack_u[index1(opKing,up,uPtypeO,isP)]+
    attack_d[index1(opKing,dp,dPtypeO,isP)]+
    attack_l[index1(opKing,lp,lPtypeO,isP)]+
    attack_r[index1(opKing,rp,rPtypeO,isP)]+
    defense_u[index1(myKing,up,uPtypeO,isP)]+
    defense_d[index1(myKing,dp,dPtypeO,isP)]+
    defense_l[index1(myKing,lp,lPtypeO,isP)]+
    defense_r[index1(myKing,rp,rPtypeO,isP)]+
    attack_nospace[index2(opKing,rook,isP)]+
    defense_nospace[index2(myKing,rook,isP)];
  return result;
}

MultiInt osl::eval::experimental::
RookEffectBase::eval(const NumEffectState &state)
{
  const CArray<Position,2> kings = {{ 
    state.getKingPosition(BLACK), 
    state.getKingPosition(WHITE), 
  }};
  MultiInt result;
  for (int i = PtypeTraits<ROOK>::indexMin; i < PtypeTraits<ROOK>::indexLimit;
       ++i)
  {
    const Piece p = state.getPieceOf(i);
    if (! p.isOnBoard()) continue;
    const Position pos=p.position();
    Position up=state.getMobility(U,i);
    Position dp=state.getMobility(D,i);
    Position lp=state.getMobility(L,i);
    Position rp=state.getMobility(R,i);
    const bool isP=p.isPromoted();
    if(p.owner()==BLACK)
      result+=evalOne<BLACK>(state,pos,kings[0],kings[1],up,dp,rp,lp,isP);
    else
      result-=evalOne<WHITE>(state,pos,kings[1],kings[0],dp,up,lp,rp,isP);
  }
  return result;
}

void osl::eval::experimental::RookEffect::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    attack_table[i][stage] = weights.value(i);
    defense_table[i][stage] = weights.value(i + ONE_DIM);
  }
}

void osl::eval::experimental::RookEffectPiece::setUp(const Weights &weights)
{
  for (size_t i = 0; i < 32; ++i)
  {
    for (int s=0; s<NStages; ++s)
      RookEffectBase::piece_table[i][s] = weights.value(i + 32*s);
  }
}

void osl::eval::experimental::
RookEffectPieceKingRelative::setUp(const Weights &weights)
{
  CArray<MultiInt, 19584> piece_attack_table;
  CArray<MultiInt, 19584> piece_defense_table;
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
    {
      piece_attack_table[i][s] = weights.value(i + ONE_DIM*2*s);
      piece_defense_table[i][s] = weights.value(i + ONE_DIM*2*s + ONE_DIM);
    }
  }
  for(int isP=0;isP<2;isP++)
    for(int y_diff=-9;y_diff<=9;y_diff++)
      for(int x_diff= -9;x_diff<=9;x_diff++){
	int i2=index2(x_diff,y_diff,isP);
	if(abs(x_diff)<9 && abs(y_diff)<9){
	  attack_nospace[i2]= 
	    -(attack_table[index(abs(x_diff),y_diff,true,isP)]+
	      attack_table[index(abs(x_diff),y_diff,false,isP)]);
	  defense_nospace[i2]= 
	    -(defense_table[index(abs(x_diff),y_diff,true,isP)]+
	      defense_table[index(abs(x_diff),y_diff,false,isP)]);
	}
	for(int ptypeo= PTYPEO_MIN;ptypeo<=PTYPEO_MAX;ptypeo++){
	  if(getPtype((PtypeO)ptypeo)==(int)PTYPE_EMPTY) continue;
	  int i1=index1(x_diff,y_diff,(PtypeO)ptypeo,isP);
	  int indexPieceH,indexPieceV;
	  int table_ptypeo=ptypeo;
	  if(getPtype((PtypeO)ptypeo)==PTYPE_EDGE) table_ptypeo=PTYPEO_EDGE;
	  if(getPtype((PtypeO)ptypeo)==PTYPE_EDGE || abs(x_diff)==9 || abs(y_diff)==9){
	    indexPieceH= 0+0+(PTYPEO_EDGE-PTYPEO_MIN)*17*9+4896+isP*9792;
	    indexPieceV= 0+0+(PTYPEO_EDGE-PTYPEO_MIN)*17*9+ isP*9792;
	  }
	  else{
	    indexPieceH= index0(abs(x_diff),-y_diff,(PtypeO)ptypeo,true,isP);
	    indexPieceV= index0(abs(x_diff),-y_diff,(PtypeO)ptypeo,false,isP);
	  }
	  attack_u[i1]=piece_attack_table[indexPieceV]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_u[i1]=piece_defense_table[indexPieceV];
	  attack_d[i1]=piece_attack_table[indexPieceV]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_d[i1]=piece_defense_table[indexPieceV];
	  if(abs(x_diff)<=8){
	    for(int y_diff_1=y_diff+1;y_diff_1<=8;y_diff_1++){
	      int i=index(abs(x_diff),y_diff_1,false,isP);
	      attack_u[i1]+=attack_table[i];
	      defense_u[i1]+=defense_table[i];
	    }
	    for(int y_diff_1=std::max(-8,y_diff);y_diff_1<=8;y_diff_1++){
	      int i=index(abs(x_diff),y_diff_1,false,isP);
	      attack_d[i1]-=attack_table[i];
	      defense_d[i1]-=defense_table[i];
	    }
	  }
	  attack_l[i1]=piece_attack_table[indexPieceH]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_l[i1]=piece_defense_table[indexPieceH];
	  attack_r[i1]=piece_attack_table[indexPieceH]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_r[i1]=piece_defense_table[indexPieceH];
	  if(abs(y_diff)<=8){
	    for(int x_diff_1=x_diff+1;x_diff_1<=8;x_diff_1++){
	      int i=index(abs(x_diff_1),y_diff,true,isP);
	      attack_r[i1]+=attack_table[i];
	      defense_r[i1]+=defense_table[i];
	    }
	    for(int x_diff_1=std::max(-8,x_diff);x_diff_1<=8;x_diff_1++){
	      int i=index(abs(x_diff_1),y_diff,true,isP);
	      attack_l[i1]-=attack_table[i];
	      defense_l[i1]-=defense_table[i];
	    }
	  }
	}
      }
}



osl::misc::CArray<MultiInt, 256> osl::eval::experimental::RookPromoteDefense::promote_defense_table;

void osl::eval::experimental::RookPromoteDefense::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      promote_defense_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

MultiInt osl::eval::experimental::
RookPromoteDefense::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<ROOK>::indexMin;
       i < PtypeTraits<ROOK>::indexLimit;
       ++i){
    const Piece rook = state.getPieceOf(i);
    if(rook.isOnBoardNotPromoted()){
      if(rook.owner()==BLACK){
	Position rookPos=rook.position();
	if(rookPos.y()>=4){
	  Position pos=state.getMobility(U,i);
	  const Piece attacked = state.getPieceAt(pos);
	  if (attacked.canMoveOn<BLACK>()){
	    const NumBitmapEffect effect = state.getEffect(pos);
	    if (effect.countEffect(WHITE) == 1){
	      PieceMask mask = effect & state.getOnBoardMask(WHITE);
	      const Piece effect_piece = state.getPieceOf(mask.takeOneBit());
	      const int index = attacked.ptype() * 16 + effect_piece.ptype();
	      result += promote_defense_table[index];
	    }
	  }
	}
      }
      else{
	Position rookPos=rook.position();
	if(rookPos.y()<=6){
	  Position pos=state.getMobility(D,i);
	  const Piece attacked = state.getPieceAt(pos);
	  if (attacked.canMoveOn<WHITE>()){
	    const NumBitmapEffect effect = state.getEffect(pos);
	    if (effect.countEffect(BLACK) == 1){
	      PieceMask mask = effect & state.getOnBoardMask(BLACK);
	      const Piece effect_piece = state.getPieceOf(mask.takeOneBit());
	      const int index = attacked.ptype() * 16 + effect_piece.ptype();
	      result -= promote_defense_table[index];
	    }
	  }
	}
      }
    }
  }
  return result;
}



osl::misc::CArray<MultiInt, 612> osl::eval::experimental::BishopEffectBase::attack_table;
osl::misc::CArray<MultiInt, 612> osl::eval::experimental::BishopEffectBase::defense_table;
osl::misc::CArray<MultiInt, 32> osl::eval::experimental::BishopEffectBase::piece_table;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::attack_ul;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::attack_ur;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::attack_dl;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::attack_dr;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::defense_ul;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::defense_ur;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::defense_dl;
osl::misc::CArray<MultiInt, 23104> osl::eval::experimental::BishopEffectBase::defense_dr;
osl::misc::CArray<MultiInt, 722> osl::eval::experimental::BishopEffectBase::attack_nospace;
osl::misc::CArray<MultiInt, 722> osl::eval::experimental::BishopEffectBase::defense_nospace;


template<osl::Player P>
inline
MultiInt osl::eval::experimental::BishopEffectBase::evalOne(
 const NumEffectState& state,
 Position bishop,
 Position myKing,
 Position opKing,
 Position ulp,
 Position urp,
 Position dlp,
 Position drp,
 bool isP)
{
  MultiInt result;
  PtypeO ulPtypeO=state.getPieceAt(ulp).ptypeO();
  PtypeO urPtypeO=state.getPieceAt(urp).ptypeO();
  PtypeO dlPtypeO=state.getPieceAt(dlp).ptypeO();
  PtypeO drPtypeO=state.getPieceAt(drp).ptypeO();
  if(P==WHITE){
    ulPtypeO=(PtypeO)(static_cast<int>(ulPtypeO)^(~15));
    urPtypeO=(PtypeO)(static_cast<int>(urPtypeO)^(~15));
    dlPtypeO=(PtypeO)(static_cast<int>(dlPtypeO)^(~15));
    drPtypeO=(PtypeO)(static_cast<int>(drPtypeO)^(~15));
    ulp=ulp.rotate180EdgeOK();
    urp=urp.rotate180EdgeOK();
    dlp=dlp.rotate180EdgeOK();
    drp=drp.rotate180EdgeOK();
    bishop=bishop.rotate180();
    myKing=myKing.rotate180();
    opKing=opKing.rotate180();
  }
  result+=attack_ul[index1(opKing,ulp,ulPtypeO,isP)]+
    attack_ur[index1(opKing,urp,urPtypeO,isP)]+
    attack_dl[index1(opKing,dlp,dlPtypeO,isP)]+
    attack_dr[index1(opKing,drp,drPtypeO,isP)]+
    defense_ul[index1(myKing,ulp,ulPtypeO,isP)]+
    defense_ur[index1(myKing,urp,urPtypeO,isP)]+
    defense_dl[index1(myKing,dlp,dlPtypeO,isP)]+
    defense_dr[index1(myKing,drp,drPtypeO,isP)]+
    attack_nospace[index2(opKing,bishop,isP)]+
    defense_nospace[index2(myKing,bishop,isP)];
  return result;
}

MultiInt osl::eval::experimental::
BishopEffectBase::eval(const NumEffectState &state)
{
  const CArray<Position,2> kings = {{ 
    state.getKingPosition(BLACK), 
    state.getKingPosition(WHITE), 
  }};

  MultiInt result;
  for (int i = PtypeTraits<BISHOP>::indexMin; i < PtypeTraits<BISHOP>::indexLimit;
       ++i)
    {
      const Piece p = state.getPieceOf(i);
      if (! p.isOnBoard()) continue;
      const Position pos=p.position();
      Position ulp=state.getMobility(UL,i);
      Position urp=state.getMobility(UR,i);
      Position dlp=state.getMobility(DL,i);
      Position drp=state.getMobility(DR,i);
      const bool isP=p.isPromoted();
      if(p.owner()==BLACK)
	result+=evalOne<BLACK>(state,pos,kings[0],kings[1],ulp,urp,dlp,drp,isP);
      else
	result-=evalOne<WHITE>(state,pos,kings[1],kings[0],drp,dlp,urp,ulp,isP);
    }
  return result;
}

void osl::eval::experimental::BishopEffect::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    attack_table[i][stage] = weights.value(i);
    defense_table[i][stage] = weights.value(i + ONE_DIM);
  }
}

void osl::eval::experimental::BishopEffectPiece::setUp(const Weights &weights)
{
  for (size_t i = 0; i < 32; ++i)
  {
    for (int s=0; s<NStages; ++s)
      BishopEffectBase::piece_table[i][s] = weights.value(i + 32*s);
  }
}


void osl::eval::experimental::
BishopEffectPieceKingRelative::setUp(const Weights &weights)
{
  CArray<MultiInt, 19584> piece_attack_table;
  CArray<MultiInt, 19584> piece_defense_table;
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s) 
    {
      piece_attack_table[i][s] = weights.value(i + ONE_DIM * 2 * s);
      piece_defense_table[i][s] = weights.value(i + ONE_DIM * 2 * s + ONE_DIM);
    }
  }
  for(int isP=0;isP<2;isP++)
    for(int y_diff=-9;y_diff<=9;y_diff++)
      for(int x_diff= -9;x_diff<=9;x_diff++){
	int i2=index2(x_diff,y_diff,isP);
	if(abs(x_diff)<9 && abs(y_diff)<9){
	  attack_nospace[i2]= 
	    -(attack_table[index(x_diff,y_diff,true,isP)]+
	      attack_table[index(x_diff,y_diff,false,isP)]);
	  defense_nospace[i2]= 
	    -(defense_table[index(x_diff,y_diff,true,isP)]+
	      defense_table[index(x_diff,y_diff,false,isP)]);
	}
	for(int ptypeo= PTYPEO_MIN;ptypeo<=PTYPEO_MAX;ptypeo++){
	  if(getPtype((PtypeO)ptypeo)==(int)PTYPE_EMPTY) continue;
	  int i1=index1(x_diff,y_diff,(PtypeO)ptypeo,isP);
	  int indexPieceUR,indexPieceUL;
	  int table_ptypeo=ptypeo;
	  if(getPtype((PtypeO)ptypeo)==PTYPE_EDGE) table_ptypeo=PTYPEO_EDGE;
	  if(getPtype((PtypeO)ptypeo)==PTYPE_EDGE || abs(x_diff)==9 || abs(y_diff)==9){
	    indexPieceUR= 0+0+(PTYPEO_EDGE-PTYPEO_MIN)*17*9+4896+isP*9792;
	    indexPieceUL= 0+0+(PTYPEO_EDGE-PTYPEO_MIN)*17*9+ isP*9792;
	  }
	  else{
	    indexPieceUR= index0(x_diff,y_diff,(PtypeO)ptypeo,true,isP);
	    indexPieceUL= index0(x_diff,y_diff,(PtypeO)ptypeo,false,isP);
	  }
	  attack_ul[i1]=piece_attack_table[indexPieceUL]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_ul[i1]=piece_defense_table[indexPieceUL];
	  attack_dr[i1]=piece_attack_table[indexPieceUL]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_dr[i1]=piece_defense_table[indexPieceUL];
	  {
	    int y_diff_1=y_diff+1, x_diff_1=x_diff-1;
	    for(;y_diff_1<=8 && x_diff_1>=-8;y_diff_1++,x_diff_1--){
	      if(std::abs(x_diff_1)<=8 && std::abs(y_diff_1)<=8){
		int i=index(x_diff_1,y_diff_1,false,isP);
		attack_ul[i1]+=attack_table[i];
		defense_ul[i1]+=defense_table[i];
	      }
	    }
	  }
	  {
	    int y_diff_1=y_diff, x_diff_1=x_diff;
	    for(;y_diff_1<=8 && x_diff_1>=-8;y_diff_1++,x_diff_1--){
	      if(std::abs(x_diff_1)<=8 && std::abs(y_diff_1)<=8){
		int i=index(x_diff_1,y_diff_1,false,isP);
		attack_dr[i1]-=attack_table[i];
		defense_dr[i1]-=defense_table[i];
	      }
	    }
	  }
	  attack_ur[i1]=piece_attack_table[indexPieceUR]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_ur[i1]=piece_defense_table[indexPieceUR];
	  attack_dl[i1]=piece_attack_table[indexPieceUR]+piece_table[table_ptypeo-PTYPEO_MIN];
	  defense_dl[i1]=piece_defense_table[indexPieceUR];
	  {
	    int y_diff_1=y_diff+1, x_diff_1=x_diff+1;
	    for(;y_diff_1<=8 && x_diff_1<=8;y_diff_1++,x_diff_1++){
	      if(std::abs(x_diff_1)<=8 && std::abs(y_diff_1)<=8){
		int i=index(x_diff_1,y_diff_1,true,isP);
		attack_ur[i1]+=attack_table[i];
		defense_ur[i1]+=defense_table[i];
	      }
	    }
	  }
	  {
	    int y_diff_1=y_diff, x_diff_1=x_diff;
	    for(;y_diff_1<=8 && x_diff_1<=8;y_diff_1++,x_diff_1++){
	      if(std::abs(x_diff_1)<=8 && std::abs(y_diff_1)<=8){
		int i=index(x_diff_1,y_diff_1,true,isP);
		attack_dl[i1]-=attack_table[i];
		defense_dl[i1]-=defense_table[i];
	      }
	    }
	  }
	}
      }
}

osl::misc::CArray<MultiInt, 32> osl::eval::experimental::BishopHead::table;
osl::misc::CArray<MultiInt, 4896> osl::eval::experimental::BishopHead::king_table;


void osl::eval::experimental::BishopHead::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
BishopHeadKingRelative::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      BishopHead::king_table[i][s] = weights.value(i + ONE_DIM*s);
  }
  for(int x_diff=0;x_diff<=8;x_diff++)
    for(int y_diff=-8;y_diff<=8;y_diff++){
      for(int i=0;i<32;i++)
	BishopHead::king_table[(i*9+x_diff)*17+y_diff+8]+=BishopHead::table[i];
    }
  const PtypeO PTYPEO_EMPTY_R=newPtypeO(WHITE,PTYPE_EMPTY);
  const PtypeO PTYPEO_EDGE_R=newPtypeO(BLACK,PTYPE_EDGE);
  for(int x_diff=0;x_diff<=8;x_diff++)
    for(int y_diff=-8;y_diff<=8;y_diff++){
      BishopHead::king_table[(ptypeOIndex(PTYPEO_EMPTY_R)*9+x_diff)*17+y_diff+8]=
	BishopHead::king_table[(ptypeOIndex(PTYPEO_EMPTY)*9+x_diff)*17+y_diff+8];
      BishopHead::king_table[(ptypeOIndex(PTYPEO_EDGE_R)*9+x_diff)*17+y_diff+8]=
	BishopHead::king_table[(ptypeOIndex(PTYPEO_EDGE)*9+x_diff)*17+y_diff+8];
    }
}

MultiInt osl::eval::experimental::
BishopHead::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<BISHOP>::indexMin;
       i < PtypeTraits<BISHOP>::indexLimit;
       ++i){
    const Piece p = state.getPieceOf(i);
    if (p.isOnBoardNotPromoted()){
      const Position pos=p.position();
      if (p.owner() == BLACK){
	if(pos.y()>=2){
	  const Position up = pos+DirectionPlayerTraits<U,BLACK>::offset();
	  if (!state.hasEffectBy(BLACK, up)){
	    const Position king = state.getKingPosition(BLACK);
	    const PtypeO ptypeo = state.getPieceAt(up).ptypeO();
	    const int index_k = indexK(BLACK, ptypeo,
				       std::abs(pos.x() - king.x()),
				       pos.y() - king.y());
	    result += king_table[index_k];
	  }
	}
      }
      else if(pos.y()<=8) {
	const Position up = pos+DirectionPlayerTraits<U,WHITE>::offset();
	if (!state.hasEffectBy(WHITE, up)){
	  const Position king = state.getKingPosition(WHITE);
	  const PtypeO ptypeo = state.getPieceAt(up).ptypeO();
	  const int index_k = indexK(WHITE, ptypeo,
				     std::abs(pos.x() - king.x()),
				     pos.y() - king.y());
	  result -= king_table[index_k];
	}
      }
    }
  }
  return result;
}


osl::misc::CArray<MultiInt, 374544> osl::eval::experimental::KingRookBishop::table;

void osl::eval::experimental::KingRookBishop::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

template<osl::Player P>
MultiInt osl::eval::experimental::
KingRookBishop::evalOne(const NumEffectState &state)
{
  const Position king=state.getKingPosition(P);
  MultiInt result;
  for (int i = PtypeTraits<ROOK>::indexMin;
       i < PtypeTraits<ROOK>::indexLimit;
       ++i)
  {
    const Piece rook = state.getPieceOf(i);
    if (!rook.isOnBoard())
    {
      continue;
    }
    for (int j = PtypeTraits<BISHOP>::indexMin;
	 j < PtypeTraits<BISHOP>::indexLimit;
	 ++j)
    {
      const Piece bishop = state.getPieceOf(j);
      if (!bishop.isOnBoard())
      {
	continue;
      }
      result += table[index<P>(king, rook, bishop)];
    }
  }
  return result;
}

MultiInt osl::eval::experimental::
KingRookBishop::eval(const NumEffectState &state)
{
  return evalOne<BLACK>(state)-evalOne<WHITE>(state);
}


osl::misc::CArray<MultiInt, 9> osl::eval::experimental::NumPiecesBetweenBishopAndKing::self_table;
osl::misc::CArray<MultiInt, 9> osl::eval::experimental::NumPiecesBetweenBishopAndKing::opp_table;
osl::misc::CArray<MultiInt, 9> osl::eval::experimental::NumPiecesBetweenBishopAndKing::all_table;

void osl::eval::experimental::
NumPiecesBetweenBishopAndKingSelf::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      NumPiecesBetweenBishopAndKing::self_table[i][s] =
	weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
NumPiecesBetweenBishopAndKingOpp::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      NumPiecesBetweenBishopAndKing::opp_table[i][s] =
	weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
NumPiecesBetweenBishopAndKingAll::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      NumPiecesBetweenBishopAndKing::all_table[i][s] =
	weights.value(i + ONE_DIM*s);
  }
}

osl::MultiInt osl::eval::experimental::
NumPiecesBetweenBishopAndKing::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<BISHOP>::indexMin;
       i < PtypeTraits<BISHOP>::indexLimit;
       ++i)
  {
    const Piece bishop = state.getPieceOf(i);
    if (!bishop.isOnBoard())
    {
      continue;
    }
    int self, opp, all;
    countBetween(state,
		 state.getKingPosition(alt(bishop.owner())),
		 bishop, self, opp, all);
    if (bishop.owner() == BLACK)
    {
      result += (self_table[self] + opp_table[opp] + all_table[all]);
    }
    else
    {
      result -= (self_table[self] + opp_table[opp] + all_table[all]);
    }
  }
  return result;
}

void osl::eval::experimental::
NumPiecesBetweenBishopAndKing::countBetween(
  const NumEffectState &state, Position king, Piece bishop,
  int &self_count, int &opp_count, int &total_count)
{
  assert(bishop.isOnBoard());
  if ((king.x() + king.y() != bishop.position().x() + bishop.position().y()) &&
	(king.x() - king.y() != bishop.position().x() - bishop.position().y()))
  {
    self_count = opp_count = total_count = 8;
    return;
  }
  Direction dir;
  assert(king.x() != bishop.position().x());
  assert(king.y() != bishop.position().y());
  if (king.x() < bishop.position().x())
  {
    if (king.y() < bishop.position().y())
    {
	dir = UR;
    }
    else
    {
	dir = DR;
    }
  }
  else
  {
    if (king.y() < bishop.position().y())
    {
	dir = UL;
    }
    else
    {
	dir = DL;
    }
  }
  const Player player = bishop.owner();
  const Direction move_dir = (player == BLACK ? dir : inverse(dir));
  self_count = opp_count = total_count = 0;
  for (Position pos = state.getMobility(dir, bishop.number());
	 pos != king; pos = Board_Table.nextPosition(player, pos, move_dir))
  {
    assert(pos.isOnBoard());
    const Piece piece = state.getPieceAt(pos);
    if (!piece.isEmpty())
    {
      ++total_count;
      if (piece.owner() == player)
	++self_count;
      else
	++opp_count;
    }
  }
}



namespace osl
{
  namespace eval
  {
    namespace experimental
    {
      template class MajorY<true, ROOK>;
      template class MajorY<false, ROOK>;
      template class MajorY<true, BISHOP>;
      template class MajorY<false, BISHOP>;
      template class RookPawn<true>;
      template class RookPawn<false>;
      template class MajorGoldSilverAttacked<false>;
      template MultiInt KingRookBishop::evalOne<BLACK>(const NumEffectState &state);
      template MultiInt KingRookBishop::evalOne<WHITE>(const NumEffectState &state);
    }
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
