#include "osl/eval/experimental/minorPiece.h"
using osl::MultiInt;

void osl::eval::experimental::PawnDropX::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PawnDropBoth::x_table[i][s] = weights.value(i+ONE_DIM*s);
  }
}

osl::misc::CArray<osl::MultiInt, 9>
osl::eval::experimental::PawnDropBoth::attack_table;
osl::misc::CArray<osl::MultiInt, 9>
osl::eval::experimental::PawnDropBoth::defense_table;
osl::misc::CArray<osl::MultiInt, 81>
osl::eval::experimental::PawnDropBoth::attack_y_table;
osl::misc::CArray<osl::MultiInt, 81>
osl::eval::experimental::PawnDropBoth::defense_y_table;
osl::misc::CArray<osl::MultiInt, 90>
osl::eval::experimental::PawnDropBoth::x_table;
osl::misc::CArray<osl::MultiInt, 18>
osl::eval::experimental::PawnDropBoth::stand_table;
osl::misc::CArray<osl::MultiInt, 90>
osl::eval::experimental::PawnDropBoth::x_stand_table;
osl::misc::CArray<osl::MultiInt, 162>
osl::eval::experimental::PawnDropBoth::y_stand_table;

void osl::eval::experimental::PawnDrop::setUp(const Weights &weights,int stage)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    PawnDropBoth::defense_table[i][stage] = weights.value(i);
    PawnDropBoth::attack_table[i][stage] = weights.value(i + ONE_DIM);
  }
}

void osl::eval::experimental::PawnDropY::setUp(const Weights &weights,int stage)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    PawnDropBoth::attack_y_table[i][stage] = weights.value(i);
    PawnDropBoth::defense_y_table[i][stage] = weights.value(i + ONE_DIM);
  }
}

void osl::eval::experimental::PawnDropPawnStand::setUp(const Weights &weights)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PawnDropBoth::stand_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::PawnDropPawnStandX::setUp(const Weights &weights)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PawnDropBoth::x_stand_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}
void osl::eval::experimental::PawnDropPawnStandY::setUp(const Weights &weights)
{
  for (int i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PawnDropBoth::y_stand_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

osl::MultiInt osl::eval::experimental::PawnDropBoth::eval(
  const NumEffectState &state)
{
  osl::MultiInt result;
  CArray<Position, 2> kings = {{state.getKingPosition<BLACK>(),
			       state.getKingPosition<WHITE>()}};
  CArray<Piece, 2> king_piece = {{state.getKingPiece<BLACK>(),
				  state.getKingPiece<WHITE>()}};
  CArray<bool, 2> has_pawn = {{state.hasPieceOnStand<PAWN>(BLACK),
			       state.hasPieceOnStand<PAWN>(WHITE)}};
  for (int x = 1; x <= 9; ++x)
  {
    if (!state.isPawnMaskSet<BLACK>(x))
    {
      const int attack_index = index(kings[WHITE], x);
      const int defense_index = index(kings[BLACK], x);
      const int attack_index_x =
	indexX<true>(king_piece[WHITE], x);
      const int defense_index_x =
	indexX<false>(king_piece[BLACK], x);
      const int attack_index_y = indexY<WHITE>(king_piece[WHITE], x);
      const int defense_index_y = indexY<BLACK>(king_piece[BLACK], x);
      result += value(attack_index, defense_index,
		      attack_index_y, defense_index_y,
		      attack_index_x, defense_index_x);
      if (has_pawn[BLACK])
      {
	result += standValue(attack_index, defense_index,
			     attack_index_y, defense_index_y,
			     attack_index_x, defense_index_x);
      }
    }
    if (!state.isPawnMaskSet<WHITE>(x))
    {
      const int attack_index = index(kings[BLACK], x);
      const int defense_index = index(kings[WHITE], x);
      const int attack_index_x =
	indexX<true>(king_piece[BLACK], x);
      const int defense_index_x =
	indexX<false>(king_piece[WHITE], x);
      const int attack_index_y = indexY<BLACK>(king_piece[BLACK], x);
      const int defense_index_y = indexY<WHITE>(king_piece[WHITE], x);
      result -= value(attack_index, defense_index,
		      attack_index_y, defense_index_y,
		      attack_index_x, defense_index_x);
      if (has_pawn[WHITE])
      {
	result -= standValue(attack_index, defense_index,
			     attack_index_y, defense_index_y,
			     attack_index_x, defense_index_x);
      }
    }
  }
  return result;
}



osl::MultiInt osl::eval::experimental::NoPawnOnStand::weight;

void osl::eval::experimental::
NoPawnOnStand::setUp(const Weights &weights,int stage)
{
  weight[stage] = weights.value(0);
}



osl::misc::CArray<osl::MultiInt, 9> osl::eval::experimental::PawnAdvance::table;

void osl::eval::experimental::
PawnAdvance::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i) {
    table[i][stage] = weights.value(i);
  }
}

osl::MultiInt osl::eval::experimental::PawnAdvance::eval(
  const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<PAWN>::indexMin;
       i < PtypeTraits<PAWN>::indexLimit; ++i)
  {
    const Piece pawn = state.getPieceOf(i);
    if (pawn.isOnBoard() && !pawn.isPromoted() &&
	cantAdvance(state, pawn))
    {
      if (pawn.owner() == BLACK)
	result += table[index(BLACK, pawn.position())];
      else
	result -= table[index(WHITE, pawn.position())];
    }
  }
  return result;
}

template <osl::Player P> inline
void osl::eval::experimental::
PawnAdvanceAll::adjust(int index, MultiInt& values)
{
  if(P==BLACK)
    values += PawnAdvance::table[index];
  else
    values -= PawnAdvance::table[index];
}

template <osl::Player P>
void osl::eval::experimental::
PawnAdvanceAll::evalWithUpdateBang(const osl::state::NumEffectState &state,
				   osl::Move moved, MultiInt& values)
{
  assert(moved.player() == P);
  if (moved.ptype() == PAWN)
  {
    if (cantAdvance(state, moved.ptypeO(), moved.to()))
    {
      adjust<P>(index(P, moved.to()), values);
      return;
    }
  }
  const Player Opponent = PlayerTraits<P>::opponent;
  Ptype captured = moved.capturePtype();
  if (captured == PAWN)
  {
    if (cantAdvance(state, moved.capturePtypeO(), moved.to()))
      adjust<P>(index(Opponent, moved.to()), values);
  }
  else if (captured != PTYPE_EMPTY)
  {
    const Piece piece = state.getPieceAt(
      moved.to() + DirectionPlayerTraits<D, Opponent>::offset());
    if (piece.isOnBoardByOwner<Opponent>() &&
	piece.ptype() == PAWN)
      adjust<P>(index(Opponent, piece.position()), values);
  }
  if (!moved.isDrop())
  {
    const Piece piece = state.getPieceAt(
      moved.from() + DirectionPlayerTraits<D, P>::offset());
    if (piece.isOnBoardByOwner<P>() &&
	piece.ptype() == PAWN)
      adjust<Opponent>(index(P, piece.position()), values);
  }
  {
    const Piece piece = state.getPieceAt(
      moved.to()+DirectionPlayerTraits<D,P>::offset());
    if (piece.isOnBoardByOwner<P>() &&
	piece.ptype() == PAWN)
      adjust<P>(index(P, piece.position()), values);
  }
}



osl::misc::CArray<osl::MultiInt, 153>
osl::eval::experimental::SilverFeatures::head_table;
osl::misc::CArray<osl::MultiInt, 9>
osl::eval::experimental::SilverFeatures::retreat_table;

void osl::eval::experimental::
SilverHeadPawnKingRelative::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      head_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

osl::MultiInt osl::eval::experimental::
SilverFeatures::eval(const NumEffectState &state)
{
  MultiInt result;
  CArray<Position, 2> kings = {{state.getKingPosition<BLACK>(),
				state.getKingPosition<WHITE>()}};
  for (int i = PtypeTraits<SILVER>::indexMin;
       i < PtypeTraits<SILVER>::indexLimit; ++i)
  {
    const Piece silver = state.getPieceOf(i);
    if (!silver.isOnBoard() || silver.isPromoted()) continue;
    if (silver.owner()==BLACK){
      result += evalOne<BLACK>(state, silver, kings);
    }
    else{
      result -= evalOne<WHITE>(state, silver, kings);
    }
  }
  return result;
}

template<osl::Player P>
inline
bool osl::eval::experimental::
SilverFeatures::canRetreat(const osl::state::NumEffectState &state,
			   const osl::Piece silver)
{
  assert(silver.isOnBoard() && !silver.isPromoted());
  assert(silver.owner()==P);
  if ((P == BLACK && silver.position().y() != 9) ||
      (P == WHITE && silver.position().y() != 1))
  {
    Position dl = silver.position()+DirectionPlayerTraits<DL,P>::offset();
    Piece pdl = state.getPieceAt(dl);
    if (!pdl.canMoveOn<P>() ||
	state.hasEffectBy(alt(P), dl))
    {
    Position dr = silver.position()+DirectionPlayerTraits<DR,P>::offset();
      Piece pdr = state.getPieceAt(dr);
      if (!pdr.canMoveOn<P>() ||
	  state.hasEffectBy(alt(P), dr))
      {
	return false;
      }
    }    
  }
  return true;
}

void osl::eval::experimental::
SilverRetreat::setUp(const Weights &weights, int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i) {
    retreat_table[i][stage] = weights.value(i);
  }
}


osl::misc::CArray<osl::MultiInt, 153>
osl::eval::experimental::GoldFeatures::knight_table;
osl::misc::CArray<osl::MultiInt, 9>
osl::eval::experimental::GoldFeatures::retreat_table;

void osl::eval::experimental::
GoldKnightKingRelative::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      knight_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

osl::MultiInt osl::eval::experimental::
GoldFeatures::eval(const NumEffectState &state)
{
  MultiInt result;
  CArray<Position, 2> kings = {{state.getKingPosition<BLACK>(),
				state.getKingPosition<WHITE>()}};
  for (int i = PtypeTraits<GOLD>::indexMin;
       i < PtypeTraits<GOLD>::indexLimit; ++i)
  {
    const Piece gold = state.getPieceOf(i);
    if (!gold.isOnBoard())
      continue;
    if (gold.owner() == BLACK)
    {
      result += evalOne<BLACK>(state, gold, kings);
    }
    else
    {
      result -= evalOne<WHITE>(state, gold, kings);
    }
  }
  return result;
}

template<osl::Player P>
inline
bool osl::eval::experimental::
GoldFeatures::canRetreat(const osl::state::NumEffectState &state,
				 const osl::Piece gold)
{
  assert(gold.isOnBoard());
  assert(P==gold.owner());

  if ((P == BLACK && gold.position().y() != 9) ||
      (P == WHITE && gold.position().y() != 1))
  {
    Position d = gold.position()+DirectionPlayerTraits<D,P>::offset();
    if ((state.getPieceAt(d).isOnBoardByOwner(P) ||
	 state.hasEffectBy(alt(P), d)))
    {
      return false;
    }
  }
  return true;
}

void osl::eval::experimental::
GoldRetreat::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i) {
    retreat_table[i][stage] = weights.value(i);
  }
}



osl::misc::CArray<MultiInt, 9> osl::eval::experimental::KnightAdvance::table;

template<osl::Player P>
inline
bool osl::eval::experimental::
KnightAdvance::cantAdvance(const osl::state::NumEffectState &state,
				  const osl::Piece knight)
{
  // knight ŨذܤˤʤȲ
  // ⤷Position(1,1)UUR˾
  assert(P==knight.owner());
  Position uul = knight.position()+DirectionPlayerTraits<UUL,P>::offset();
  const Piece puul = state.getPieceAt(uul);
  if (!puul.canMoveOn<P>())
  {
    Position uur = knight.position()+DirectionPlayerTraits<UUR,P>::offset();
    const Piece puur = state.getPieceAt(uur);
    if (!puur.canMoveOn<P>())
      return true;
  }
  return false;
}

void osl::eval::experimental::
KnightAdvance::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i) {
    table[i][stage] = weights.value(i);
  }
}

MultiInt osl::eval::experimental::KnightAdvance::eval(
  const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<KNIGHT>::indexMin;
       i < PtypeTraits<KNIGHT>::indexLimit; ++i)
  {
    const Piece knight = state.getPieceOf(i);
    if (!knight.isOnBoard() || knight.isPromoted()) continue;
    if (knight.owner() == BLACK){
      if(cantAdvance<BLACK>(state,knight))
	result += table[index(BLACK, knight.position())];
    }
    else if(cantAdvance<WHITE>(state,knight)){
      result -= table[index(WHITE, knight.position())];
    }
  }
  return result;
}


MultiInt osl::eval::experimental::AllGold::weight;

void osl::eval::experimental::
AllGold::setUp(const Weights &weights,int stage)
{
  weight[stage] = weights.value(0);
}



osl::misc::CArray<MultiInt, 144> osl::eval::experimental::PtypeY::table;

void osl::eval::experimental::
PtypeY::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i][stage] = weights.value(i);
  }
}

MultiInt osl::eval::experimental::PtypeY::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = 0; i < Piece::SIZE; ++i)
  {
    const Piece p = state.getPieceOf(i);
    if (!p.isOnBoard())
      continue;
    if (p.owner() == BLACK)
      result += table[index(BLACK,p.ptype(),p.position())];
    else
      result -= table[index(WHITE,p.ptype(),p.position())];
  }
  return result;
}

template<osl::Player P>
MultiInt osl::eval::experimental::
PtypeY::evalWithUpdate(const NumEffectState &, Move moved,
		       MultiInt const& last_value)
{
  MultiInt result(last_value);

  if (!moved.isDrop())
  {
    if (P == BLACK)
      result -= table[index(BLACK, moved.oldPtype(), moved.from())];
    else
      result += table[index(WHITE, moved.oldPtype(), moved.from())];
  }
  Ptype captured = moved.capturePtype();
  if (captured != PTYPE_EMPTY)
  {
    const MultiInt weight =
      table[index(alt(P), captured, moved.to())];
    if (P == BLACK)
      result += weight;
    else
      result -= weight;
  }
  {
    if (P == BLACK)
      result += table[index(BLACK, moved.ptype(), moved.to())];
    else
      result -= table[index(WHITE, moved.ptype(), moved.to())];
  }

  return result;
}


osl::misc::CArray<MultiInt, 80> osl::eval::experimental::PtypeX::table;

void osl::eval::experimental::
PtypeX::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i][stage] = weights.value(i);
  }
}

MultiInt osl::eval::experimental::PtypeX::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = 0; i < Piece::SIZE; ++i)
  {
    const Piece p = state.getPieceOf(i);
    if (!p.isOnBoard())
      continue;
    if (p.owner() == BLACK)
      result += table[index(BLACK,p.ptype(),p.position())];
    else
      result -= table[index(WHITE,p.ptype(),p.position())];
  }
  return result;
}

template<osl::Player P>
MultiInt osl::eval::experimental::
PtypeX::evalWithUpdate(const NumEffectState &, Move moved,
		       MultiInt const& last_value)
{
  MultiInt result(last_value);

  if (!moved.isDrop())
    {
      if (P == BLACK)
	result -= table[index(BLACK, moved.oldPtype(), moved.from())];
      else
	result += table[index(WHITE, moved.oldPtype(), moved.from())];
      Ptype captured = moved.capturePtype();
      if (captured != PTYPE_EMPTY)
	{
	  if (P == BLACK)
	    result += table[index(WHITE, captured, moved.to())];
	  else
	    result -= table[index(BLACK, captured, moved.to())];
	}
    }
  if (P == BLACK)
    result += table[index(BLACK, moved.ptype(), moved.to())];
  else
    result -= table[index(WHITE, moved.ptype(), moved.to())];
  return result;
}



#ifndef MINIMAL
template <bool Opening>
osl::misc::CArray<int, osl::PTYPEO_SIZE> osl::eval::experimental::PawnAndEmptyAndPiece<Opening>::table;
template <bool Opening>
osl::misc::CArray<int, osl::PTYPEO_SIZE*9> osl::eval::experimental::PawnAndEmptyAndPiece<Opening>::y_table;

template <bool Opening>
void osl::eval::experimental::
PawnAndEmptyAndPiece<Opening>::setUp(const Weights &weights)
{
  for (size_t i = 0; i < DIM_BASE; ++i)
  {
    table[i] = weights.value(i);
  }
  for (size_t i = 0; i < DIM_Y; ++i)
  {
    y_table[i] = weights.value(DIM_BASE + 1);
  }
}

template <bool Opening>
int osl::eval::experimental::PawnAndEmptyAndPiece<Opening>::eval(
  const NumEffectState &state)
{
  int result = 0;
  for (int i = PtypeTraits<PAWN>::indexMin;
       i < PtypeTraits<PAWN>::indexLimit; ++i)
  {
    const Piece pawn = state.getPieceOf(i);
    if (pawn.isOnBoard() && !pawn.isPromoted() &&
	((pawn.owner() == BLACK && pawn.position().y() >= 3) ||
	 (pawn.owner() == WHITE && pawn.position().y() <= 7)))
    {
      if (pawn.owner() == BLACK &&
	  state.getPieceAt(Position(pawn.position().x(),
				    pawn.position().y() - 1)).isEmpty())
      {
	result += value(pawn,
			state.getPieceAt(Position(pawn.position().x(),
						  pawn.position().y() - 2)));
      }
      else if (pawn.owner() == WHITE &&
	       state.getPieceAt(Position(pawn.position().x(),
					 pawn.position().y() + 1)).isEmpty())
      {
	result -= value(pawn,
			state.getPieceAt(Position(pawn.position().x(),
						  pawn.position().y() + 2)));
      }
    }
  }
  return result;
}

template <bool Opening>
int osl::eval::experimental::
PawnAndEmptyAndPiece<Opening>::pawnBonusFrom(const NumEffectState &state,
					     const Player player,
					     const Move move)
{
  if (state.isPawnMaskSet(player, move.from().x()))
  {
    Position d = Board_Table.nextPosition(player, move.from(), D);
    const Piece piece_d = state.getPieceAt(d);
    if ((move.capturePtype() == PTYPE_EMPTY && d == move.to())
	|| piece_d.isEmpty())
    {
      const Piece piece =
	state.getPieceAt(Board_Table.nextPosition(player, d, D));
      if (piece.isOnBoardByOwner(player) &&
	  piece.ptype() == PAWN)
      {
	int weight = value(piece.owner(),
			   NEW_PTYPEO(move.player(),
				      move.oldPtype()),
			   piece.position());
	if (d != move.to())
	  weight -= value(piece, Piece::EMPTY());
	if (piece.owner() == BLACK)
	  return weight;
	else
	  return -weight;
      }
    }
    else if (piece_d.isOnBoardByOwner(player) &&
	     piece_d.ptype() == PAWN)
    {
      const Position up = Board_Table.nextPosition(player, move.from(), U);
      const Piece piece = state.getPieceAt(up);
      if (move.to() != up &&!piece.isEdge())
      {
	const int weight = move.to() == up ?
	  value(piece_d.owner(),
		(move.capturePtype() == PTYPE_EMPTY ?
		 PTYPEO_EMPTY : move.capturePtypeO()), piece_d.position()) :
	  value(piece_d, piece);
	if (piece_d.owner() == BLACK)
	  return -weight;
	else
	  return weight;
      }
    }
  }
  return 0;
}

template <bool Opening>
int osl::eval::experimental::
PawnAndEmptyAndPiece<Opening>::pawnBonusCapture(const NumEffectState &state,
						const Player player,
						const Move move)
{
  if (state.isPawnMaskSet(player, move.to().x()))
  {
    Position d = Board_Table.nextPosition(player, move.to(), D);
    if (state.getPieceAt(d).isEmpty() && (d != move.from()))
    {
      const Piece piece =
	state.getPieceAt(Board_Table.nextPosition(player, d, D));
      if (piece.isOnBoardByOwner(player) &&
	  piece.ptype() == PAWN)
      {
	const int weight = value(piece.owner(),
				NEW_PTYPEO(alt(move.player()),
					   move.capturePtype()),
				 piece.position());
	if (piece.owner() == BLACK)
	  return weight;
	else
	  return -weight;
      }
    }
  }
  return 0;
}

template <bool Opening>
int osl::eval::experimental::
PawnAndEmptyAndPiece<Opening>::pawnBonusTo(const NumEffectState &state,
					   const Player player,
					   const Move move)
{
  if (state.isPawnMaskSet(player, move.to().x()))
  {
    Position d = Board_Table.nextPosition(player, move.to(), D);
    const Piece piece_d = state.getPieceAt(d);
    if (piece_d.isEmpty())
    {
      const Piece piece =
	state.getPieceAt(Board_Table.nextPosition(player, d, D));
      if (piece.isOnBoardByOwner(player) &&
	  piece.ptype() == PAWN)
      {
	int weight = value(piece,
			   state.getPieceAt(move.to()));
	if (move.capturePtype() == PTYPE_EMPTY && move.from() != d)
	{
	  weight -= value(piece, Piece::EMPTY());
	}
	if (piece.owner() == BLACK)
	  return weight;
	else
	  return -weight;
      }
    }
    else if (piece_d.isOnBoardByOwner(player) &&
	     piece_d.ptype() == PAWN && move.capturePtype() == PTYPE_EMPTY)
    {
      const Position up = Board_Table.nextPosition(player, move.to(), U);
      const Piece piece = state.getPieceAt(up);
      if (up != move.from() && !piece.isEdge())
      {
	const int weight = value(piece_d, piece);
	if (piece_d.owner() == BLACK)
	  return -weight;
	else
	  return +weight;
      }
    }
  }
  return 0;
}

template <bool Opening>
int osl::eval::experimental::
PawnAndEmptyAndPiece<Opening>::evalWithUpdate(
  const NumEffectState &state, Move moved,
  int last_value)
{
  Ptype captured = moved.capturePtype();

  if (!moved.isDrop())
  {
    if (moved.oldPtype() == PAWN)
    {
      if (captured == PTYPE_EMPTY)
      {
	const Position pos =
	  Board_Table.nextPosition(moved.player(), moved.to(), U);
	if (pos.isOnBoard())
	{
	  const int weight = value(moved.player(),
				   state.getPieceAt(pos).ptypeO(),
				   moved.to());
	  if (moved.player() == BLACK)
	    last_value -= weight;
	  else
	    last_value += weight;
	}
      }
    }
    else
    {
      last_value -= pawnBonusFrom(state, moved.player(), moved);
    }
    last_value -= pawnBonusFrom(state, alt(moved.player()), moved);
  }

  if (captured != PTYPE_EMPTY)
  {
    if (captured == PAWN)
    {
      Position d = Board_Table.nextPosition(moved.player(), moved.to(), D);
      if (moved.from() != d && state.getPieceAt(d).isEmpty())
      {
	const Position target = Board_Table.nextPosition(moved.player(),
							 d, D);
	const Piece piece = state.getPieceAt(target);
	if (!piece.isEdge())
	{
	  const int weight = value(alt(moved.player()),
				   target == moved.from() ?
				   moved.oldPtypeO() :
				   piece.ptypeO(),
				   moved.to());
	  if (moved.player() == BLACK)
	    last_value += weight;
	  else
	    last_value -= weight;
	}
      }
    }
    else
    {
      last_value -= pawnBonusCapture(state, alt(moved.player()), moved);
    }
    last_value -= pawnBonusCapture(state, moved.player(), moved);
  }

  if (moved.ptype() == PAWN)
  {
    const Position pos =
      Board_Table.nextPosition(moved.player(), moved.to(), U);
    if (state.getPieceAt(pos).isEmpty())
    {
      const Position target =
	Board_Table.nextPosition(moved.player(), pos, U);

      if (target.isOnBoard())
      {
	const int weight = value(moved.player(),
				 state.getPieceAt(target).ptypeO(),
				 moved.to());
	if (moved.player() == BLACK)
	  last_value += weight;
	else
	  last_value -= weight;
      }
    }
  }
  else
  {
    last_value += pawnBonusTo(state, moved.player(), moved);
  }
  last_value += pawnBonusTo(state, alt(moved.player()), moved);

  return last_value;
}
#endif

MultiInt osl::eval::experimental::KnightCheck::weight;
osl::misc::CArray<MultiInt, 9> osl::eval::experimental::KnightCheck::y_table;

void osl::eval::experimental::KnightCheck::setUp(const Weights &weights,int stage)
{
  KnightCheck::weight[stage] = weights.value(0);
}

void osl::eval::experimental::
KnightCheckY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      KnightCheck::y_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

MultiInt osl::eval::experimental::
KnightCheck::eval(const NumEffectState &state)
{
  MultiInt result;
  if (canCheck<BLACK>(state))
  {
    const int index_y = indexY<BLACK>(state.getKingPosition<BLACK>().y());
    result += value(index_y);
  }
  if (canCheck<WHITE>(state))
  {
    const int index_y = indexY<WHITE>(state.getKingPosition<WHITE>().y());
    result -= value(index_y);
  }
  return result;
}

osl::misc::CArray<MultiInt, 1024> osl::eval::experimental::PawnPtypeOPtypeO::table;
osl::misc::CArray<MultiInt, 9216> osl::eval::experimental::PawnPtypeOPtypeO::y_table;

void osl::eval::experimental::
PawnPtypeOPtypeO::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
PawnPtypeOPtypeOY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PawnPtypeOPtypeO::y_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

MultiInt osl::eval::experimental::
PawnPtypeOPtypeO::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<PAWN>::indexMin;
       i < PtypeTraits<PAWN>::indexLimit; ++i)
  {
    Piece pawn = state.getPieceOf(i);
    if (pawn.isOnBoard() && !pawn.isPromoted())
    {
      const Position up = Board_Table.nextPosition(pawn.owner(),
						   pawn.position(), U);
      const Position up_up = Board_Table.nextPosition(pawn.owner(),
						      up, U);
      PtypeO up_p =
	(up.isOnBoard() ? state.getPieceAt(up).ptypeO() : PTYPEO_EDGE);
      PtypeO up_up_p =
	(up_up.isOnBoard() ? state.getPieceAt(up_up).ptypeO() : PTYPEO_EDGE);
      const int idx = index(pawn.owner(), up_p, up_up_p);
      const int idx_y = indexY(pawn.owner(), up_p, up_up_p,
			       pawn.position().y());
      if (pawn.owner() == BLACK)
	result += table[idx] + y_table[idx_y];
      else
	result -= table[idx] + y_table[idx_y];
    }
  }
  return result;
}

template<osl::Player P>
MultiInt osl::eval::experimental::
PawnPtypeOPtypeO::evalWithUpdate(const NumEffectState &state, Move moved,
				 const CArray2d<int, 2, 9> &pawns,
				 const MultiInt &last_value)
{
  assert(moved.player()==P);
  MultiInt result(last_value);
  if (!moved.isDrop())
  {
    if (moved.oldPtype() == PAWN)
    {
      const Position up_up = moved.to() + DirectionPlayerTraits<U,P>::offset();
      const PtypeO up_up_p =
	(up_up.isOnBoard() ? state.getPieceAt(up_up).ptypeO() : PTYPEO_EDGE);
      const int i = index(P, moved.capturePtypeOSafe(), up_up_p);
      const int i_y = indexY(P, moved.capturePtypeOSafe(),
			     up_up_p, moved.from().y());
      if (P == BLACK)
	result -= table[i]+y_table[i_y];
      else
	result += table[i]+y_table[i_y];
    }
    if (pawns[BLACK][moved.from().x() - 1] != 0)
    {
      if (pawns[BLACK][moved.from().x() - 1] ==
	  moved.from().y() + 1)
      {
	const Position up_up = moved.from() + DirectionPlayerTraits<U,BLACK>::offset();
	const PtypeO up_up_p =
	  (up_up.isOnBoard() ? (up_up == moved.to() ? moved.capturePtypeOSafe() :
				state.getPieceAt(up_up).ptypeO()) :
	   PTYPEO_EDGE);
	const int i = index(BLACK, moved.oldPtypeO(), up_up_p);
	const int i_y = indexY(BLACK, moved.oldPtypeO(), up_up_p,
			      moved.from().y() + 1);
	result -= table[i]+y_table[i_y];
	if (up_up != moved.to())
	{
	  const int new_i = index(BLACK, PTYPEO_EMPTY, up_up_p);
	  const int new_i_y = indexY(BLACK, PTYPEO_EMPTY, up_up_p,
				     moved.from().y() + 1);
	  result += table[new_i]+y_table[new_i_y];
	}
      }
      if (pawns[BLACK][moved.from().x() - 1] ==
	  moved.from().y() + 2)
      {
	const Position up = moved.from() + DirectionPlayerTraits<D,BLACK>::offset();
	const PtypeO up_p =
	  (up.isOnBoard() ? (up == moved.to() ? moved.capturePtypeOSafe() :
			     state.getPieceAt(up).ptypeO()) : PTYPEO_EDGE);
	const int i = index(BLACK, up_p, moved.oldPtypeO());
	const int i_y = indexY(BLACK, up_p, moved.oldPtypeO(),
			       moved.from().y() + 2);
	result -= table[i]+y_table[i_y];
	if (moved.to() != up)
	{
	  const int new_i = index(BLACK, up_p, PTYPEO_EMPTY);
	  const int new_i_y = indexY(BLACK, up_p, PTYPEO_EMPTY,
				     moved.from().y() + 2);
	  result += table[new_i]+y_table[new_i_y];
	}
      }
    }
    if (pawns[WHITE][moved.from().x() - 1] != 0)
    {
      if (pawns[WHITE][moved.from().x() - 1] ==
	  moved.from().y() - 1)
      {
	const Position up_up = moved.from() + DirectionPlayerTraits<U,WHITE>::offset(); 	
	const PtypeO up_up_p =
	  (up_up.isOnBoard() ? (up_up == moved.to() ? moved.capturePtypeOSafe() :
				state.getPieceAt(up_up).ptypeO()) :
	   PTYPEO_EDGE);
	const int i = index(WHITE, moved.oldPtypeO(), up_up_p);
	const int i_y = indexY(WHITE, moved.oldPtypeO(), up_up_p,
			       moved.from().y() - 1);
	result += table[i]+y_table[i_y];
	if (moved.to() != up_up)
	{
	  const int new_i = index(WHITE, PTYPEO_EMPTY, up_up_p);
	  const int new_i_y = indexY(WHITE, PTYPEO_EMPTY, up_up_p,
				     moved.from().y() - 1);
	  result -= table[new_i]+y_table[new_i_y];
	}
      }
      if (pawns[WHITE][moved.from().x() - 1] ==
	  moved.from().y() - 2)
      {
	const Position up = moved.from() + DirectionPlayerTraits<D,WHITE>::offset(); 	
	const PtypeO up_p =
	  (up.isOnBoard() ? (up == moved.to() ? moved.capturePtypeOSafe() :
			     state.getPieceAt(up).ptypeO()) : PTYPEO_EDGE);
	const int i = index(WHITE, up_p, moved.oldPtypeO());
	const int i_y = indexY(WHITE, up_p, moved.oldPtypeO(),
			       moved.from().y() - 2);
	result += table[i]+y_table[i_y];
	if (moved.to() != up)
	{
	  const int new_i = index(WHITE, up_p, PTYPEO_EMPTY);
	  const int new_i_y = indexY(WHITE, up_p, PTYPEO_EMPTY,
				     moved.from().y() - 2);
	  result -= table[new_i]+y_table[new_i_y];
	}
      }
    }
  }
  Ptype captured = moved.capturePtype();
  if (captured == PAWN)
  {
    const Position up = moved.to() + DirectionPlayerTraits<D,P>::offset(); 	
    const Position up_up = up + DirectionPlayerTraits<D,P>::offset(); 	
    const PtypeO up_p =
      (up.isOnBoard() ? (up == moved.from() ? moved.oldPtypeO() :
			 state.getPieceAt(up).ptypeO()) : PTYPEO_EDGE);
    const PtypeO up_up_p =
      (up_up.isOnBoard() ? (up_up == moved.from() ? moved.oldPtypeO() :
			    state.getPieceAt(up_up).ptypeO()) : PTYPEO_EDGE);
    const int i = index(alt(P), up_p, up_up_p);
    const int i_y = indexY(alt(P), up_p, up_up_p,
			   moved.to().y());
    if (P == BLACK)
    {
      result += table[i]+y_table[i_y];
    }
    else
    {
      result -= table[i]+y_table[i_y];
    }
  }
  if (moved.ptype() == PAWN)
  {
    const Position up = moved.to() + DirectionPlayerTraits<U,P>::offset();
    const Position up_up = up + DirectionPlayerTraits<U,P>::offset();
    const PtypeO up_p =
      (up.isOnBoard() ? (up == moved.from() ? moved.oldPtypeO() :
			 state.getPieceAt(up).ptypeO()) : PTYPEO_EDGE);
    const PtypeO up_up_p =
      (up_up.isOnBoard() ? (up_up == moved.from() ? moved.oldPtypeO() :
			    state.getPieceAt(up_up).ptypeO()) : PTYPEO_EDGE);
    const int i = index(P, up_p, up_up_p);
    const int i_y = indexY(P, up_p, up_up_p, moved.to().y());
    if (P == BLACK)
    {
      result += table[i]+y_table[i_y];
    }
    else
    {
      result -= table[i]+y_table[i_y];
    }
  }
  if (pawns[BLACK][moved.to().x() - 1] != 0)
  {
    if (pawns[BLACK][moved.to().x() - 1] ==
	moved.to().y() + 1)
    {
      const Position up_up = moved.to() + DirectionPlayerTraits<U,BLACK>::offset();
      const PtypeO up_up_p =
	(up_up.isOnBoard() ? state.getPieceAt(up_up).ptypeO() :
	 PTYPEO_EDGE);
      const int i = index(BLACK, moved.ptypeO(), up_up_p);
      const int i_y = indexY(BLACK, moved.ptypeO(), up_up_p,
			     moved.to().y() + 1);
      result += table[i]+y_table[i_y];
      if (moved.isDrop() || moved.from() != up_up)
      {
	const int old_i = index(BLACK, moved.capturePtypeOSafe(), up_up_p);
	const int old_i_y = indexY(BLACK, moved.capturePtypeOSafe(),
				   up_up_p, moved.to().y() + 1);
	result -= table[old_i]+y_table[old_i_y];
      }
    }
    if (pawns[BLACK][moved.to().x() - 1] ==
	moved.to().y() + 2)
    {
      const Position up = moved.to() + DirectionPlayerTraits<D,BLACK>::offset();
      const PtypeO up_p =
	(up.isOnBoard() ? state.getPieceAt(up).ptypeO() : PTYPEO_EDGE);
      const int i = index(BLACK, up_p, moved.ptypeO());
      const int i_y = indexY(BLACK, up_p, moved.ptypeO(), moved.to().y() + 2);
      result += table[i]+y_table[i_y];
      if (moved.isDrop() || up != moved.from())
      {
	const int old_i = index(BLACK, up_p, moved.capturePtypeOSafe());
	const int old_i_y = indexY(BLACK, up_p, moved.capturePtypeOSafe(),
				   moved.to().y() + 2);
	result -= table[old_i]+y_table[old_i_y];
      }
    }
  }
  if (pawns[WHITE][moved.to().x() - 1] != 0)
  {
    if (pawns[WHITE][moved.to().x() - 1] ==
	moved.to().y() - 1)
    {
      const Position up_up = moved.to() + DirectionPlayerTraits<U,WHITE>::offset();
      const PtypeO up_up_p =
	(up_up.isOnBoard() ? state.getPieceAt(up_up).ptypeO() :
	 PTYPEO_EDGE);
      const int i = index(WHITE, moved.ptypeO(), up_up_p);
      const int i_y = indexY(WHITE, moved.ptypeO(), up_up_p,
			     moved.to().y() - 1);
      result -= table[i]+y_table[i_y];
      if (up_up != moved.from())
      {
	const int old_i = index(WHITE, moved.capturePtypeOSafe(), up_up_p);
	const int old_i_y = indexY(WHITE, moved.capturePtypeOSafe(), up_up_p,
				   moved.to().y() - 1);
	result += table[old_i]+y_table[old_i_y];
      }
    }
    if (pawns[WHITE][moved.to().x() - 1] ==
	moved.to().y() - 2)
    {
      const Position up = moved.to() + DirectionPlayerTraits<D,WHITE>::offset();
      const PtypeO up_p =
	(up.isOnBoard() ? state.getPieceAt(up).ptypeO() : PTYPEO_EDGE);
      const int i = index(WHITE, up_p, moved.ptypeO());
      const int i_y = indexY(WHITE, up_p, moved.ptypeO(), moved.to().y() - 2);
      result -= table[i]+y_table[i_y];
      if (moved.isDrop() || up != moved.from())
      {
	const int old_i = index(WHITE, up_p, moved.capturePtypeOSafe());
	const int old_i_y = indexY(WHITE, up_p, moved.capturePtypeOSafe(),
				   moved.to().y() - 2);
	result += table[old_i]+y_table[old_i_y];
      }
    }
  }
  return result;
}



osl::misc::CArray<MultiInt, 9> osl::eval::experimental::PromotedMinorPieces::table;
osl::misc::CArray<MultiInt, 162> osl::eval::experimental::PromotedMinorPieces::y_table;

void osl::eval::experimental::
PromotedMinorPieces::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
PromotedMinorPiecesY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PromotedMinorPieces::y_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

template <int Sign>
inline void osl::eval::experimental::
PromotedMinorPieces::adjust(int index, int index_attack, int index_defense,
			    MultiInt &result)
{
  if(Sign>0)
    result+= table[index] + y_table[index_attack] + y_table[index_defense];
  else
    result-= table[index] + y_table[index_attack] + y_table[index_defense];
}
template <osl::Player P>
void osl::eval::experimental::
PromotedMinorPieces::evalOne(const NumEffectState &state,
			     const PieceMask promoted,
			     MultiInt &result)
{
  PieceMask attack = promoted & state.getOnBoardMask(P);
  const Position king = state.getKingPosition<PlayerTraits<P>::opponent>();
  const Position self_king = state.getKingPosition<P>();
  int min_left = -10;
  int min_right = 10;
  while (attack.any())
  {
    const Piece p = state.getPieceOf(attack.takeOneBit());
    const int x_diff = (P == BLACK ? p.position().x() - king.x() :
			king.x() - p.position().x());
    if (x_diff <= 0)
    {
      if (x_diff > min_left)
      {
	if (min_left != -10)
	{
	  if (P == BLACK)
	    adjust<1>(-min_left, indexY<true, P>(king, -min_left),
		      indexY<false, P>(self_king, -min_left), result);
	  else
	    adjust<-1>(-min_left, indexY<true, P>(king, -min_left),
		       indexY<false, P>(self_king, -min_left), result);
	}
	min_left = x_diff;
      }
      else
      {
	if (P == BLACK)
	  adjust<1>(-x_diff, indexY<true, P>(king, -x_diff),
		    indexY<false, P>(self_king, -x_diff),
		    result);
	else
	  adjust<-1>(-x_diff, indexY<true, P>(king, -x_diff),
		     indexY<false, P>(self_king, -x_diff),
		     result);
      }
    }
    if (x_diff >= 0)
    {
      if (x_diff < min_right)
      {
	if (min_right != 10)
	{
	  if (P == BLACK)
	    adjust<1>(min_right, indexY<true, P>(king, min_right),
		      indexY<false, P>(self_king, min_right),
		      result);
	  else
	    adjust<-1>(min_right, indexY<true, P>(king, min_right),
		       indexY<false, P>(self_king, min_right),
		       result);
	}
	min_right = x_diff;
      }
      else if (x_diff != 0)
      {
	if (P == BLACK)
	  adjust<1>(x_diff, indexY<true, P>(king, x_diff),
		    indexY<false, P>(self_king, x_diff),
		    result);
	else
	  adjust<-1>(x_diff, indexY<true, P>(king, x_diff),
		     indexY<false, P>(self_king, x_diff),
		     result);
      }
    }
  }
}

MultiInt osl::eval::experimental::
PromotedMinorPieces::eval(const NumEffectState &state)
{
  MultiInt result;
  PieceMask promoted_pieces = state.promotedPieces();
  promoted_pieces.clearBit<ROOK>();
  promoted_pieces.clearBit<BISHOP>();
  if (promoted_pieces.none())
    return result;

  evalOne<BLACK>(state, promoted_pieces, result);
  evalOne<WHITE>(state, promoted_pieces, result);
  return result;
}

MultiInt osl::eval::experimental::
PromotedMinorPieces::evalWithUpdate(const NumEffectState &state,
				    Move moved,
				    const MultiInt &last_values)
{
  Ptype captured = moved.capturePtype();
  if (moved.ptype() == KING ||
      (isPromoted(moved.ptype()) && !isMajor(moved.ptype())) ||
      (captured != PTYPE_EMPTY && isPromoted(captured) &&
       !isMajor(captured)))
    return eval(state);

  return last_values;
}


osl::misc::CArray<MultiInt, 64> osl::eval::experimental::NonPawnAttacked::table;
osl::misc::CArray<MultiInt, 19584> osl::eval::experimental::NonPawnAttacked::king_table;

void osl::eval::experimental::NonPawnAttacked::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::NonPawnAttackedKingRelative::setUp(
  const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      NonPawnAttacked::king_table[i][s] = weights.value(i + ONE_DIM*s);
  }
  for(int x_diff=0;x_diff<9;x_diff++)
    for(int y_diff= -8;y_diff<=8;y_diff++)
      for(int has_support=0;has_support<2;has_support++)
	for(int same_turn=0;same_turn<2;same_turn++)
	  for(int ptype=0;ptype<PTYPE_SIZE;ptype++){
	    int index=((ptype + (same_turn ? 0 : PTYPE_SIZE) +
			(has_support ? 0 : PTYPE_SIZE*2))* 9 + x_diff) * 17 +
	      y_diff + 8;
	    int index0=ptype + (same_turn ? 0 : PTYPE_SIZE) +
	      (has_support ? 0 : PTYPE_SIZE * 2);
	    NonPawnAttacked::king_table[index] += NonPawnAttacked::table[index0];
	  }
}

template <int Sign>
void osl::eval::experimental::
NonPawnAttacked::adjust(int black, int white,
			int black_turn_king_attack,
			int black_turn_king_defense,
			int white_turn_king_attack,
			int white_turn_king_defense,
			MultiIntPair &result)
{
  if(Sign>0){
    result[BLACK] += king_table[black_turn_king_attack] +
      king_table[black_turn_king_defense];
    result[WHITE] += king_table[white_turn_king_attack] +
      king_table[white_turn_king_defense];
  }
  else{
    result[BLACK] -= king_table[black_turn_king_attack] +
      king_table[black_turn_king_defense];
    result[WHITE] -= king_table[white_turn_king_attack] +
      king_table[white_turn_king_defense];
  }
}

void osl::eval::experimental::
NonPawnAttacked::eval(const NumEffectState &state, MultiIntPair& result)
{
  result = MultiIntPair();
  CArray<Position, 2> kings = {{state.getKingPosition<BLACK>(),
				state.getKingPosition<WHITE>()}};
  PieceMask black_attacked = state.effectedMask(WHITE) & state.getOnBoardMask(BLACK);
  black_attacked.reset(KingTraits<BLACK>::index);
  mask_t black_ppawn = state.promotedPieces().getMask<PAWN>() & black_attacked.selectBit<PAWN>();
  black_attacked.clearBit<PAWN>();
  black_attacked.orMask(PtypeFuns<PAWN>::indexNum, black_ppawn);
  PieceMask black_with_support = state.effectedMask(BLACK) & black_attacked;
  PieceMask black_without_support = (~state.effectedMask(BLACK)) & black_attacked;
  while (black_with_support.any())
  {
    const Piece piece = state.getPieceOf(black_with_support.takeOneBit());
    const int index_black_turn = index(true, true, piece.ptype());
    const int index_white_turn = index(false, true, piece.ptype());
    const int index_king_black_turn_attack =
      indexK<true>(kings[WHITE], true, true, piece);
    const int index_king_white_turn_attack =
      indexK<true>(kings[WHITE], false, true, piece);
    const int index_king_black_turn_defense =
      indexK<false>(kings[BLACK], true, true, piece);
    const int index_king_white_turn_defense =
      indexK<false>(kings[BLACK], false, true, piece);
    adjust<1>(index_black_turn, index_white_turn,
	      index_king_black_turn_attack, index_king_black_turn_defense,
	      index_king_white_turn_attack, index_king_white_turn_defense,
	      result);
  }
  while (black_without_support.any())
  {
    const Piece piece = state.getPieceOf(black_without_support.takeOneBit());
    const int index_black_turn = index(true, false, piece.ptype());
    const int index_white_turn = index(false, false, piece.ptype());
    const int index_king_black_turn_attack =
      indexK<true>(kings[WHITE], true, false, piece);
    const int index_king_white_turn_attack =
      indexK<true>(kings[WHITE], false, false, piece);
    const int index_king_black_turn_defense =
      indexK<false>(kings[BLACK], true, false, piece);
    const int index_king_white_turn_defense =
      indexK<false>(kings[BLACK], false, false, piece);
    adjust<1>(index_black_turn, index_white_turn,
	      index_king_black_turn_attack, index_king_black_turn_defense,
	      index_king_white_turn_attack, index_king_white_turn_defense,
	      result);
  }

  PieceMask white_attacked = state.effectedMask(BLACK) & state.getOnBoardMask(WHITE);
  white_attacked.reset(KingTraits<WHITE>::index);
  mask_t white_ppawn = state.promotedPieces().getMask<PAWN>() & white_attacked.selectBit<PAWN>();
  white_attacked.clearBit<PAWN>();
  white_attacked.orMask(PtypeFuns<PAWN>::indexNum, white_ppawn);
  PieceMask white_with_support = state.effectedMask(WHITE) & white_attacked;
  PieceMask white_without_support = (~state.effectedMask(WHITE)) & white_attacked;
  while (white_with_support.any())
  {
    const Piece piece = state.getPieceOf(white_with_support.takeOneBit());
    const int index_black_turn = index(false, true, piece.ptype());
    const int index_white_turn = index(true, true, piece.ptype());
    const int index_king_black_turn_attack =
      indexK<true>(kings[BLACK], false, true, piece);
    const int index_king_white_turn_attack =
      indexK<true>(kings[BLACK], true, true, piece);
    const int index_king_black_turn_defense =
      indexK<false>(kings[WHITE], false, true, piece);
    const int index_king_white_turn_defense =
      indexK<false>(kings[WHITE], true, true, piece);
    adjust<-1>(index_black_turn, index_white_turn,
	      index_king_black_turn_attack, index_king_black_turn_defense,
	      index_king_white_turn_attack, index_king_white_turn_defense,
	      result);
  }
  while (white_without_support.any())
  {
    const Piece piece = state.getPieceOf(white_without_support.takeOneBit());
    const int index_black_turn = index(false, false, piece.ptype());
    const int index_white_turn = index(true, false, piece.ptype());
    const int index_king_black_turn_attack =
      indexK<true>(kings[BLACK], false, false, piece);
    const int index_king_white_turn_attack =
      indexK<true>(kings[BLACK], true, false, piece);
    const int index_king_black_turn_defense =
      indexK<false>(kings[WHITE], false, false, piece);
    const int index_king_white_turn_defense =
      indexK<false>(kings[WHITE], true, false, piece);
    adjust<-1>(index_black_turn, index_white_turn,
	      index_king_black_turn_attack, index_king_black_turn_defense,
	      index_king_white_turn_attack, index_king_white_turn_defense,
	      result);
  }
}

template<osl::Player P>
void osl::eval::experimental::
NonPawnAttacked::evalWithUpdateBang(
  const NumEffectState &state,
  Move moved,
  const CArray<PieceMask, 2> &effected,
  MultiIntPair &result)
{
  if (moved.ptype() == KING)
  {
    eval(state, result);
    return;
  }

  CArray<PieceMask, 2> effected_mask = effected;
  effected_mask[0].clearBit<KING>();
  effected_mask[1].clearBit<KING>();
  CArray<PieceMask, 2> new_mask = {{
      state.effectedMask(BLACK),
      state.effectedMask(WHITE)
    }};

  mask_t black_ppawn =
    new_mask[0].selectBit<PAWN>() & state.promotedPieces().getMask<PAWN>();
  mask_t white_ppawn =
    new_mask[1].selectBit<PAWN>() & state.promotedPieces().getMask<PAWN>();
  new_mask[0].clearBit<PAWN>();
  new_mask[1].clearBit<PAWN>();
  new_mask[0].orMask(PtypeFuns<PAWN>::indexNum, black_ppawn);
  new_mask[1].orMask(PtypeFuns<PAWN>::indexNum, white_ppawn);
  new_mask[0].clearBit<KING>();
  new_mask[1].clearBit<KING>();
  CArray<Position, 2> kings = {{ state.getKingPosition<BLACK>(),
				 state.getKingPosition<WHITE>() }};
  const Piece p = state.getPieceAt(moved.to());
  assert(p.owner()==P);
  if (!moved.isDrop())
  {
    if (effected_mask[alt(P)].test(p.number()))
    {
      const bool has_support = effected_mask[P].test(p.number());
      const int index_black_turn = index(BLACK == P,
					 has_support, moved.oldPtype());
      const int index_white_turn = index(WHITE == P,
					 has_support, moved.oldPtype());
      const int index_king_black_turn_attack =
	indexK<true>(kings[alt(P)], BLACK == P,
		     has_support, moved.from(), P, moved.oldPtype());
      const int index_king_white_turn_attack =
	indexK<true>(kings[alt(P)], WHITE == P,
		     has_support, moved.from(), P, moved.oldPtype());
      const int index_king_black_turn_defense =
	indexK<false>(kings[P], BLACK == P,
		      has_support, moved.from(), P, moved.oldPtype());
      const int index_king_white_turn_defense =
	indexK<false>(kings[P], WHITE == P,
		      has_support, moved.from(), P, moved.oldPtype());
      if (P == BLACK)
	adjust<-1>(index_black_turn, index_white_turn,
		   index_king_black_turn_attack, index_king_black_turn_defense,
		   index_king_white_turn_attack, index_king_white_turn_defense,
		   result);
      else
	adjust<1>(index_black_turn, index_white_turn,
		  index_king_black_turn_attack, index_king_black_turn_defense,
		  index_king_white_turn_attack, index_king_white_turn_defense,
		  result);
    }
  }
  if (new_mask[alt(P)].test(p.number()))
  {
    const bool has_support = new_mask[P].test(p.number());
    const int index_black_turn = index(BLACK == P,
				       has_support, p.ptype());
    const int index_white_turn = index(WHITE == P,
				       has_support, p.ptype());
    const int index_king_black_turn_attack =
      indexK<true>(kings[alt(P)], BLACK == P,
		   has_support, p);
    const int index_king_white_turn_attack =
      indexK<true>(kings[alt(P)], WHITE == P,
		   has_support, p);
    const int index_king_black_turn_defense =
      indexK<false>(kings[P], BLACK == P,
		    has_support, p);
    const int index_king_white_turn_defense =
      indexK<false>(kings[P], WHITE == P,
		    has_support, p);
    if (P == BLACK)
      adjust<1>(index_black_turn, index_white_turn,
		index_king_black_turn_attack, index_king_black_turn_defense,
		index_king_white_turn_attack, index_king_white_turn_defense,
		result);
    else
      adjust<-1>(index_black_turn, index_white_turn,
		 index_king_black_turn_attack, index_king_black_turn_defense,
		 index_king_white_turn_attack, index_king_white_turn_defense,
		 result);
  }
  const Ptype captured = moved.capturePtype();
  if (captured != PTYPE_EMPTY && captured != PAWN)
  {
    PieceMask captured_mask =
      effected_mask[P] & (~state.getOnBoardMask(BLACK)) &
      (~state.getOnBoardMask(WHITE));
    
    const bool has_support = effected_mask[alt(P)].test(captured_mask.takeOneBit());
    const int index_black_turn =
      index(WHITE == P, has_support, captured);
    const int index_white_turn =
      index(BLACK == P, has_support, captured);
    const int index_king_black_turn_attack =
      indexK<true>(kings[P], WHITE == P,
		   has_support, moved.to(), alt(P), captured);
    const int index_king_white_turn_attack =
      indexK<true>(kings[P], BLACK == P,
		   has_support, moved.to(), alt(P), captured);
    const int index_king_black_turn_defense =
      indexK<false>(kings[alt(P)], WHITE == P,
		    has_support, moved.to(), alt(P), captured);
    const int index_king_white_turn_defense =
      indexK<false>(kings[alt(P)], BLACK == P,
		    has_support, moved.to(), alt(P), captured);
    if (P == BLACK)
      adjust<1>(index_black_turn, index_white_turn,
		index_king_black_turn_attack, index_king_black_turn_defense,
		index_king_white_turn_attack, index_king_white_turn_defense,
		result);
    else
      adjust<-1>(index_black_turn, index_white_turn,
		 index_king_black_turn_attack, index_king_black_turn_defense,
		 index_king_white_turn_attack, index_king_white_turn_defense,
		 result);
  }

  updateEffectChanged<BLACK>(state, effected_mask, new_mask, p.number(),
			     result);
  updateEffectChanged<WHITE>(state, effected_mask, new_mask, p.number(),
			     result);
}


osl::misc::CArray<MultiInt, 9> osl::eval::experimental::KnightHead::table;
osl::misc::CArray<MultiInt, 144> osl::eval::experimental::KnightHead::opp_table;

void osl::eval::experimental::
KnightHead::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
KnightHeadOppPiecePawnOnStand::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      KnightHead::opp_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

MultiInt osl::eval::experimental::
KnightHead::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<KNIGHT>::indexMin;
       i < PtypeTraits<KNIGHT>::indexLimit;
       ++i)
  {
    const Piece knight = state.getPieceOf(i);
    if (knight.isOnBoard() && !knight.isPromoted())
    {
      const Position up = Board_Table.nextPosition(knight.owner(),
						   knight.position(), U);
      const Piece up_piece = state.getPieceAt(up);
      if ((up_piece.isEmpty() && state.hasPieceOnStand<PAWN>(knight.owner()) &&
	   !state.isPawnMaskSet(knight.owner(), knight.position().x()) &&
	   state.countEffect(knight.owner(), up) >=
	   state.countEffect(alt(knight.owner()), up)) ||
	  (state.hasEffectByPtypeStrict<PAWN>(knight.owner(), up) &&
	   (up_piece.isEmpty() || up_piece.owner() != knight.owner()) &&
	   state.countEffect(knight.owner(), up) >
	   state.countEffect(alt(knight.owner()), up)))
      {
	const int y = knight.position().y();
	if (knight.owner() == BLACK)
	{
	  result += table[y - 1];
	}
	else
	{
	  result -= table[9 - y];
	}
      }
      else if (up_piece.isPiece() && up_piece.owner() != knight.owner() &&
	       state.hasPieceOnStand<PAWN>(up_piece.owner()))
      {
	const int y = (knight.owner() == BLACK ? knight.position().y() :
		       10 - knight.position().y());
	const int index = up_piece.ptype() * 9 + y - 1;
	if (knight.owner() == BLACK)
	{
	  result += opp_table[index];
	}
	else
	{
	  result -= opp_table[index];
	}
      }
    }
  }
  return result;
}


osl::misc::CArray<MultiInt, 1024> osl::eval::experimental::NonPawnAttackedPtype::table;

void osl::eval::experimental::
NonPawnAttackedPtype::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
NonPawnAttackedPtype::eval(const NumEffectState &state,
			   CArray<PieceMask, 40> &attacked_mask,
			   MultiIntPair &result)
{
  result = MultiIntPair();
  PieceMask black_attacked = state.effectedMask(WHITE) & state.getOnBoardMask(BLACK);
  black_attacked.reset(KingTraits<BLACK>::index);
  mask_t black_ppawn = state.promotedPieces().getMask<PAWN>() & black_attacked.selectBit<PAWN>();
  black_attacked.clearBit<PAWN>();
  black_attacked.orMask(PtypeFuns<PAWN>::indexNum, black_ppawn);
  while (black_attacked.any())
  {
    const Piece piece = state.getPieceOf(black_attacked.takeOneBit());
    const bool with_support = state.effectedMask(BLACK).test(piece.number());
    PieceMask attacking =
      state.getEffect(piece.position()) & state.getOnBoardMask(WHITE);
    attacked_mask[piece.number()] = attacking;
    
    while (attacking.any())
    {
      const Piece attack = state.getPieceOf(attacking.takeOneBit());
      const int index_black_turn = index(true, with_support,
					 piece.ptype(), attack.ptype());
      const int index_white_turn = index(false, with_support,
					 piece.ptype(), attack.ptype());
      adjust<1>(index_black_turn, index_white_turn, result);
    }
  }
  PieceMask white_attacked = state.effectedMask(BLACK) & state.getOnBoardMask(WHITE);
  white_attacked.reset(KingTraits<WHITE>::index);
  mask_t white_ppawn = state.promotedPieces().getMask<PAWN>() & white_attacked.selectBit<PAWN>();
  white_attacked.clearBit<PAWN>();
  white_attacked.orMask(PtypeFuns<PAWN>::indexNum, white_ppawn);
  while (white_attacked.any())
  {
    const Piece piece = state.getPieceOf(white_attacked.takeOneBit());
    const bool with_support = state.effectedMask(WHITE).test(piece.number());
    PieceMask attacking =
      state.getEffect(piece.position()) & state.getOnBoardMask(BLACK);
    attacked_mask[piece.number()] = attacking;
    while (attacking.any())
    {
      const Piece attack = state.getPieceOf(attacking.takeOneBit());
      const int index_black_turn = index(false, with_support,
					 piece.ptype(), attack.ptype());
      const int index_white_turn = index(true, with_support,
					 piece.ptype(), attack.ptype());
      adjust<-1>(index_black_turn, index_white_turn, result);
    }
  }
}

template<osl::Player P>
void osl::eval::experimental::
NonPawnAttackedPtype::evalWithUpdateBang(
  const NumEffectState &state,
  Move moved,
  const CArray<PieceMask, 2> &effected,
  CArray<PieceMask, 40> &attacked_mask,
  MultiIntPair &result)
{
  CArray<PieceMask, 2> effected_mask = effected;
  effected_mask[0].clearBit<KING>();
  effected_mask[1].clearBit<KING>();
  CArray<PieceMask, 2> new_mask = {{
      state.effectedMask(BLACK),
      state.effectedMask(WHITE)
    }};
  mask_t black_ppawn =
    new_mask[0].selectBit<PAWN>() & state.promotedPieces().getMask<PAWN>();
  mask_t white_ppawn =
    new_mask[1].selectBit<PAWN>() & state.promotedPieces().getMask<PAWN>();
  new_mask[0].clearBit<PAWN>();
  new_mask[1].clearBit<PAWN>();
  new_mask[0].orMask(PtypeFuns<PAWN>::indexNum, black_ppawn);
  new_mask[1].orMask(PtypeFuns<PAWN>::indexNum, white_ppawn);
  new_mask[0].clearBit<KING>();
  new_mask[1].clearBit<KING>();
  const Piece p = state.getPieceAt(moved.to());
  assert(p.owner()==P);
  assert(moved.player()==P);
  const Ptype captured = moved.capturePtype();
  int captured_number = -1;
  if (captured != PTYPE_EMPTY && captured != PAWN)
  {
    PieceMask captured_mask =
      effected_mask[P] & (~state.getOnBoardMask(BLACK)) &
      (~state.getOnBoardMask(WHITE));
    captured_number = captured_mask.takeOneBit();
  }
  if (!moved.isDrop() && moved.oldPtype() != PAWN)
  {
    if (effected_mask[alt(P)].test(p.number()))
    {
      const bool has_support = effected_mask[P].test(p.number());
      PieceMask attacking = attacked_mask[p.number()];
      if (captured_number != -1)
      {
	if (attacking.test(captured_number))
	{
	  if (P == BLACK)
	  {
	    evalOnePiece<false>(P, moved.oldPtype(), captured,
				has_support, result);
	  }
	  else
	  {
	    evalOnePiece<true>(P, moved.oldPtype(), captured,
			       has_support, result);
	  }
	  attacking.reset(captured_number);
	}
      }
      while (attacking.any())
      {
	const Piece attack = state.getPieceOf(attacking.takeOneBit());
	if (P == BLACK)
	{
	  evalOnePiece<false>(P, moved.oldPtype(), attack.ptype(),
			      has_support, result);
	}
	else
	{
	  evalOnePiece<true>(P, moved.oldPtype(), attack.ptype(),
			      has_support, result);
	}
      }
    }
  }
  if (new_mask[alt(P)].test(p.number()))
  {
    const bool has_support = new_mask[P].test(p.number());
    PieceMask attacking =
      state.getEffect(moved.to()) & state.getOnBoardMask(alt(P));
    attacked_mask[p.number()] = attacking;
    while (attacking.any())
    {
      const Piece attack = state.getPieceOf(attacking.takeOneBit());
      if (P == BLACK)
      {
	evalOnePiece<true>(P, p.ptype(), attack.ptype(),
			   has_support, result);
      }
      else
      {
	evalOnePiece<false>(P, p.ptype(), attack.ptype(),
			    has_support, result);
      }
    }
  }
  if (captured_number != -1)
  {
    const bool has_support = effected_mask[alt(P)].test(captured_number);
    PieceMask attacking = attacked_mask[captured_number];
    if (attacking.test(p.number()))
    {
      if (P == BLACK)
      {
	evalOnePiece<true>(alt(P), captured, moved.oldPtype(),
			   has_support, result);
      }
      else
      {
	evalOnePiece<false>(alt(P), captured, moved.oldPtype(),
			   has_support, result);
      }
      attacking.reset(p.number());
    }
    while (attacking.any())
    {
      const Piece attack = state.getPieceOf(attacking.takeOneBit());
      if (P == BLACK)
      {
	evalOnePiece<true>(alt(P), captured, attack.ptype(),
			   has_support, result);
      }
      else
      {
	evalOnePiece<false>(alt(P), captured, attack.ptype(),
			   has_support, result);
      }
    }
  }
  updateChanged<BLACK>(state, p, moved, captured_number,
		       effected_mask, new_mask, attacked_mask, result);
  updateChanged<WHITE>(state, p, moved, captured_number,
		       effected_mask, new_mask, attacked_mask, result);
}

osl::misc::CArray<MultiInt, 160>
osl::eval::experimental::PtypeCount::table;
osl::misc::CArray<MultiInt, 2240>
osl::eval::experimental::PtypeCount::xy_table;
osl::misc::CArray<MultiInt, 2240>
osl::eval::experimental::PtypeCount::xy_attack_table;
osl::misc::CArray<MultiInt, 2240>
osl::eval::experimental::PtypeCount::xy_table_diff;
osl::misc::CArray<MultiInt, 2240>
osl::eval::experimental::PtypeCount::xy_attack_table_diff;
void osl::eval::experimental::PtypeCount::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::PtypeCountXY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PtypeCount::xy_table[i][s] = weights.value(i + ONE_DIM*s);
  }
  for(int i=PTYPE_BASIC_MIN;i<=PTYPE_MAX;i++){
    Ptype ptype=static_cast<Ptype>(i);
    int indexMin=Ptype_Table.getIndexMin(ptype);
    int size=Ptype_Table.getIndexLimit(ptype)-indexMin;
    for(int x=0;x<5;x++){
      for(int j=0;j<size;j++){
	for(int k=0;k<160;k+=40){
	  PtypeCount::xy_table[(indexMin+j+k)*5+x]+=PtypeCount::table[indexMin+j+k];
	}
      }
    }
  }
  for(int i=PTYPE_BASIC_MIN;i<=PTYPE_MAX;i++){
    Ptype ptype=static_cast<Ptype>(i);
    int indexMin=Ptype_Table.getIndexMin(ptype);
    int size=Ptype_Table.getIndexLimit(ptype)-indexMin;
    for(int x=0;x<5;x++){
      for(int k=0;k<160;k+=40)
	PtypeCount::xy_table_diff[(indexMin+k)*5+x]=PtypeCount::xy_table[(indexMin+k)*5+x];
      for(int j=1;j<size;j++){
	for(int k=0;k<160;k+=40)
	  PtypeCount::xy_table_diff[(indexMin+k+j)*5+x]=PtypeCount::xy_table[(indexMin+k+j)*5+x]-PtypeCount::xy_table[(indexMin+k+j-1)*5+x];
      }
    }
    for(int y=0;y<9;y++){
      for(int k=0;k<160;k+=40)
	PtypeCount::xy_table_diff[800+(indexMin+k)*9+y]=PtypeCount::xy_table[800+(indexMin+k)*9+y];
      for(int j=1;j<size;j++){
	for(int k=0;k<160;k+=40)
	  PtypeCount::xy_table_diff[800+(indexMin+k+j)*9+y]=PtypeCount::xy_table[800+(indexMin+k+j)*9+y]-PtypeCount::xy_table[800+(indexMin+k+j-1)*9+y];
      }
    }
  }
}

void osl::eval::experimental::PtypeCountXYAttack::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      PtypeCount::xy_attack_table[i][s] = weights.value(i + ONE_DIM*s);
  }
  for(int i=PTYPE_BASIC_MIN;i<=PTYPE_MAX;i++){
    Ptype ptype=static_cast<Ptype>(i);
    int indexMin=Ptype_Table.getIndexMin(ptype);
    int size=Ptype_Table.getIndexLimit(ptype)-indexMin;
    for(int x=0;x<5;x++){
      for(int k=0;k<160;k+=40)
	PtypeCount::xy_attack_table_diff[(indexMin+k)*5+x]=PtypeCount::xy_attack_table[(indexMin+k)*5+x];
      for(int j=1;j<size;j++){
	for(int k=0;k<160;k+=40)
	  PtypeCount::xy_attack_table_diff[(indexMin+k+j)*5+x]=PtypeCount::xy_attack_table[(indexMin+k+j)*5+x]-PtypeCount::xy_attack_table[(indexMin+k+j-1)*5+x];
      }
    }
    for(int y=0;y<9;y++){
      for(int k=0;k<160;k+=40)
	PtypeCount::xy_attack_table_diff[800+(indexMin+k)*9+y]=PtypeCount::xy_attack_table[800+(indexMin+k)*9+y];
      for(int j=1;j<size;j++){
	for(int k=0;k<160;k+=40)
	  PtypeCount::xy_attack_table_diff[800+(indexMin+k+j)*9+y]=PtypeCount::xy_attack_table[800+(indexMin+k+j)*9+y]-PtypeCount::xy_attack_table[800+(indexMin+k+j-1)*9+y];
      }
    }
  }
}

template<osl::Player P,osl::Ptype T>
MultiInt osl::eval::experimental::PtypeCount::
evalPlayerPtype(const osl::CArray2d<int, 2, osl::PTYPE_SIZE> &ptype_count,
		const osl::CArray2d<int, 2, osl::PTYPE_SIZE> &ptype_board_count,
		const osl::CArray<int,2> &kings_x,
		const osl::CArray<int,2> &kings_y)
{
  MultiInt out;
  int i=PlayerTraits<P>::index;
  int j=static_cast<int>(T);
  if (ptype_count[i][j] != 0)
  {
    const int index_x = indexCountX<T>(ptype_count[i][j], kings_x[i]);
    const int index_y = indexCountY<T>(ptype_count[i][j], kings_y[i]);
    const int index_x_attack =
      indexCountX<T>(ptype_count[i][j], kings_x[1-i]);
    const int index_y_attack =
      indexCountY<T>(ptype_count[i][j], kings_y[1-i]);
    if (P == BLACK)
    {
      out += xy_table[index_x] + xy_table[index_y];
      out += xy_attack_table[index_x_attack] +
	xy_attack_table[index_y_attack];
    }
    else
    {
      out -= (xy_table[index_x] + xy_table[index_y]);
      out -= (xy_attack_table[index_x_attack] +
	      xy_attack_table[index_y_attack]);
    }
    if (ptype_board_count[i][j] != 0)
    {
      const int index_x =
	indexBoardCountX<T>(ptype_board_count[i][j], kings_x[i]);
      const int index_y =
	indexBoardCountY<T>(ptype_board_count[i][j], kings_y[i]);
      const int index_x_attack =
	indexBoardCountX<T>(ptype_board_count[i][j], kings_x[(i + 1) & 1]);
      const int index_y_attack =
	indexBoardCountY<T>(ptype_board_count[i][j], kings_y[(i + 1) & 1]);
      if (P == BLACK)
      {
	out += xy_table[index_x] + xy_table[index_y];
	out += xy_attack_table[index_x_attack] +
	  xy_attack_table[index_y_attack];
      }
      else
      {
	out -= (xy_table[index_x] + xy_table[index_y]);
	out -= (xy_attack_table[index_x_attack] +
		xy_attack_table[index_y_attack]);
      }
    }
  }
  return out;
}

void osl::eval::experimental::PtypeCount::eval(
  const NumEffectState &state,
  const CArray2d<int, 2, PTYPE_SIZE> &ptype_count,
  const CArray2d<int, 2, PTYPE_SIZE> &ptype_board_count,
  MultiInt &out)
{
  out.clear();
  CArray<int, 2> kings_x = {{ state.getKingPosition<BLACK>().x(),
			      state.getKingPosition<WHITE>().x() }};
  CArray<int, 2> kings_y = {{ state.getKingPosition<BLACK>().y(),
			      10 - state.getKingPosition<WHITE>().y() }};
  if (kings_x[0] > 5)
    kings_x[0] = 10 - kings_x[0];
  if (kings_x[1] > 5)
    kings_x[1] = 10 - kings_x[1];
  out = 
    evalPlayerPtype<BLACK,PPAWN>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PLANCE>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PKNIGHT>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PSILVER>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PBISHOP>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PROOK>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,GOLD>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,PAWN>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,LANCE>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,KNIGHT>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,SILVER>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,BISHOP>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<BLACK,ROOK>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PPAWN>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PLANCE>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PKNIGHT>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PSILVER>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PBISHOP>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PROOK>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,GOLD>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,PAWN>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,LANCE>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,KNIGHT>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,SILVER>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,BISHOP>(ptype_count,ptype_board_count,kings_x,kings_y)+
    evalPlayerPtype<WHITE,ROOK>(ptype_count,ptype_board_count,kings_x,kings_y);
}

template<osl::Player P>
void osl::eval::experimental::PtypeCount::evalWithUpdateBang(
  const NumEffectState &state,
  Move last_move,
  CArray2d<int, 2, PTYPE_SIZE> &ptype_count,
  CArray2d<int, 2, PTYPE_SIZE> &ptype_board_count,
  MultiInt &last_value_and_out,
  unsigned int &ptypeo_mask)
{
  assert(last_move.player()==P);
  const Player altP=PlayerTraits<P>::opponent;
  CArray<int, 2> kings_x = {{ state.getKingPosition<BLACK>().x(),
			      state.getKingPosition<WHITE>().x() }};
  CArray<int, 2> kings_y = {{ state.getKingPosition<BLACK>().y(),
			      10 - state.getKingPosition<WHITE>().y() }};
  if (kings_x[0] > 5)
    kings_x[0] = 10 - kings_x[0];
  if (kings_x[1] > 5)
    kings_x[1] = 10 - kings_x[1];

  if (last_move.ptype() == KING)
  {
    const Ptype capturedPtype = last_move.capturePtype();
    if (capturedPtype != PTYPE_EMPTY)
    {
      const PtypeO capturedPtypeO = last_move.capturePtypeO();
      if(--ptype_count[altP][capturedPtype]==0)
	ptypeo_mask &= ~(1<<(last_move.capturePtypeO()-PTYPEO_MIN));
      --ptype_board_count[altP][capturedPtype];
      const Ptype base_captured = unpromote(capturedPtype);
      ++ptype_count[P][base_captured];
      ptypeo_mask |= (1<<(captured(capturedPtypeO)-PTYPEO_MIN));
    }
    eval(state, ptype_count, ptype_board_count, last_value_and_out);
    return;
  }
  
  MultiInt sum;
  if (last_move.isDrop())
  {
    const int count = ++ptype_board_count[P][last_move.ptype()];
    sum = valueBoardAll(last_move.ptype(),count,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP]);
  }
  else{
    Ptype capturedPtype = last_move.capturePtype();
    if (capturedPtype != PTYPE_EMPTY)
    {
      const int count = --ptype_count[altP][capturedPtype];
      if(count==0)
	ptypeo_mask &= ~(1<<(last_move.capturePtypeO()-PTYPEO_MIN));
      const int board_count = --ptype_board_count[altP][capturedPtype];
      const Ptype base_captured = unpromote(capturedPtype);
      const int c_count = ++ptype_count[P][base_captured];
      ptypeo_mask |= 1<<(captured(last_move.capturePtypeO())-PTYPEO_MIN);
      sum=valueAll(capturedPtype,count+1,kings_x[altP],kings_y[altP],kings_x[P],kings_y[P])+
	valueBoardAll(capturedPtype,board_count+1,kings_x[altP],kings_y[altP],kings_x[P],kings_y[P])+
	valueAll(base_captured,c_count,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP]);
    }
    if (last_move.isPromote())
    {
      const Ptype old_ptype = last_move.oldPtype();
      const Ptype new_ptype = last_move.ptype();
      const int base_count = --ptype_count[P][old_ptype];
      const int base_board_count = --ptype_board_count[P][old_ptype];
      const int count = ++ptype_count[P][new_ptype];
      const int board_count = ++ptype_board_count[P][new_ptype];
      if(base_count==0)
	ptypeo_mask &= ~(1<<(last_move.oldPtypeO()-PTYPEO_MIN));
      ptypeo_mask |= (1<<(last_move.ptypeO()-PTYPEO_MIN));
      sum+=valueAll(new_ptype,count,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP])+
	valueBoardAll(new_ptype,board_count,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP])-
	valueAll(old_ptype,base_count+1,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP])-
	valueBoardAll(old_ptype,base_board_count+1,kings_x[P],kings_y[P],kings_x[altP],kings_y[altP]);
    }
  }
  if(P==BLACK) last_value_and_out+= sum;
  else last_value_and_out-= sum;
}

void osl::eval::experimental::
LanceEffectPieceKingRelative::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

osl::misc::CArray<MultiInt, 9792>
osl::eval::experimental::LanceEffectPieceKingRelative::table;

MultiInt osl::eval::experimental::
LanceEffectPieceKingRelative::eval(const NumEffectState &state)
{
  MultiInt result;
  for (int i = PtypeTraits<LANCE>::indexMin;
       i < PtypeTraits<LANCE>::indexLimit;
       ++i)
  {
    const Piece lance = state.getPieceOf(i);
    if (lance.isOnBoard() && !lance.isPromoted())
    {
      const Position self_king = state.getKingPosition(lance.owner());
      const Position opp_king = state.getKingPosition(alt(lance.owner()));
      Position p = state.getMobility(lance.owner() == BLACK ? U : D,
				     lance.number());
      if (!p.isOnBoard())
      {
	const int index1 = 0 + 0 + (PTYPEO_EDGE - PTYPEO_MIN) * 17 * 9;
	const int index2 = 0 + 0 + (PTYPEO_EDGE - PTYPEO_MIN) * 17 * 9 + 4896;
	if (lance.owner() == BLACK)
	{
	  result += table[index1];
	  result += table[index2];
	}
	else
	{
	  result -= table[index1];
	  result -= table[index2];
	}
      }
      else
      {
	const int index1 = index(lance.owner(), p, opp_king,
				 state.getPieceAt(p).ptypeO(), true);
	const int index2 = index(lance.owner(), p, self_king,
				 state.getPieceAt(p).ptypeO(), false);
	if (lance.owner() == BLACK)
	{
	  result += table[index1];
	  result += table[index2];
	}
	else
	{
	  result -= table[index1];
	  result -= table[index2];
	}
      }
    }
  }
  return result;
}

osl::misc::CArray<MultiInt, 1440>
osl::eval::experimental::PtypeYPawnY::table;

void osl::eval::experimental::PtypeYPawnY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s = 0; s < NStages; ++s)
    {
      table[i][s] = weights.value(i + ONE_DIM*s);
    }
  }
}

osl::MultiInt osl::eval::experimental::
PtypeYPawnY::eval(const NumEffectState &state,
		  const CArray2d<int, 2, 9> &pawns)
{
  MultiInt result;
  for (int i = 0; i < Piece::SIZE; ++i)
  {
    const Piece piece = state.getPieceOf(i);
    // only skip pawn, not ppawns
    if (piece.ptype() == PAWN)
      continue;
    if (!piece.isOnBoard())
      continue;

    const int idx = index(piece.owner(), piece.ptype(), piece.position().y(),
			  pawns[piece.owner()][piece.position().x() - 1]);
    if (piece.owner() == BLACK)
    {
      result += table[idx];
    }
    else
    {
      result -= table[idx];
    }
  }

  return result;
}

template<osl::Player P>
void osl::eval::experimental::
PtypeYPawnY::evalWithUpdateBang(const NumEffectState &state,
				Move moved,
				const CArray2d<int, 2, 9> &pawns,
				MultiInt& last_value)
{
  Ptype captured = moved.capturePtype();
  assert(P==moved.player());

  if (moved.oldPtype() == PAWN)
  {
    const int x = moved.to().x();
    const int old_pawn_y = (moved.isDrop() ? 0 : moved.from().y());
    const int new_pawn_y = pawns[P][moved.to().x() - 1];
    for (int y = 1; y <= 9; ++y)
    {
      const Piece p = state.getPieceAt(Position(x, y));
      if (y == moved.to().y())
      {
	if (p.ptype() == PPAWN)
	{
	  const int idx_new = index(P, p.ptype(), y, new_pawn_y);
	  if (P == BLACK)
	  {
	    last_value += table[idx_new];
	  }
	  else
	  {
	    last_value -= table[idx_new];
	  }   
	}
      }
      else if (!p.isEmpty() && p.owner() == P)
      {
	const int idx_old = index(P, p.ptype(), y, old_pawn_y);
	const int idx_new = index(P, p.ptype(), y, new_pawn_y);
	if (P == BLACK)
	{
	  last_value -= table[idx_old];
	  last_value += table[idx_new];
	}
	else
	{
	  last_value += table[idx_old];
	  last_value -= table[idx_new];
	}   
      }
    }
  }
  else
  {
    if (!moved.isDrop())
    {
      const int pawn_y = pawns[P][moved.from().x() - 1];
      const int idx = index(P, moved.oldPtype(), moved.from().y(),
			    pawn_y);
      if (P == BLACK)
      {
	last_value -= table[idx];
      }
      else
      {
	last_value += table[idx];
      }
    }
    {
      const int pawn_y = pawns[P][moved.to().x() - 1];
      const int idx = index(P, moved.ptype(), moved.to().y(),
			    pawn_y);
      if (P == BLACK)
      {
	last_value += table[idx];
      }
      else
      {
	last_value -= table[idx];
      }
    }
  }

  if (captured != PTYPE_EMPTY)
  {
    if (captured == PAWN)
    {
      const int old_pawn_y = moved.to().y();
      const int new_pawn_y = 0;
      const int x = moved.to().x();
      for (int y = 1; y <= 9; ++y)
      {
	const Piece p = state.getPieceAt(Position(x, y));
	if (!p.isEmpty() && p.owner() == alt(P))
	{
	  const int idx_old = index(alt(P), p.ptype(), y,
				    old_pawn_y);
	  const int idx_new = index(alt(P), p.ptype(), y,
				    new_pawn_y);
	  if (P == BLACK)
	  {
	    last_value += table[idx_old];
	    last_value -= table[idx_new];
	  }
	  else
	  {
	    last_value -= table[idx_old];
	    last_value += table[idx_new];
	  }   
	}
      }
    }
    else
    {
      const int pawn_y = pawns[alt(P)][moved.to().x() - 1];
      const int idx = index(alt(P), captured, moved.to().y(),
			    pawn_y);
      if (P == BLACK)
      {
	last_value += table[idx];
      }
      else
      {
	last_value -= table[idx];
      }
    }
  }
}

osl::misc::CArray<osl::MultiInt, 1215>
osl::eval::experimental::GoldAndSilverNearKing::table;
osl::misc::CArray<osl::MultiInt, 9720>
osl::eval::experimental::GoldAndSilverNearKing::combination_table;

void osl::eval::experimental::
GoldAndSilverNearKing::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
GoldAndSilverNearKingCombination::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      GoldAndSilverNearKing::combination_table[i][s] =
	weights.value(i + ONE_DIM*s);
  }
}

template <osl::Player P>
osl::MultiInt osl::eval::experimental::
GoldAndSilverNearKing::evalOne(const NumEffectState &state,
			       const CArray2d<int, 2, 3> &gs_count)
{
  MultiInt result;
  int total = 0;
  const Position king = state.getKingPosition<P>();
  for (size_t i = 0; i < gs_count.size2(); ++i)
  {
    total += gs_count[P][i];
    if (total != 0)
    {
      result += table[index<P>(king, i, total)];
    }
  }
  result += combination_table[
    indexCombination<P>(king, gs_count[P][0],
			gs_count[P][1], gs_count[P][2])];
  return P == BLACK ? result : -result;
}

osl::MultiInt osl::eval::experimental::
GoldAndSilverNearKing::eval(const NumEffectState &state,
			    const CArray2d<int, 2, 3> &gs_count)
{
  return evalOne<BLACK>(state, gs_count) + evalOne<WHITE>(state, gs_count);
}


osl::misc::CArray<osl::MultiInt, 8192>
osl::eval::experimental::PtypeCombination::table;

void osl::eval::experimental::
PtypeCombination::setUp(const Weights &weights)
{
  static CArray<MultiInt, 8192> orig_table;
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s = 0; s < NStages; ++s)
    {
      orig_table[i][s] = weights.value(i + ONE_DIM*s);
    }
  }
  for(int i=0;i<8192;i++){
    int pawn=(i>>12)&1;
    int ppawn=(i>>6)&1;
    int lance=(i>>11)&1;
    int plance=(i>>5)&1;
    int knight=(i>>10)&1;
    int pknight=(i>>4)&1;
    int silver=(i>>9)&1;
    int psilver=(i>>3)&1;
    int bishop=(i>>8)&1;
    int pbishop=(i>>2)&1;
    int rook=(i>>7)&1;
    int prook=(i>>1)&1;
    int gold=(i>>0)&1;
    int newIndex=ppawn|(plance<<1)|(pknight<<2)|(psilver<<3)|(pbishop<<4)|
      (prook<<5)|(gold<<6)|(pawn<<7)|(lance<<8)|(knight<<9)|(silver<<10)|
      (bishop<<11)|(rook<<12);
    table[newIndex]=orig_table[i];
  }
}

osl::MultiInt osl::eval::experimental::
PtypeCombination::eval(unsigned int ptypeo_mask)
{
  return evalOne<BLACK>(ptypeo_mask) + evalOne<WHITE>(ptypeo_mask);
}


namespace osl
{
  namespace eval
  {
    namespace experimental
    {
      template void PawnAdvanceAll::
      evalWithUpdateBang<BLACK>(const NumEffectState &, Move,MultiInt&);
      template void PawnAdvanceAll::
      evalWithUpdateBang<WHITE>(const NumEffectState &, Move,MultiInt&);
      template MultiInt PtypeY::
      evalWithUpdate<BLACK>(const NumEffectState &, Move, MultiInt const&);
      template MultiInt PtypeY::
      evalWithUpdate<WHITE>(const NumEffectState &, Move, MultiInt const&);
      template MultiInt PtypeX::
      evalWithUpdate<BLACK>(const NumEffectState &, Move, MultiInt const&);
      template MultiInt PtypeX::
      evalWithUpdate<WHITE>(const NumEffectState &, Move, MultiInt const&);
      template MultiInt PawnPtypeOPtypeO::
      evalWithUpdate<BLACK>(const NumEffectState &, Move, const CArray2d<int, 2, 9> &, MultiInt const&);
      template MultiInt PawnPtypeOPtypeO::
      evalWithUpdate<WHITE>(const NumEffectState &, Move, const CArray2d<int, 2, 9> &, MultiInt const&);

      template void osl::eval::experimental::NonPawnAttacked::
      evalWithUpdateBang<BLACK>(const NumEffectState &state,
				Move moved,
				const CArray<PieceMask, 2> &effected,
				MultiIntPair &result);
      template void osl::eval::experimental::NonPawnAttacked::
      evalWithUpdateBang<WHITE>(const NumEffectState &state,
				Move moved,
				const CArray<PieceMask, 2> &effected,
				MultiIntPair &result);
      template void osl::eval::experimental::NonPawnAttackedPtype::
      evalWithUpdateBang<BLACK>(
	const NumEffectState &state,
	Move moved,
	const CArray<PieceMask, 2> &effected,
	CArray<PieceMask, 40> &attacked_mask,
	MultiIntPair &result);
      template void osl::eval::experimental::NonPawnAttackedPtype::
      evalWithUpdateBang<WHITE>(
	const NumEffectState &state,
	Move moved,
	const CArray<PieceMask, 2> &effected,
	CArray<PieceMask, 40> &attacked_mask,
	MultiIntPair &result);
      template void osl::eval::experimental::PtypeYPawnY::
      evalWithUpdateBang<BLACK>(const NumEffectState &state,
				Move moved,
				const CArray2d<int, 2, 9> &pawns,
				MultiInt& last_value);
      template void osl::eval::experimental::PtypeYPawnY::
      evalWithUpdateBang<WHITE>(const NumEffectState &state,
				Move moved,
				const CArray2d<int, 2, 9> &pawns,
				MultiInt& last_value);
      template void PtypeCount::
      evalWithUpdateBang<BLACK>(const NumEffectState &state,Move last_move,
				CArray2d<int, 2, PTYPE_SIZE> &ptype_count,
				CArray2d<int, 2, PTYPE_SIZE> &ptype_board_count,
				MultiInt &last_value_and_out,
				unsigned int &ptypeo_mask);
      template void PtypeCount::
      evalWithUpdateBang<WHITE>(const NumEffectState &state,Move last_move,
				CArray2d<int, 2, PTYPE_SIZE> &ptype_count,
				CArray2d<int, 2, PTYPE_SIZE> &ptype_board_count,
				MultiInt &last_value_and_out,
				unsigned int &ptypeo_mask);
#ifndef MINIMAL
      template class PawnAndEmptyAndPiece<true>;
      template class PawnAndEmptyAndPiece<false>;
#endif
    }
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
