#include "osl/pieceStand.h"
#include "osl/eval/experimental/pieceStand.h"

osl::misc::CArray<int, osl::PTYPE_SIZE>
osl::eval::experimental::SimplePieceStand::table;

void osl::eval::experimental::
SimplePieceStand::setUp(const Weights &weights)
{
  table.fill(0);
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[osl::PieceStand::order[i]] = weights.value(i);
  }
}

int osl::eval::experimental::
SimplePieceStand::eval(const NumEffectState &state) const
{
  int value = 0;
  for (size_t i = 0; i < osl::PieceStand::order.size(); ++i)
  {
    value += table[osl::PieceStand::order[i]] *
      (state.countPiecesOnStand(osl::BLACK, osl::PieceStand::order[i]) -
       state.countPiecesOnStand(osl::WHITE, osl::PieceStand::order[i]));
  }
  return value;
}

int osl::eval::experimental::
SimplePieceStand::evalWithUpdate(const NumEffectState &,
				 Move moved, int last_value) const
{
  if (moved.isPass())
    return last_value;
  int value = last_value;
  osl::Ptype captured = moved.capturePtype();
  if (moved.isDrop())
  {
    if (moved.player() == BLACK)
      value -= table[moved.ptype()];
    else
      value += table[moved.ptype()];
  }
  else if (captured != PTYPE_EMPTY)
  {
    if (moved.player() == BLACK)
      value += table[unpromote(captured)];
    else
      value -= table[unpromote(captured)];
  }
  return value;
}


osl::misc::CArray<osl::MultiInt, osl::Piece::SIZE>
osl::eval::experimental::PieceStand::table;

void osl::eval::experimental::
PieceStand::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i][stage] = weights.value(i);
  }
}

osl::MultiInt osl::eval::experimental::PieceStand::eval(
  const osl::state::NumEffectState &state)
{
  osl::MultiInt result;
  for (size_t i = 0; i < osl::PieceStand::order.size(); ++i)
  {
    const Ptype ptype = osl::PieceStand::order[i];
    const int black_count =
      state.countPiecesOnStand(osl::BLACK, osl::PieceStand::order[i]);
    const int white_count =
      state.countPiecesOnStand(osl::WHITE, osl::PieceStand::order[i]);
    for (int j = 0; j < black_count; ++j)
    {
      result += table[Ptype_Table.getIndexMin(ptype) + j];
    }
    for (int j = 0; j < white_count; ++j)
    {
      result -= table[Ptype_Table.getIndexMin(ptype) + j];
    }
  }
  return result;
}



osl::misc::CArray<osl::MultiInt, 21>
osl::eval::experimental::NonPawnPieceStand::table;

void osl::eval::experimental::
NonPawnPieceStand::setUp(const Weights &weights,int stage)
{
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i][stage] = weights.value(i);
  }
}

osl::MultiInt osl::eval::experimental::
NonPawnPieceStand::eval(int black_count, int white_count)
{
  return table[black_count] - table[white_count];
}


osl::misc::CArray<osl::MultiInt, 5625> osl::eval::experimental::NonPawnPieceStandCombination::table;
osl::misc::CArray<osl::MultiInt, 5625> osl::eval::experimental::NonPawnPieceStandCombination::check_table;

void osl::eval::experimental::
NonPawnPieceStandCombination::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
CanCheckNonPawnPieceStandCombination::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      NonPawnPieceStandCombination::check_table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

osl::MultiInt osl::eval::experimental::
NonPawnPieceStandCombination::eval(const NumEffectState &state,
				   const CArray<bool, 2> &can_check)
{
  const int black_index = index(state.countPiecesOnStand<ROOK>(BLACK),
				state.countPiecesOnStand<BISHOP>(BLACK),
				state.countPiecesOnStand<GOLD>(BLACK),
				state.countPiecesOnStand<SILVER>(BLACK),
				state.countPiecesOnStand<KNIGHT>(BLACK),
				state.countPiecesOnStand<LANCE>(BLACK));
  const int white_index = index(state.countPiecesOnStand<ROOK>(WHITE),
				state.countPiecesOnStand<BISHOP>(WHITE),
				state.countPiecesOnStand<GOLD>(WHITE),
				state.countPiecesOnStand<SILVER>(WHITE),
				state.countPiecesOnStand<KNIGHT>(WHITE),
				state.countPiecesOnStand<LANCE>(WHITE));
  MultiInt result;
  result = table[black_index] - table[white_index];
  if (can_check[PlayerTraits<WHITE>::index])
  {
    result += check_table[black_index];
  }
  if (can_check[PlayerTraits<BLACK>::index])
  {
    result -= check_table[white_index];
  }
  return result;
}

osl::MultiInt osl::eval::experimental::
NonPawnPieceStandCombination::evalWithUpdate(
  const NumEffectState &state,
  Move moved,
  const MultiInt &last_value,
  const CArray<bool, 2> &could_check,
  const CArray<bool, 2> &can_check)
{
  if (!moved.isDrop() && moved.capturePtype() == PTYPE_EMPTY &&
      could_check[0] == can_check[0] && could_check[1] == can_check[1])
  {
    return last_value;
  }
  return eval(state, can_check);
}


osl::misc::CArray<osl::MultiInt, 44> osl::eval::experimental::NonPawnPieceStandTurn::table;

void osl::eval::experimental::
NonPawnPieceStandTurn::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s)
      table[i][s] = weights.value(i + ONE_DIM*s);
  }
}

void osl::eval::experimental::
NonPawnPieceStandTurn::eval(const NumEffectState &state, MultiIntPair& result)
{
  result = MultiIntPair();
  for (size_t i = 0; i < osl::PieceStand::order.size(); ++i)
  {
    const Ptype ptype = osl::PieceStand::order[i];
    if (ptype == PAWN)
      continue;
    const int black_count =
      state.countPiecesOnStand(osl::BLACK, osl::PieceStand::order[i]);
    const int white_count =
      state.countPiecesOnStand(osl::WHITE, osl::PieceStand::order[i]);
    for (int j = 0; j < black_count; ++j)
    {
      const int index_black = index(BLACK, BLACK, ptype, j);
      const int index_white = index(BLACK, WHITE, ptype, j);
      result[BLACK] += table[index_black];
      result[WHITE] += table[index_white];
    }
    for (int j = 0; j < white_count; ++j)
    {
      const int index_black = index(WHITE, BLACK, ptype, j);
      const int index_white = index(WHITE, WHITE, ptype, j);
      result[BLACK] -= table[index_black];
      result[WHITE] -= table[index_white];
    }
  }
}

template<osl::Player P>
void osl::eval::experimental::
NonPawnPieceStandTurn::evalWithUpdateBang(
  const NumEffectState &state,
  Move moved, MultiIntPair &result)
{
  assert(P==moved.player());
  if (!moved.isDrop() && moved.capturePtype() == PTYPE_EMPTY)
    return;

  if (moved.isDrop())
  {
    const Ptype ptype = moved.ptype();
    if (ptype == PAWN)
      return;
    const int count =
      state.countPiecesOnStand(P, moved.ptype());
    const int index_black = index(P, BLACK, moved.ptype(), count);
    const int index_white = index(P, WHITE, moved.ptype(), count);
    if(P==BLACK){
      result[BLACK] -= table[index_black];
      result[WHITE] -= table[index_white];
    }
    else{
      result[BLACK] += table[index_black];
      result[WHITE] += table[index_white];
    }
  }
  if (moved.capturePtype() != PTYPE_EMPTY &&
      unpromote(moved.capturePtype()) != PAWN)
  {
    Ptype ptype = unpromote(moved.capturePtype());
    const int count = state.countPiecesOnStand(P, ptype) - 1;
    const int index_black = index(P, BLACK, ptype, count);
    const int index_white = index(P, WHITE, ptype, count);
    if(P==BLACK){
      result[BLACK] += table[index_black];
      result[WHITE] += table[index_white];
    }
    else{
      result[BLACK] -= table[index_black];
      result[WHITE] -= table[index_white];
    }
  }
}


osl::misc::CArray<osl::MultiInt, 360> osl::eval::experimental::PieceStandY::y_attack_table;
osl::misc::CArray<osl::MultiInt, 360> osl::eval::experimental::PieceStandY::y_defense_table;;
osl::misc::CArray<osl::MultiInt, 9*7*19> osl::eval::experimental::PieceStandY::y_attack_table_sum;
osl::misc::CArray<osl::MultiInt, 9*7*19> osl::eval::experimental::PieceStandY::y_defense_table_sum;

void osl::eval::experimental::
PieceStandY::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    for (int s=0; s<NStages; ++s) 
    {
      y_attack_table[i][s] = weights.value(i + ONE_DIM * 2 * s);
      y_defense_table[i][s] = weights.value(i + ONE_DIM * 2 * s + ONE_DIM);
    }
  }
  for (int i=0;i<7;i++){
    Ptype ptype=osl::PieceStand::order[i];
    int ptypeSize=Ptype_Table.getIndexLimit(ptype)-Ptype_Table.getIndexMin(ptype);
    for(int king_y=1;king_y<=9;king_y++){
      MultiInt attack_sum, defense_sum;
      for(int count=0;count<=ptypeSize;count++){
#if 0
	int oldIndex=(king_y - 1) * 40 + Ptype_Table.getIndexMin(ptype) + count;
	int newIndex=(king_y - 1) * 7*19 + i*19 + count;
#else
	int oldIndex=index(ptype,BLACK,Position(5,king_y),count);
	int newIndex=index(i,BLACK,Position(5,king_y),count);
#endif
	y_attack_table_sum[newIndex]=attack_sum;
	y_defense_table_sum[newIndex]=defense_sum;
	if(count==ptypeSize) break;
	attack_sum += y_attack_table[oldIndex];
	defense_sum += y_defense_table[oldIndex];
      }
    }
  }
}

inline 
void osl::eval::experimental::PieceStandY::updateResult(osl::state::NumEffectState const& state,osl::MultiInt &result,int i, osl::Ptype ptype, osl::misc::CArray<osl::Position,2> const&kings)
{
  const int black_count = state.countPiecesOnStand(BLACK, ptype);
  const int white_count = state.countPiecesOnStand(WHITE, ptype);
  const int attack_index_1 = PieceStandY::index(i, BLACK, kings[WHITE], black_count);
  const int attack_index_2 = PieceStandY::index(i, WHITE, kings[BLACK], white_count);
  const int defense_index_1 = PieceStandY::index(i, BLACK, kings[BLACK], black_count);
  const int defense_index_2 = PieceStandY::index(i, WHITE, kings[WHITE], white_count);
  result += y_attack_table_sum[attack_index_1] - y_attack_table_sum[attack_index_2] +
    y_defense_table_sum[defense_index_1] - y_defense_table_sum[defense_index_2];
}

osl::MultiInt osl::eval::experimental::
PieceStandY::eval(const NumEffectState &state)
{
  MultiInt result;
  const CArray<Position,2> kings = {{ 
      state.getKingPosition(BLACK),
      state.getKingPosition(WHITE),
    }};
  updateResult(state,result,0,ROOK,kings);
  updateResult(state,result,1,BISHOP,kings);
  updateResult(state,result,2,GOLD,kings);
  updateResult(state,result,3,SILVER,kings);
  updateResult(state,result,4,KNIGHT,kings);
  updateResult(state,result,5,LANCE,kings);
  updateResult(state,result,6,PAWN,kings);
  return result;
}

template<osl::Player P>
osl::MultiInt osl::eval::experimental::
PieceStandY::evalWithUpdate(
  const NumEffectState &state,
  Move moved, const MultiInt &last_value)
{
  if (moved.ptype() == KING)
    return eval(state);

  MultiInt result(last_value);
  if (moved.isDrop())
  {
    const Ptype ptype = moved.ptype();
    const int count =
      state.countPiecesOnStand(P, ptype);
    const int attack_index = index(ptype, P,
				   state.getKingPosition(alt(P)),
				   count);
    const int defense_index = index(ptype, P,
				    state.getKingPosition(P),
				    count); 
    if(P==BLACK)
      result -= y_attack_table[attack_index] +y_defense_table[defense_index];
    else
      result += y_attack_table[attack_index] +y_defense_table[defense_index];
  }
  if (moved.capturePtype() != PTYPE_EMPTY)
  {
    Ptype ptype = unpromote(moved.capturePtype());
    const int count = state.countPiecesOnStand(P, ptype)-1;
    const int attack_index = index(ptype, P,
				   state.getKingPosition(alt(P)),
				   count);
    const int defense_index = index(ptype, P,
				    state.getKingPosition(P),
				    count);
    if(P==BLACK)
      result += y_attack_table[attack_index] +y_defense_table[defense_index];
    else
      result -= y_attack_table[attack_index] +y_defense_table[defense_index];
  }
  return result;
}

osl::misc::CArray<osl::MultiInt, 16384> osl::eval::experimental::PieceStandCombinationBoth::table;

void osl::eval::experimental::
PieceStandCombinationBoth::setUp(const Weights &weights)
{
  for (size_t i = 0; i < ONE_DIM; ++i)
  {
    int low = (i & 0x7F);
    int high = (i >> 7);
    if (low == high)
      continue;
    for (int s = 0; s < NStages; ++s)
    {
      table[i][s] = weights.value(i + ONE_DIM*s);
      if (high > low)
      {
	table[(low << 7) | high][s] = -table[i][s];
      }
    }
  }
}

osl::MultiInt osl::eval::experimental::
PieceStandCombinationBoth::eval(const NumEffectState &state)
{
  int black_index = 0;
  int white_index = 0;
  black_index |= ((state.hasPieceOnStand<ROOK>(BLACK) ? 1 : 0) << 6);
  black_index |= ((state.hasPieceOnStand<BISHOP>(BLACK) ? 1 : 0) << 5);
  black_index |= ((state.hasPieceOnStand<GOLD>(BLACK) ? 1 : 0) << 4);
  black_index |= ((state.hasPieceOnStand<SILVER>(BLACK) ? 1 : 0) << 3);
  black_index |= ((state.hasPieceOnStand<KNIGHT>(BLACK) ? 1 : 0) << 2);
  black_index |= ((state.hasPieceOnStand<LANCE>(BLACK) ? 1 : 0) << 1);
  black_index |= ((state.hasPieceOnStand<PAWN>(BLACK) ? 1 : 0) << 0);
  white_index |= ((state.hasPieceOnStand<ROOK>(WHITE) ? 1 : 0) << 6);
  white_index |= ((state.hasPieceOnStand<BISHOP>(WHITE) ? 1 : 0) << 5);
  white_index |= ((state.hasPieceOnStand<GOLD>(WHITE) ? 1 : 0) << 4);
  white_index |= ((state.hasPieceOnStand<SILVER>(WHITE) ? 1 : 0) << 3);
  white_index |= ((state.hasPieceOnStand<KNIGHT>(WHITE) ? 1 : 0) << 2);
  white_index |= ((state.hasPieceOnStand<LANCE>(WHITE) ? 1 : 0) << 1);
  white_index |= ((state.hasPieceOnStand<PAWN>(WHITE) ? 1 : 0) << 0);
  return table[(black_index << 7) | white_index];
}


namespace osl
{
  namespace eval
  {
    namespace experimental
    {
      template void NonPawnPieceStandTurn::evalWithUpdateBang<BLACK>(const NumEffectState &, Move, MultiIntPair &);
      template void NonPawnPieceStandTurn::evalWithUpdateBang<WHITE>(const NumEffectState &, Move, MultiIntPair &);
      template MultiInt PieceStandY::evalWithUpdate<BLACK>(const NumEffectState &, Move, const MultiInt &);
      template MultiInt PieceStandY::evalWithUpdate<WHITE>(const NumEffectState &, Move, const MultiInt &);
    }
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
