#include "osl/eval/experimental/pin.h"
using osl::MultiInt;

osl::misc::CArray<int, osl::PTYPE_SIZE>
osl::eval::experimental::SimplePin::table;

osl::misc::CArray2d<MultiInt, osl::PTYPE_SIZE, 17 * 9>
osl::eval::experimental::Pin::table;

void osl::eval::experimental::
SimplePin::setUp(const Weights &weights)
{
  table.fill(0);
  for (size_t i = 0; i < weights.dimension(); ++i)
  {
    table[i] = weights.value(i);
  }
}

int osl::eval::experimental::
SimplePin::eval(const NumEffectState &state,
		PieceMask black_mask, PieceMask white_mask) const
{
  int value = 0;
  while (black_mask.any())
  {
    const osl::Piece piece = state.getPieceOf(black_mask.takeOneBit());
    value += table[piece.ptype()];
  }
  while (white_mask.any())
  {
    const osl::Piece piece = state.getPieceOf(white_mask.takeOneBit());
    value -= table[piece.ptype()];
  }
  return value;
}



void osl::eval::experimental::
Pin::setUp(const Weights &weights,int stage)
{
  for (int i = PTYPE_PIECE_MIN; i <= PTYPE_MAX; ++i)
  {
    for (int y = 0; y <= 16; ++y)
    {
      for (int x = 0; x <= 8; ++x)
      {
	const int distance = x * 17 + y;
	table[i][distance][stage] =
	  weights.value((i - PTYPE_PIECE_MIN) * 17 * 9 + distance);
      }
    }
  }
}

MultiInt osl::eval::experimental::
Pin::eval(const NumEffectState &state,
	  PieceMask black_mask, PieceMask white_mask)
{
  MultiInt value;
  const Position black_king = state.getKingPosition<BLACK>();
  const Position white_king = state.getKingPosition<WHITE>();
  while (black_mask.any())
  {
    const osl::Piece piece = state.getPieceOf(black_mask.takeOneBit());
    value += table[piece.ptype()][index(black_king, piece)];
  }
  while (white_mask.any())
  {
    const osl::Piece piece = state.getPieceOf(white_mask.takeOneBit());
    value -= table[piece.ptype()][index(white_king, piece)];
  }
  return value;
}

