/* numEffectState.cc
 */
#include "osl/state/numEffectState.h"
#include "osl/state/numEffectState.tcc"
#include "osl/effect/numSimpleEffect.tcc"

#include <iostream>

bool osl::state::operator==(const NumEffectState& st1,
			    const NumEffectState& st2)
{
  assert(st1.isConsistent(true));
  assert(st2.isConsistent(true));
  if (!(st1.effects == st2.effects)) 
    return false;
  if (!(st1.onBoardMask == st2.onBoardMask)) 
    return false;
  if (!(st1.promoted == st2.promoted)) 
    return false;
  if (!(st1.pin_or_open == st2.pin_or_open)) 
    return false;
  if (!(st1.king_mobility == st2.king_mobility)) 
    return false;
  return (static_cast<const SimpleState&>(st1)
	  == static_cast<const SimpleState&>(st2));
}

osl::state::
NumEffectState::NumEffectState(const SimpleState& st) 
  : SimpleState(st),effects(st)
{
  onBoardMask[0].resetAll();
  onBoardMask[1].resetAll();
  promoted.resetAll();
  effects.effected_mask[0].resetAll();
  effects.effected_mask[1].resetAll();
  effects.effected_changed_mask[0].resetAll();
  effects.effected_changed_mask[1].resetAll();
  for(int num=0;num<40;num++){
    Piece p=getPieceOf(num);
    if (p.isOnBoard()){
      getOnBoardMask(p.owner()).set(num);
      if (p.isPromoted())
	promoted.set(num);
      for(int i=0;i<2;i++){
	Player pl=indexToPlayer(i);
	if(hasEffectBy(pl,p.position()))
	{
	  effects.effected_mask[i].set(num);
	  effects.effected_changed_mask[i].set(num);
	}
      }
    }
  }
  makePinOpen(BLACK);
  makePinOpen(WHITE);
  valid_king8info[BLACK] = valid_king8info[WHITE] = false;
}
osl::state::
NumEffectState::~NumEffectState() 
{
}

const osl::checkmate::King8Info osl::state::
NumEffectState::king8Info(Player king) const
{
  if (! valid_king8info[king]) {
    king8info_cache[king] = King8Info::make(alt(king), *this).value;
    valid_king8info[king] = true;
  }
  return King8Info(king8info_cache[king]);
}
void osl::state::NumEffectState::
doSimpleMove(Position from, Position to, int promoteMask)
{
  Piece oldPiece;
  int num;
  PtypeO oldPtypeO, newPtypeO;
  CArray<PieceMask,2> pin_or_open_backup;
  KingMobility king_mobility_backup;
  PieceMask promoted_backup;
  CArray<PieceMask,2> effected_mask_backup;
  CArray<PieceMask,2> effected_changed_mask_backup;
  mobility::MobilityTable mobilityTable;
  if (getTurn()==BLACK){
    apply_move::ApplyDoUndoSimpleMove<BLACK,NumEffectState >
      ::prologue(*this, from, to, promoteMask, 
		 oldPiece, num, oldPtypeO, newPtypeO, 
		 pin_or_open_backup, king_mobility_backup,
		 promoted_backup, effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  else{
    apply_move::ApplyDoUndoSimpleMove<WHITE,NumEffectState >
      ::prologue(*this, from, to, promoteMask, 
		 oldPiece, num, oldPtypeO, newPtypeO, 
		 pin_or_open_backup, king_mobility_backup,
		 promoted_backup, effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  if (promoteMask!=0 && num < PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(),from);
}
void osl::state::NumEffectState::
doCaptureMove(Position from, Position to, Piece target, int promoteMask)
{
  Piece oldPiece;
  PtypeO oldPtypeO, capturePtypeO, newPtypeO;
  int num0, num1, num1Index;
  mask_t num1Mask;
  CArray<PieceMask,2> pin_or_open_backup;
  KingMobility king_mobility_backup;
  PieceMask promoted_backup;
  CArray<PieceMask,2> effected_mask_backup;
  CArray<PieceMask,2> effected_changed_mask_backup;
  mobility::MobilityTable mobilityTable;
  if(getTurn()==BLACK){
    apply_move::ApplyDoUndoCaptureMove<BLACK,NumEffectState >
      ::prologue(*this, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		 capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask, 
		 pin_or_open_backup, king_mobility_backup,
		 promoted_backup, effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  else{
    apply_move::ApplyDoUndoCaptureMove<WHITE,NumEffectState >
      ::prologue(*this, from, to, target, promoteMask, oldPiece, oldPtypeO, 
		 capturePtypeO, newPtypeO, num0, num1, num1Index,num1Mask, 
		 pin_or_open_backup, king_mobility_backup,
		 promoted_backup, effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  const Ptype capturePtype=target.ptype();
  if (capturePtype==PAWN)
    clearPawn(alt(getTurn()),to);
  if (promoteMask!=0 && num0<PtypeTraits<PAWN>::indexLimit)
    clearPawn(getTurn(),from);
}

void osl::state::NumEffectState::
doDropMove(Position to,Ptype ptype)
{
  Piece oldPiece;
  PtypeO ptypeO;
  int num, numIndex;
  mask_t numMask;
  CArray<PieceMask,2> pin_or_open_backup;
  KingMobility king_mobility_backup;
  CArray<PieceMask,2> effected_mask_backup;
  CArray<PieceMask,2> effected_changed_mask_backup;
  mobility::MobilityTable mobilityTable;
  if(getTurn()==BLACK){
    apply_move::ApplyDoUndoDropMove<BLACK,NumEffectState >
      ::prologue(*this, to, ptype, oldPiece, num, ptypeO, numIndex, numMask, 
		 pin_or_open_backup, king_mobility_backup,
		 effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  else{
    apply_move::ApplyDoUndoDropMove<WHITE,NumEffectState >
      ::prologue(*this, to, ptype, oldPiece, num, ptypeO, numIndex, numMask, 
		 pin_or_open_backup, king_mobility_backup,
		 effected_mask_backup, effected_changed_mask_backup,mobilityTable);
  }
  if (ptype==PAWN)
    setPawn(getTurn(),to);
}

#ifndef MINIMAL
bool osl::state::NumEffectState::isConsistent(bool showError) const
{
  if (!SimpleState::isConsistent(showError)) 
  {
    if (showError)
      std::cerr << "error before effect\n";
    return false;
  }
  effect::NumSimpleEffectTable effects1(*this);
  if (!(effects1==effects))
  {
    if (showError)
    {
      std::cerr << "Effect error 1" << std::endl;
      std::cerr << *this;
      for(int y=1;y<=9;y++)
	for(int x=9;x>0;x--)
	{
	  Position pos(x,y);
	  if (!(effects1.getEffect(pos)==effects.getEffect(pos)))
	  {
	    std::cerr << pos << ",real=" << effects.getEffect(pos) << ",ideal=" << effects1.getEffect(pos) << std::endl;
	  }
	}
      for(int num=0;num<=39;num++){
	for(int i=0;i<8;i++){
	  Direction d=static_cast<Direction>(i);
	  if(effects.effectedNumTable[num][d]!=effects1.effectedNumTable[num][d]){
	    std::cerr << "piece=" << getPieceOf(num) << ",num=" << num << ",d=" << d << ",v1=" << effects.effectedNumTable[num][d] << ",v2=" << effects1.effectedNumTable[num][d] << std::endl;
	  }
	}
      }
      std::cerr << effects.effectedNumTable << std::endl;
    }
    return false;
  }
  for (int z=0; z<2; ++z) {
    const Player p = indexToPlayer(z);
    const PieceMask pin2 = effect_util::Pin::make(*this, p);
    if (pin(p) != pin2) {
      if (showError)
	std::cerr << "pin for " << p << " differs " << pin(p) << " " << pin2 << "\n";
      return false;
    }
    King8Info king8info2 = King8Info::make(alt(p), *this);
    if (king8Info(p).value != king8info2.value) {
      if (showError)
	std::cerr << "king8info for " << p << " differs \n" << king8Info(p) << "\n" << king8info2 << "\n";
      return false;
    }      
  }
  for (int i=0; i<Piece::SIZE; ++i) {
    const Piece p = getPieceOf(i);
    if (p.isOnBoard()) {
      if (promoted.test(i) != p.isPromoted()) {
	if (showError)
	  std::cerr << "promoted differs " << p << " " << promoted << " " << promoted.test(i) << "\n";
	return false;
      }
    }
  }
  return true;
}
#endif

bool osl::state::NumEffectState::isConsistent(const NumEffectState& prev, Move moved, bool show_error) const
{
#ifdef MINIMAL
  show_error=false;
#endif
  // test changedEffects
  const CArray<BoardMask,2> changed_squares
    = {{ changedEffects(BLACK), changedEffects(WHITE) }};
  const BoardMask changed_all = changed_squares[BLACK] | changed_squares[WHITE];
  CArray<BoardMask, Piece::SIZE> each_effect, prev_effect;
  for (int i=0; i<Piece::SIZE; ++i) {
    each_effect[i].clear();
    prev_effect[i].clear();
  }
  for (int x=1; x<=9; ++x) {
    for (int y=1; y<=9; ++y) {
      const Position sq(x, y);
      for (int i=0; i<Piece::SIZE; ++i) {
	if (getEffect(sq).test(i))
	  each_effect[i].set(sq);
	if (prev.getEffect(sq).test(i))
	  prev_effect[i].set(sq);
      }
      if (! changed_all.test(sq))
      {
	if (getEffect(sq) != prev.getEffect(sq)) {
#ifndef MINIMAL
	  if (show_error)
	    std::cerr << "changedEffects unset\n" << *this << moved << sq << "\n";	  
#endif
	  return false;
	}
      }
      for (int i=0; i<2; ++i) 
      {
	const Player pl = indexToPlayer(i);
	if (! changed_squares[pl].test(sq))
	{
	  if ((getEffect(sq) & getOnBoardMask(pl))
	      != (prev.getEffect(sq) & prev.getOnBoardMask(pl))) {
#ifndef MINIMAL
	    if (show_error)
	      std::cerr << "changedEffects unset for " << pl << "\n" << *this << moved << sq << "\n";
#endif
	    return false;
	  }
	}
      }
    }
  }
  // test changedPieces()
  const NumBitmapEffect changed_effect_pieces = changedPieces(); 
  for (int i=0; i<Piece::SIZE; ++i) {
    if (each_effect[i] == prev_effect[i])
      continue;
    if (! changed_effect_pieces.test(i)) {
#ifndef MINIMAL
      if (show_error)
	std::cerr << "changedPieces() unset\n" << *this << moved << i 
		  << " " << each_effect[i] << " != " <<  prev_effect[i] << "\n";
#endif
      return false;
    }
  }
  // test effectedChanged(Player pl)
  for (int i=0; i<Piece::SIZE; ++i) 
  {
    for (int j=0; j<2; ++j) 
    {
      const Player pl = indexToPlayer(j);
      if (prev.getPieceOf(i).position() == moved.to())
	continue;		// captured
      if (prev.effectedMask(pl).test(i) != effectedMask(pl).test(i)) {
	if (! effectedChanged(pl).test(i)) {
#ifndef MINIMAL
	  if (show_error)
	    std::cerr << "effectedChanged(" << pl << ") unset\n" << *this << moved << i 
		      << " " << prev.effectedChanged(pl) << " != " << prev.effectedChanged(WHITE) << "\n";
#endif
	  return false;
	}
      }
    }
  }
  return true;
}

template <bool show_error>
bool
#if (defined __GNUC__) && (! defined GPSONE)
__attribute__ ((used))
#endif
osl::state::NumEffectState::isAlmostValidMove(Move move) const{
  assert(move.isValid());
  assert(move.isNormal());
  assert(this->getTurn() == move.player());
  assert(isValidMoveByRule(move, true));

  const Position from=move.from();
  const Position to=move.to();
  const Piece toPiece=this->getPieceAt(to);
  const Ptype ptype=move.ptype();
  const Player turn = move.player();
    
  if (from.isPieceStand()){ // Ǥļ
    // åȤ
    if (! toPiece.isEmpty()) {
      if (show_error) std::cerr << "drop on to piece : " << move << std::endl;
      return false;
    }
    // ⤽⤽ζäƤ뤫?
    if (! this->hasPieceOnStand(turn,ptype)) {
      if (show_error) std::cerr << turn << " don't have : " << ptype << std::endl;
      return false;
    }
    // Υå
    if (ptype==PAWN && isPawnMaskSet(turn, to.x())) {
      if (show_error) std::cerr << " Double Pawn : " << move << std::endl;
      return false;
    }
  }
  else{
    const Piece from_piece = this->getPieceAt(from);
    // fromˤΤζ
    if (from_piece.isEmpty() || (from_piece.owner() != turn)){
      if (show_error) 
	std::cerr << " No such piece0 : " << move << std::endl;
      return false;
    }
    // promoteƤpromoteǽ
    if (move.isPromote()){
      // fromˤΤζ
      if (from_piece.ptype() != unpromote(move.ptype()))
	{
	  if (show_error) 
	    std::cerr << " No such piece1  : " << move << std::endl;
	  return false;
	}
      if (from_piece.isPromotedNotKingGold()){
	if (show_error) 
	  std::cerr << " can't promote promoted piece : " << move << std::endl;
	return false;
      }
    }
    else{
      // fromˤΤζ
      if (from_piece.ptype() != move.ptype()){
	if (show_error) 
	  std::cerr << " No such piece2  : " << move << std::endl;
	return false;
      }
    }
    // toˤΤζ𤫶?
    if (!toPiece.isEmpty() && toPiece.owner()==turn) {
      if (show_error) std::cerr << " No move on  : " << move << std::endl;
      return false;
    }
    if(!hasEffectByPiece(from_piece,to)){
      if (show_error) {
	std::cerr << " No such move2 : " << move << std::endl;
      }
      return false;
    }
    // capturePtypeפƤ뤫?
    if (toPiece.ptype()!=move.capturePtype()) {
      if (show_error) std::cerr << " Not such capture : " << move 
				<< std::endl << *this;
      return false;
    }
  }

  assert(isValidMoveByRule(move, true));
  return true;
}

bool osl::state::NumEffectState::
isAlmostValidMove(Move move,bool show_error) const{
#ifdef MINIMAL
  show_error=false;
#endif
  if(show_error)
    return isAlmostValidMove<true>(move);
  else
    return isAlmostValidMove<false>(move);
}

#ifndef MINIMAL
void osl::state::NumEffectState::showEffect(std::ostream& os) const
{
  os<< static_cast<SimpleState const&>(*this);
  for(int y=1;y<=9;y++){
    os << 'P' << y;  
    for(int x=9;x>0;x--){
      Position pos(x,y);
      os << record::csa::show(getPieceAt(pos)) << getEffect(pos);
    }
    os << std::endl;
  }
  // ɽ
  for(int num=0;num<Piece::SIZE;num++){
    if (standMask(BLACK).test(num)){
      os << "P+00" << record::csa::show(Piece_Table.getPtypeOf(num))
	 << std::endl;
    }
    else if (standMask(WHITE).test(num)){
      os << "P-00" << record::csa::show(Piece_Table.getPtypeOf(num))
	 << std::endl;
    }
  }
}
#endif

osl::container::PieceMask osl::state::NumEffectState::
makePinOpen(osl::Position target,osl::Player defense)
{
  PieceMask pins;
  if(target.isPieceStand()) return pins;
  PieceMask mask=getOnBoardMask(alt(defense));
  makePinOpenDir<UL>(target,pins,mask,defense);
  makePinOpenDir<U>(target,pins,mask,defense);
  makePinOpenDir<UR>(target,pins,mask,defense);
  makePinOpenDir<L>(target,pins,mask,defense);
  makePinOpenDir<R>(target,pins,mask,defense);
  makePinOpenDir<DL>(target,pins,mask,defense);
  makePinOpenDir<D>(target,pins,mask,defense);
  makePinOpenDir<DR>(target,pins,mask,defense);
  return pins;
}

void osl::state::NumEffectState::
makePinOpen(osl::Player defense)
{
  pin_or_open[defense]=makePinOpen(getKingPosition(defense),defense);
}

namespace osl
{
  // explicit template instantiation

  template bool NumEffectState:: 
  hasEffectByWithRemove<BLACK>(Position, Position) const;
  template bool NumEffectState:: 
  hasEffectByWithRemove<WHITE>(Position, Position) const;

}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
